
/*
 * cis.c - user interface to XTVSIS based on X window.
 */

/*
 * Copyright (c) 1991, 1993 Old Dominion University.
 * Copyright (c) 1991, 1993 University of North Carolina at Chapel Hill.
 * All rights reserved.
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  Old Dominion University and the University of North
 * Carolina at Chapel Hill make no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 */
/*
 * written by : Sun W. Chang
 * date       : April, 1992
 * location   : Old Dominion University, Norfolk Virginia
 *
 * NOTE : All X window modules and parts were written by John Rives at ODU
 *     with minor changes to them in order to interface with network 
 *     modules written by Sun W. Chang.
 *     Unless specifically documented with authors name, all modules are
 *     written by Sun W. Chang.
 */
/* modifications made by  :  Pam Skillman
*  date                   :  December, 1992
*  location               :  Old Dominion University, Norfolk Virginia
*/
/* modifications made by  :  Angela M. Yeung
 * date                   :  April, 1993
 * location               :  Old Dominion University, Norfolk Virginia
 */

#define   DEBUGINF      0&DEBUGON



#include <stdio.h>
#include <errno.h>
#include <malloc.h>
#include <strings.h>
#include <string.h>
#include <signal.h>
#include <sys/wait.h>

#include <sys/time.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>

#include <netinet/in.h>
#include <netdb.h>

#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/List.h>
#include <X11/Xaw/Cardinals.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/Dialog.h>

#include "cis.h"

/***********  begin modification by cai ************/
#include <X11/Xutil.h>
#include "dflag.bitmap"
#include "rflag.bitmap"
#include "conf.icon.bitmap"

void     redisplay();
void     resize();
Pixmap create_flag();
void change_flag();
void SetHints();
void  OptionChange();
void  WMCallBack();
void  WMCancel();
void  WMAccept();


typedef struct {
   Pixmap     pix;
   GC         gc;
   Dimension  width, height;
} 
pixmap_data;

pixmap_data  data[2];
static int Update_FLAG = FALSE;
static int Icon_FLAG = FALSE;
static int NewData_FLAG = FALSE;
Widget       FlagForm;

/***********  end modification by cai ************/

/* by John Rives */
#include "cis_resources.h"
/* end John Rives */


#define   ALL_LIST_LINES      100
#define   DETAIL_LIST_LINES   100
#define RingBell   0

String   all_list[ALL_LIST_LINES];
String   detail_list[DETAIL_LIST_LINES];


void       StartXtvChairman();
void       JoinAction();
void       xtvsis_service();
void       send_REQUEST_SESSION_INFO();
XtCallbackProc    StartSessionReturn();
XtCallbackProc    JoinSessionReturn();
XtCallbackProc    WMReturn();
int      send_REQUEST_LAST_UPDATE_TIME();
void   sig_child();
/* functions for adding an intermediate step for get all conf info -- Angela Yeung */
void   DoneAllInfo();                         /* func to pop down a dialog */
void   Show_Diff();
void   update_conf_ls();                        /* update the current conf listing window */


/* struct of info database */
struct info_db{
   char             *conf_id;
   char             *info;
   struct info_db   *next;
};

struct info_db *build_db();                       /* func to build the info database */
struct info_db *NewList, *OldList;                /* list of all conferences */
struct info_db *AddList, *DelList, *ModList;   /* list of new, deleted and modified conferences */
char   conf_pkt[PACKET_SIZE];
/* End of Angela Yeung */


struct xtvsis_addr {
   struct sockaddr_in   addr;
   struct xtvsis_addr   *next;
} 
*xslist;

struct session {
   char                    *session_id;
   char                    *info;
   struct sockaddr_in      addr;
   time_t                  caltime;
   struct session     *next;
} 
*slist;

struct sockaddr_in   myaddr;
int         sockfd;
int         accept_XTVSIS_LIST;
int         accept_SESSION_INFO;
int         accept_LAST_UPDATE_TIME = TRUE;
time_t         last_update_time;
/*  begin modifications by Pam Skillman */
int                     XTVLIST_VERSION, last_version_from_xtvsis;
char                    LastVersion[VERSION_SIZE];
/*  end modifications by Pam Skillman  */

int                     First_Time= 1;   /* 1st time call send_REQ_SESS_INFO, Angela Yeung */

char dialog_str[256];      /* dialog box string     */
char WM_dialog_str[256];      /* WM dialog box string     */
char   *INITIAL_MSG_ALL = {
   "....... Contacting CIS Servers ......"};
char   *INITIAL_MSG_DETAIL = {
   "....................................."};
char   *NO_SESSION = {
   "No XTV Conferences are executing"};
char   *ALL_LIST_HEADER = {
   "XTV Information for ALL Conferences"};
char   *NEW_LIST_HEADER = {
   "XTV Information for NEW Conferences"};
char   *DELETED_LIST_HEADER = {
   "XTV Information for DELETED Conferences"};
char   *MODIFIED_LIST_HEADER = {
   "XTV Information for MODIFIED Conferences"};
char   *DETAIL_LIST_HEADER = {
   "XTV Information for Selected Conference"};
char   *NO_INFO = {
   "No Detailed Info Available"};


/* by John Rives */
XtAppContext       AppCon;        /* X application context           */

static XtActionsRec actionsTable[] = {
   {
      "StartSession",   (XtActionProc) StartSessionReturn   }
   ,
   {
      "JoinSession", (XtActionProc)   JoinSessionReturn   }
   ,
   {
      "WMChoice", (XtActionProc)   WMReturn   }
   ,
   NULL
};

/* toplevel */
Widget      topLevel;

/* form widgets */
Widget      TopForm;
Widget      ButtonForm;
Widget      AllSessionForm;
Widget      OneSessionForm;
Widget      NoSelectionForm;
Widget      NoDaemonForm;
Widget      QuitForm;
Widget      FlagForm;
Widget      AllInfoForm;

/* labels */
Widget      NoDaemonLabel;
Widget      NoSelectionLabel;
Widget      OptionsLabel;

/* command buttons */
Widget      StartButton;
Widget      AllInfoButton;
Widget      JoinButton;
Widget      OneSessionButton;
Widget      QuitButton;
Widget      AcceptStartButton;
Widget      CancelStartButton;
Widget      NoDaemonOKButton;
Widget      JoinAcceptButton;
Widget      JoinCancelButton;
Widget      OKButton;
Widget      NoQuitButton;
Widget      YesQuitButton;
Widget      WMButton;
Widget      AcceptWMButton;
Widget      CancelWMButton;
Widget		NewInfoButton;
Widget		DeletedInfoButton;
Widget		ModifiedInfoButton;
Widget		UpdateInfoButton;
Widget		DoneInfoButton;

/* dialog boxes */
Widget      StartDialog;
Widget      JoinDialog;
Widget      WMDialog;

/* transient popup shells */
Widget      TranShellStart;
Widget      TranShellNoDaemon;
Widget      TranShellJoin;
Widget      TranShellNoSelection;
Widget      TranShellQuit;
Widget      WMTranShell;
Widget      TranShellAllInfo;

/* list widget */
Widget      OneListWidget;
Widget      SessionList;

/* viewport widgets */
Widget      SessionListViewport;
Widget      OneViewport;

/* togle widgets, for otions (wahab) */
Widget Board;
Widget Talk;
Widget Virtual;
Widget Fast;


/* end John Rives */

/* Make this global so other functions can access it, Dr. wahab */
time_t                  last_update_time_from_xtvsis; 
/* options structure (wahab) */
struct {
   int b;                        /* 0 = board on, 1 = no board */
   int t;                        /* 0 = talk on, 1 = no board */
   int v;                        /* 0 = VS is on , 1 = no VS */
   int f;                        /* 0 = all on, 1 all off    */
} 
Option;


/***************************************************************************
 * main(argc, argv)
 **************************************************************************/
main(argc, argv)
int     argc;
char    *argv[];
{

   Arg     arg;

   last_update_time = time(NULL);
   last_update_time_from_xtvsis = time(NULL);

   build_xslist();

   if (xslist) {    /* create socket only when one or more XTVSIS */
      if (create_socket_for_xtvsis() < 0)     /* socket failed */
         xslist = NULL;          /* prevent any networking operation */
   }

   /*  begin modifications by Pam Skillman  */
   send_ID_LIST_REQUEST_TO_ALL();
   /*  end modifications by Pam Skillman  */

   send_REQUEST_XTVSIS_LIST();

   send_REQUEST_SESSION_INFO(NULL, NULL, NULL);

   alarm(TIMEOUT_LAST_UPDATE_TIME);
   signal(SIGALRM, send_REQUEST_LAST_UPDATE_TIME);


   /* by John Rives */
   topLevel = XtAppInitialize(&AppCon, "cis", NULL, 0,
   &argc, argv, FallbackResources, NULL, 0);
   XtSetArg(arg, XtNtitle, "XTV User Interface");
   XtSetValues(topLevel, &arg, 1);

   create_widgets();

   create_callbacks();

   XtRealizeWidget(topLevel);

   SetHints();

   XtAppMainLoop(AppCon);
   /* end John Rives */

}


/******************************************************************************
 * quitit((w, client_data, call_data) -- by John Rives
 *    displays popup menu to exit program
 *****************************************************************************/
static void quitit(w, client_data, call_data)
Widget w;
caddr_t client_data, call_data;
{
   int winx, winy;
   int rwinx, rwiny;
   unsigned int maskret;
   Window rootret, childret;
   Arg arg[2];

   XQueryPointer(XtDisplay(topLevel), XtWindow(topLevel), &rootret,
   &childret, &rwinx, &rwiny, &winx, &winy, &maskret);

   /* move the popup shell to pointer position */
   XtSetArg(arg[0], XtNx, rwinx);
   XtSetArg(arg[1], XtNy, rwiny);
   XtSetValues(TranShellQuit, arg, 2);

   XtSetSensitive(QuitButton, FALSE);
   XtPopup(TranShellQuit, XtGrabNonexclusive);
#if   RingBell
   XBell(XtDisplay(topLevel), 5);
# endif
}


/******************************************************************************
 * QuitAction(w, client_data, call_data) -- by John Rives
 *    determines result of exit popup menu
 ******************************************************************************/
static void QuitAction(w, client_data, call_data)
Widget w;
caddr_t client_data, call_data;
{
   XtPopdown(TranShellQuit);
   XtSetSensitive(QuitButton, TRUE);

   if (w == YesQuitButton) {
      if (xslist)
         close(sockfd);
      exit(0);
   }
}


/******************************************************************************
 * LGetArgs(CommandArgs,s) -- by John Rives
 *    parses string into command line argument
 ******************************************************************************/
void LGetArgs(CommandArgs,s)
char *CommandArgs[];
char *s;
{
   int i;

   i=0;
   while (*s) {   /* skip white spaces */
      while ((*s == ' ') || (*s == '\t'))
         *s++ = NULL;

      CommandArgs[i++] = s;

      if (*s == '"') {   /* quoted string - find matching end quote */
         CommandArgs[i-1] = ++s;   /* do not copy the starting '"' */
         while (*s != NULL && *s != '"')
            s++;
         if (*s == '"')
            *s++ = NULL;     /* do not copy the ending '"' */
      }
      else
         /* skip over this word to next white space */
         while ((*s != NULL) && (*s != ' ') && (*s != '\t'))
            s++;
   }

   CommandArgs[i]=NULL;
}


/******************************************************************************
 * StartSessionReturn((w, event) -- by John Rives
 *    callback when user preeses return key to start a session
 ******************************************************************************/
XtCallbackProc StartSessionReturn(w, event)
Widget w;
XEvent *event;
{
   StartXtvChairman(w, NULL, NULL);
}


/******************************************************************************
 * JoinSessionReturn(w, event) -- by John Rives
 *   callback when user preeses return key to join a session
******************************************************************************/
XtCallbackProc JoinSessionReturn(w, event)
Widget w;
XEvent *event;
{
   JoinAction(JoinAcceptButton, NULL, NULL);
}

/******************************************************************************
 * WMReturn((w, event) -- by wahab
 *      callback when user preeses return key to specify Window Manager
 ******************************************************************************/
XtCallbackProc WMReturn(w, event)
Widget w;
XEvent *event;
{
   WMAccept(w, NULL, NULL);
}



/******************************************************************************
 * NoDaemonOKCancel(w, client_data, call_data) -- by John Rives
 *    calback when OK button pressed on NoDaemon popup shell
 ******************************************************************************/
void NoDaemonOKCancel(w, client_data, call_data)
Widget w;
caddr_t client_data, call_data;
{
   XtPopdown(TranShellNoDaemon);
   XtSetSensitive(AllInfoButton, TRUE);
}


/******************************************************************************
 * OKCancel(w, client_data, call_data) -- by John Rives
 *    callback when user presses OK button
 ******************************************************************************/
void OKCancel(w, client_data, call_data)
Widget w;
caddr_t client_data, call_data;
{
   XtPopdown(TranShellNoSelection);
   /* this callback could be due to either one or join button */
   XtSetSensitive(OneSessionButton, TRUE);
   XtSetSensitive(JoinButton, TRUE);
}


/******************************************************************************
 * CancelStart(w, client_data, call_data) -- by John Rives
 *    callback when cancel button pressed for starting session
 ******************************************************************************/
void CancelStart(w, client_data, call_data)
Widget w;
caddr_t client_data, call_data;
{
   XtPopdown(TranShellStart);
   XtSetSensitive(StartButton, TRUE);
}

/******************************************************************************
 * DoneAllInfo(w, client_data, call_data) -- by wahab 
 *    callback when Done button pressed for AllInfo dialog
 ******************************************************************************/
void DoneAllInfo(w, client_data, call_data)
Widget w;
caddr_t client_data, call_data;
{
   XtPopdown(TranShellAllInfo);
   if (Update_FLAG)
   	XtSetSensitive(AllInfoButton, TRUE);
   else
	XtSetSensitive(AllInfoButton, FALSE);
}



/******************************************************************************
 * StartXtvChairman(w, client_data, call_data) -- by John Rives
 *    callback when Start button pressed for starting a session
 ******************************************************************************/
void StartXtvChairman(w, client_data, call_data)
Widget w;
caddr_t client_data, call_data;
{

   /* copy any arguments entered in the dialog box */
   if (*XawDialogGetValueString(StartDialog))
      sprintf(dialog_str, "xtv -S \"%s\" %s", 
      XawDialogGetValueString(StartDialog), WM_dialog_str);
   else
      sprintf(dialog_str, "xtv %s", WM_dialog_str);
   /* wahab */
# if DEBUGINF
   fprintf(stderr,"the value of strings WM and both are %s %s\n",
   WM_dialog_str, dialog_str);
# endif
   if(Option.v)
      strcat(dialog_str, " -V ");

   if(Option.b)
      strcat(dialog_str, " -B ");
   if(Option.t)
      strcat(dialog_str, " -T ");


   siginterrupt(SIGCHLD, 1);
   signal(SIGCHLD, sig_child);
   if (fork() == 0) {      /* child */
      char **CommandArgs[20];

      LGetArgs(CommandArgs, dialog_str);
      execvp("xtv", CommandArgs);
   }

   XtPopdown(TranShellStart);
   XtSetSensitive(StartButton, TRUE);
}


/******************************************************************************
 * JoinStart(w, client_data, call_data) -- by John Rives
 *    callback function called when Join button pressed 
 ***************************************************************************/
void JoinStart(w, client_data, call_data)
Widget w;
caddr_t client_data, call_data;
{
   int     i;
   int winx, winy;
   int rwinx, rwiny;
   unsigned int maskret;
   Window rootret, childret;
   Widget wid;
   Arg arg[2];
   XawListReturnStruct *list_item;

   /* ensure that the user has selected a session */

   list_item = XawListShowCurrent(SessionList);
   /* first item (list_indesx = 0) is  description of the fields, not valid */
   if (list_item->list_index == XAW_LIST_NONE || list_item->list_index == 0)
      wid = TranShellNoSelection;
   else
      wid = TranShellJoin;

   XQueryPointer(XtDisplay(topLevel), XtWindow(topLevel), &rootret,
   &childret, &rwinx, &rwiny, &winx, &winy, &maskret);

   /* move the popup shell to pointer position */
   i = 0;
   XtSetArg(arg[i], XtNx, rwinx);   
   i++;
   XtSetArg(arg[i], XtNy, rwiny);   
   i++;
   XtSetValues(wid, arg, i);

   XtSetSensitive(JoinButton, FALSE);
   XtPopup(wid, XtGrabNonexclusive);
#if   RingBell
   XBell(XtDisplay(topLevel), 5);
# endif
}


/******************************************************************************
 * JoinAction(w, client_data, call_data) -- by John Rives
 *    determines action for join popup menu
 ****************************************************************************/
void JoinAction(w, client_data, call_data)
Widget w;
caddr_t client_data, call_data;
{

   siginterrupt(SIGCHLD, 1);
   signal(SIGCHLD, sig_child);

   if (w == JoinAcceptButton) {
      XawListReturnStruct *list_item;

      list_item = XawListShowCurrent(SessionList);
      if (list_item->list_index == XAW_LIST_NONE || list_item->list_index == 0) {
         /* do nothing since no name has been seleted */
         /* should not get here, but to make sure     */
      }
      else if (fork() == 0) {      /* child */
         char *host_ptr;
         char *end_ptr;
         char  cmd_str[256];
         char **CommandArgs[20];

         /* find the hostname and session number */
         host_ptr = list_item->string;

         end_ptr = host_ptr;
         while (*end_ptr != ':')          /* find ':' delimeter    */
            end_ptr++;
         *end_ptr++ = ' ';                /* change `:` to a space */
         while (*++end_ptr != ' ');       /* find next space       */
         *end_ptr = '\0';                 /* null-terminate string */

         /* IMPORTANT!!! all arguments, e.g. -v, MUST preceed the  */
         /* hostname and sesion number for a participant!!!        */
         if (*XawDialogGetValueString(JoinDialog))
            sprintf(cmd_str, "xtv -M %s %s",
            XawDialogGetValueString(JoinDialog), host_ptr);
         else
            sprintf(cmd_str, "xtv %s", host_ptr);

         LGetArgs(CommandArgs, cmd_str);

         execvp("xtv", CommandArgs);
      }
   }

   XtPopdown(TranShellJoin);
   XtSetSensitive(JoinButton, TRUE);

}


/***************************************************************************
 * start(w, client_data, call_data) -- by John Rives
 *    displays popup menu when user presses start new session button
 ******************************************************************************/
void start(w, client_data, call_data)
Widget w;
caddr_t client_data, call_data;
{
   int      winx, winy, rwinx, rwiny;
   unsigned int    maskret;
   Window       rootret, childret;
   Arg       arg[2];

   XQueryPointer(XtDisplay(topLevel), XtWindow(topLevel), &rootret,
   &childret, &rwinx, &rwiny, &winx, &winy, &maskret);

   /* move the popup shell to pointer position */
   XtSetArg(arg[0], XtNx, rwinx);
   XtSetArg(arg[1], XtNy, rwiny);
   XtSetValues(TranShellStart, arg, 2);

   XtSetSensitive(StartButton, FALSE);
   XtPopup(TranShellStart, XtGrabNonexclusive);
#if   RingBell
   XBell(XtDisplay(topLevel), 5);
# endif
}

/***************************************************************************
 * startAllInfo(w, client_data, call_data) -- by wahab
 *    displays popup menu when user presses AllInfo button
 ******************************************************************************/
void startAllInfo(w, client_data, call_data)
Widget w;
caddr_t client_data, call_data;
{
   int      winx, winy, rwinx, rwiny;
   unsigned int    maskret;
   Window       rootret, childret;
   Arg       arg[2];

   XQueryPointer(XtDisplay(topLevel), XtWindow(topLevel), &rootret,
   &childret, &rwinx, &rwiny, &winx, &winy, &maskret);

   /* move the popup shell to pointer position */
   XtSetArg(arg[0], XtNx, rwinx);
   XtSetArg(arg[1], XtNy, rwiny);
   XtSetValues(TranShellAllInfo, arg, 2);

   XtSetSensitive(AllInfoButton, FALSE);
   XtPopup(TranShellAllInfo, XtGrabNonexclusive);
#if   RingBell
   XBell(XtDisplay(topLevel), 5);
# endif
}



/******************************************************************************
 * detail_session_info(w, client_data, call_data)
 *    callback when one session info button pressed
 ******************************************************************************/
void detail_session_info(w, client_data, call_data)
Widget w;
caddr_t client_data, call_data;
{
   XawListReturnStruct *list_item;

   list_item = XawListShowCurrent(SessionList);
   /* check for no selection of first item, which is list descriptoin */
   if (list_item->list_index == XAW_LIST_NONE || list_item->list_index == 0) {
      int   winx, winy, rwinx, rwiny;
      unsigned int maskret;
      Window rootret, childret;
      Arg arg[2];

      XQueryPointer(XtDisplay(topLevel), XtWindow(topLevel), &rootret,
      &childret, &rwinx, &rwiny, &winx, &winy, &maskret);

      /* move the popup shell to pointer position */
      XtSetArg(arg[0], XtNx, rwinx);
      XtSetArg(arg[1], XtNy, rwiny);
      XtSetValues(TranShellNoSelection, arg, 2);

      /* show no selection popup menu */
      XtSetSensitive(OneSessionButton, FALSE);
      XtPopup(TranShellNoSelection, XtGrabNonexclusive);
#if   RingBell
      XBell(XtDisplay(topLevel), 5);
# endif
   }
   else {
      char   *c1,*c2;
      int   i=0;
      struct session   *p;

      detail_list[i++] = DETAIL_LIST_HEADER;
      for (p=slist; p ; p=p->next)
         if (strcmp(p->session_id, list_item->string) == 0)
            break;

      if (p) {   /* found */
         char   info[PACKET_SIZE], *c1, *c2;
         char   chairman[100];
         char   subject[110];
         static char   *participant = {
            "PARTICIPANTS:\n"         };

         strcpy(info, p->info);
         c1 = info;

         /*
                   * assemble chairman line
                   */
         strcpy(chairman, "CHAIRMAN: ");
         for (c2=chairman; *c2 != '\0' ; c2++);
         for (; *c1 != ':' ; c1++,c2++)
            *c2 = *c1;
         *c2 = '\0';
         detail_list[i++] = chairman;

         /*
                   * assemble subject line
                   */
         strcpy(subject, "SUBJECT: ");
         for (c2=subject; *c2 != '\0' ; c2++);
         for (c1++; *c1 != ':' ; c1++,c2++)
            *c2 = *c1;
         *c2 = '\0';
         detail_list[i++] = subject;

         /*
                   * participants list
                   */
         detail_list[i++] = participant;
         detail_list[i++] = ++c1;
      }
      else {
         detail_list[i++] = NO_INFO;
      }

      XawListChange(OneListWidget, detail_list, i, 0, TRUE);
      XtSetSensitive(OneListWidget, TRUE);
      XtSetSensitive(OneViewport, TRUE);
#if   RingBell
      XBell(XtDisplay(topLevel), 5);
# endif
   }
}


/*****************************************************************************
 * send_REQUEST_SESSION_INFO(w, client_data, call_data)
 *   send REQUEST_SESSION_INFO when AllSessionButton is pressed.
 ******************************************************************************/
void send_REQUEST_SESSION_INFO(w, client_data, call_data)
Widget   w;
caddr_t   client_data, call_data;
{
   struct xtvsis_addr   *p;
   char         packet = REQUEST_SESSION_INFO;
   p = xslist;

   if (p == NULL) {
      int      winx, winy, rwinx, rwiny;
      unsigned int   maskret;
      Window      rootret, childret;
      Arg      arg[2];

      XQueryPointer(XtDisplay(topLevel), XtWindow(topLevel), &rootret,
      &childret, &rwinx, &rwiny, &winx, &winy, &maskret);

      /* move the popup shell to pointer position */
      XtSetArg(arg[0], XtNx, rwinx);
      XtSetArg(arg[1], XtNy, rwiny);
      XtSetValues(TranShellNoDaemon, arg, 2);

      XtSetSensitive(AllInfoButton, FALSE);
      
      XtPopup(TranShellNoDaemon, XtGrabNonexclusive);
#if   RingBell
      XBell(XtDisplay(topLevel), 5);
# endif
   }
   else if (NewData_FLAG || First_Time ) {
      for (p=xslist; p ; p=p->next)  
         sendto(sockfd, &packet, 1, 0,
         (struct sockaddr *) &(p->addr), sizeof(p->addr));
      accept_SESSION_INFO = TRUE;
   } else if (Update_FLAG && !First_Time)
	startAllInfo(AllInfoButton, NULL, NULL);
}


/******************************************************************************
 * create_widgets() -- by John Rives
 *    creates the widgets of the program
 ******************************************************************************/
create_widgets()
{
   int  i, n;
   char str[80];
   Arg  args[20];
   Arg          wargs[10];
   XGCValues    values;


   i = 0;

   XtSetValues(topLevel, NULL, 0);

   /* top form */
   TopForm = XtCreateManagedWidget( "TopForm", formWidgetClass, topLevel,
   NULL, 0);

   /* to buttom form and widgets */
   ButtonForm = XtCreateManagedWidget( "ButtonForm", formWidgetClass, TopForm,
   NULL, 0);

   StartButton = XtCreateManagedWidget( "StartButton", commandWidgetClass,
   ButtonForm, NULL, 0);

   AllInfoButton = XtCreateManagedWidget( "AllInfoButton",
   commandWidgetClass, ButtonForm, NULL, 0);
   XtSetSensitive(AllInfoButton, FALSE);

   QuitButton = XtCreateManagedWidget( "QuitButton", commandWidgetClass,
   ButtonForm, NULL, 0);

   /* second form and widgets */
   AllSessionForm = XtCreateManagedWidget( "AllSessionForm",
   formWidgetClass, TopForm, NULL, 0);

   JoinButton = XtCreateManagedWidget( "JoinButton", commandWidgetClass,
   AllSessionForm, NULL, 0);

   OneSessionButton = XtCreateManagedWidget( "OneSessionButton",
   commandWidgetClass, AllSessionForm, NULL, 0);

   SessionListViewport = 
       XtCreateManagedWidget( "SessionListViewport", viewportWidgetClass,
   AllSessionForm, NULL, 0);


   all_list[0] = INITIAL_MSG_ALL;

   i = 0;
   XtSetArg(args[i], XtNlist, (XtArgVal) all_list);  
   i++;
   XtSetArg(args[i], XtNnumberStrings, (XtArgVal) 1);  
   i++;
   SessionList = XtCreateManagedWidget( "SessionList", listWidgetClass,
   SessionListViewport, args, i );
   XtSetSensitive(SessionListViewport, TRUE);

   /* third form and widgets */
   OneSessionForm = XtCreateManagedWidget( "OneSessionForm",
   formWidgetClass, TopForm, NULL, 0);

   OneViewport = 
       XtCreateManagedWidget( "OneViewport", viewportWidgetClass,
   OneSessionForm, NULL, 0 );

  i = 0;
   detail_list[0] = INITIAL_MSG_DETAIL;
   XtSetArg(args[i], XtNnumberStrings, (XtArgVal) 1);  i++;
   XtSetArg(args[i], XtNlist, (XtArgVal) detail_list);  i++;
   OneListWidget =
         XtCreateManagedWidget( "OneListWidget", listWidgetClass,
                                     OneViewport, args, i );


   /*****************  begin modification by cai *******************/

   /* fourth form and widgets */

   /*
      * Create a widget in which to display the flag.
      */

   FlagForm = XtVaCreateManagedWidget("FlagForm",
   labelWidgetClass,TopForm,
   XtNheight,   64,
   XtNwidth,    64,
   NULL);

   /*
     * Use the foreground and background colors
     * of the canvas to create a graphics context.
     */

   n = 0;
   XtSetArg(wargs[n], XtNforeground, &values.foreground);
   n++;
   XtSetArg(wargs[n], XtNbackground, &values.background);
   n++;
   XtGetValues(FlagForm, wargs, n);
   data[0].gc = XtGetGC(FlagForm, GCForeground | GCBackground, &values);
   data[1].gc = XtGetGC(FlagForm, GCForeground | GCBackground, &values);  
      /*
      * Create the pixmap conatinign the flag. Store the 
      * pixmap, as well as the size of the pixmap in the struct.
      */
     data[1].width = rflag_width;
     data[1].height = rflag_height;
     data[1].pix = create_flag(FlagForm, data[0].gc, rflag_bits, 
                            rflag_width, rflag_height );
   
     data[0].width = rflag_width;
     data[0].height = rflag_height;
     data[0].pix = create_flag(FlagForm, data[1].gc, dflag_bits, 
                            dflag_width, dflag_height );
   
   /*****************  end modification by cai *******************/

   /* create transient shell for join xtv session button */

   /* transient shell should always be unmapped when created */
   i = 0;
   XtSetArg(args[i], XtNmappedWhenManaged, (XtArgVal) FALSE);  
   i++;
   TranShellJoin = XtCreatePopupShell("TranShellJoin",
   transientShellWidgetClass, topLevel, args, i);


   i = 0;
   XtSetArg(args[i], XtNvalue, dialog_str);  
   i++;
   JoinDialog = XtCreateManagedWidget( "JoinDialog", dialogWidgetClass,
   TranShellJoin, args, i);

   JoinAcceptButton = XtCreateManagedWidget( "JoinAcceptButton",
   commandWidgetClass, JoinDialog, NULL, 0);

   JoinCancelButton = XtCreateManagedWidget( "JoinCancelButton",
   commandWidgetClass, JoinDialog, NULL, 0);


  /* create transient shell for start xtv session button */

   /* transient shell should always be unmapped when created */
   i = 0;
   XtSetArg(args[i], XtNmappedWhenManaged, (XtArgVal) FALSE);
   i++;
   TranShellAllInfo = XtCreatePopupShell("TranShellAllInfo",
   transientShellWidgetClass, topLevel, args, i);

   AllInfoForm = XtCreateManagedWidget( "AllInfoForm", formWidgetClass,
   TranShellAllInfo, NULL, 0);

   NewInfoButton = XtCreateManagedWidget( "NewInfoButton",
   commandWidgetClass, AllInfoForm, NULL, 0);
   DeletedInfoButton = XtCreateManagedWidget( "DeletedInfoButton",
   commandWidgetClass, AllInfoForm, NULL, 0);
   ModifiedInfoButton = XtCreateManagedWidget( "ModifiedInfoButton",
   commandWidgetClass, AllInfoForm, NULL, 0);
   UpdateInfoButton = XtCreateManagedWidget( "UpdateInfoButton",
   commandWidgetClass, AllInfoForm, NULL, 0);
   DoneInfoButton = XtCreateManagedWidget( "DoneInfoButton",
   commandWidgetClass, AllInfoForm, NULL, 0);



  /* Add Call back for each button */

   XtAddCallback(NewInfoButton, XtNcallback, Show_Diff, NULL);
   XtAddCallback(DeletedInfoButton, XtNcallback, Show_Diff, NULL);
   XtAddCallback(ModifiedInfoButton, XtNcallback, Show_Diff, NULL);
   XtAddCallback(UpdateInfoButton, XtNcallback, update_conf_ls, conf_pkt);
   XtAddCallback(DoneInfoButton, XtNcallback, DoneAllInfo, NULL);
   /* create transient shell for start xtv session button */

   /* transient shell should always be unmapped when created */
   i = 0;
   XtSetArg(args[i], XtNmappedWhenManaged, (XtArgVal) FALSE);  
   i++;
   TranShellStart = XtCreatePopupShell("TranShellStart",
   transientShellWidgetClass, topLevel, args, i);

   i = 0;
   XtSetArg(args[i], XtNvalue, dialog_str);  
   i++;
   StartDialog = XtCreateManagedWidget( "StartDialog", dialogWidgetClass,
   TranShellStart, args, i);

   AcceptStartButton = XtCreateManagedWidget( "AcceptStartButton",
   commandWidgetClass, StartDialog, NULL, 0);
   CancelStartButton = XtCreateManagedWidget( "CancelStartButton",
   commandWidgetClass, StartDialog, NULL, 0);

   OptionsLabel = XtCreateManagedWidget(
   "OptionsLabel",
   labelWidgetClass,
   StartDialog,
   NULL,
   0
       );
   n = 0;
   XtSetArg(args[n], XtNfromVert, AcceptStartButton);
   n++;
   XtSetArg(args[n], XtNfromHoriz, NULL);
   n++;
   XtSetValues(OptionsLabel, args, n);

   Option.b = 0;
   Option.t = 0;
   Option.v = 0;
   Option.f = Option.b && Option.t && Option.v;


   Board =  XtCreateManagedWidget( "BoardButton",
   toggleWidgetClass, StartDialog, NULL, 0);
   n = 0;
   XtSetArg(args[n], XtNfromVert, OptionsLabel);
   n++;
   XtSetArg(args[n], XtNfromHoriz, NULL );
   n++;
   XtSetArg(args[n], XtNstate, Option.b);
   n++;
   XtSetValues(Board, args, n);
   XtAddCallback(Board, XtNcallback, OptionChange, 'B');

   Talk =  XtCreateManagedWidget( "TalkButton",
   toggleWidgetClass, StartDialog, NULL, 0);
   n = 0;
   XtSetArg(args[n], XtNfromVert, OptionsLabel);
   n++;
   XtSetArg(args[n], XtNfromHoriz,  Board);
   n++;
   XtSetArg(args[n], XtNstate, Option.t);
   n++;
   XtSetValues(Talk, args, n);
   XtAddCallback(Talk, XtNcallback, OptionChange, 'T');

   Fast =  XtCreateManagedWidget( "FastButton",
   toggleWidgetClass, StartDialog, NULL, 0);
   n = 0;
   XtSetArg(args[n], XtNfromVert, Talk);
   n++;
   XtSetArg(args[n], XtNfromHoriz,  Board);
   n++;
   XtSetArg(args[n], XtNstate, Option.f);
   n++;
   XtSetValues(Fast, args, n);
   XtAddCallback(Fast, XtNcallback, OptionChange, 'F');

   Virtual =  XtCreateManagedWidget( "VirtualButton",
   toggleWidgetClass, StartDialog, NULL, 0);
   n = 0;
   XtSetArg(args[n], XtNfromVert, OptionsLabel);
   n++;
   XtSetArg(args[n], XtNfromHoriz,  Talk);
   n++;
   XtSetArg(args[n], XtNstate, Option.v);
   n++;
   XtSetValues(Virtual, args, n);
   XtAddCallback(Virtual, XtNcallback, OptionChange, 'V');

   WMButton = XtCreateManagedWidget("WMButton", commandWidgetClass, 
   StartDialog, NULL, 0);

   n = 0;
   XtSetArg(args[n], XtNmappedWhenManaged, (XtArgVal) FALSE);  
   n++;
   WMTranShell = XtCreatePopupShell("WMTranShell",
   transientShellWidgetClass, topLevel, args, n);

   n = 0;
   XtSetArg(args[n], XtNvalue, WM_dialog_str);  
   n++;
   WMDialog = XtCreateManagedWidget( "WMDialog", dialogWidgetClass,
   WMTranShell, args, n);

   n = 0;
   XtSetArg(args[n], XtNfromVert, Virtual);
   n++;
   XtSetArg(args[n], XtNfromHoriz,  Fast);
   n++;
   XtSetValues(WMButton, args, n);


   XtAddCallback(WMButton, XtNcallback, WMCallBack, NULL);

   AcceptWMButton = XtCreateManagedWidget( "AcceptWMButton",
   commandWidgetClass, WMDialog, NULL, 0);

   CancelWMButton = XtCreateManagedWidget( "CancelWMButton",
   commandWidgetClass, WMDialog, NULL, 0);


   XtAddCallback(AcceptWMButton, XtNcallback, WMAccept, NULL);
   XtAddCallback(CancelWMButton, XtNcallback, WMCancel, NULL);








   /* create transient shell for no session choosen warning message */

   /* transient shell should always be unmapped when created */
   i = 0;
   XtSetArg(args[i], XtNmappedWhenManaged, (XtArgVal) FALSE);  
   i++;
   TranShellNoSelection = XtCreatePopupShell("TranShellNoSelection",
   transientShellWidgetClass, topLevel, args, i);

   NoSelectionForm = XtCreateManagedWidget( "NoSelectionForm", formWidgetClass,
   TranShellNoSelection, NULL, 0);

   NoSelectionLabel = XtCreateManagedWidget( "NoSelectionLabel",
   labelWidgetClass, NoSelectionForm, NULL, 0);

   OKButton = XtCreateManagedWidget( "OKButton", commandWidgetClass,
   NoSelectionForm, NULL, 0);

   /* create transient shell for no daemon running */

   /* transient shell should always be unmapped when created */
   i = 0;
   XtSetArg(args[i], XtNmappedWhenManaged, (XtArgVal) FALSE);  
   i++;
   TranShellNoDaemon = XtCreatePopupShell("TranShellNoDaemon",
   transientShellWidgetClass, topLevel, args, i);

   NoDaemonForm = XtCreateManagedWidget( "NoDaemonForm", formWidgetClass,
   TranShellNoDaemon, NULL, 0);

   NoDaemonLabel = XtCreateManagedWidget( "NoDaemonLabel",
   labelWidgetClass, NoDaemonForm, NULL, 0);

   NoDaemonOKButton = XtCreateManagedWidget( "NoDaemonOKButton",
   commandWidgetClass, NoDaemonForm, NULL, 0);

   /* create transient shell for quit button */

   /* transient shell should always be unmapped when created */
   i = 0;
   XtSetArg(args[i], XtNmappedWhenManaged, (XtArgVal) FALSE);  
   i++;
   TranShellQuit = XtCreatePopupShell("TranShellQuit",
   transientShellWidgetClass, topLevel, args, i);

   QuitForm = XtCreateManagedWidget( "QuitForm", formWidgetClass,
   TranShellQuit, NULL, 0);

   YesQuitButton = XtCreateManagedWidget( "YesQuitButton",
   commandWidgetClass, QuitForm, NULL, 0);

   NoQuitButton = XtCreateManagedWidget( "NoQuitButton", commandWidgetClass,
   QuitForm, NULL, 0);
}


/******************************************************************************
 * create_callbacks() -- by John Rives
 *    create callbacks for the widgets 
 ******************************************************************************/
create_callbacks()
{
   XtAddCallback(QuitButton, XtNcallback, quitit, NULL);
   XtAddCallback(StartButton, XtNcallback, start, NULL);
   XtAddCallback(AllInfoButton, XtNcallback, send_REQUEST_SESSION_INFO, NULL);
   XtAddCallback(JoinButton, XtNcallback, JoinStart, NULL);
   XtAddCallback(OneSessionButton, XtNcallback, detail_session_info, NULL);
   XtAddCallback(JoinAcceptButton, XtNcallback, JoinAction, NULL);
   XtAddCallback(JoinCancelButton, XtNcallback, JoinAction, NULL);
   XtAddCallback(AcceptStartButton, XtNcallback, StartXtvChairman, NULL);
   XtAddCallback(CancelStartButton, XtNcallback, CancelStart, NULL);
   XtAddCallback(OKButton, XtNcallback, OKCancel, NULL);
   XtAddCallback(NoDaemonOKButton, XtNcallback, NoDaemonOKCancel, NULL);
   XtAddCallback(YesQuitButton, XtNcallback, QuitAction, NULL);
   XtAddCallback(NoQuitButton, XtNcallback, QuitAction, NULL);

   /* add callback for input from the datagram socket */
   XtAppAddInput(AppCon, sockfd, XtInputReadMask, xtvsis_service, NULL);

   XtAppAddActions(AppCon, actionsTable, XtNumber(actionsTable));

   /*****************  begin modification by playai *******************/
   /* add event handler  for  flag   use  */
   XtAddEventHandler(FlagForm, ExposureMask,False,redisplay, NULL); 
   XtAddEventHandler(FlagForm, ResizeRedirectMask,False, resize,NULL); 
   /*****************  end modification by cai *******************/

}


/*************************************************************************
 * build_xslist()
 *      builds XTVSIS list from "xtvsis.list" file.
 *      "xtvsis.list" contains hostname:port per line.
 *
 *      calling routines : main() -- when started
 *                         xtvsis_service() -- when new XTVSIS_LIST packet
 *                                      is received from a XTVSIS.
 *************************************************************************/
build_xslist()
{
   FILE                    *fp;
   char                    aline[80], *c1, *c2;
   char                    hostname[HOSTNAME_SIZE];
   int                     port;
   struct hostent          *gethostbyname();
   struct hostent          *host;
   struct xtvsis_addr      *p;

   xslist = NULL;

   if ((fp=fopen(XTVSIS_LIST_file, "r")) == NULL)
      return;

   for (; fgets(aline, 80, fp) ;) {
      /*  begin modifications by Pam Skillman  */
      /*  first line in file is version number  */

      if (strstr(aline,":") == NULL)  {
         XTVLIST_VERSION = atoi(aline);

         /*  end modifications by Pam Skillman  */
         continue;  
      }

      for (c1=aline, c2=hostname; *c1!=':' ; c1++,c2++)
         *c2 = *c1;
      *c2 = '\0';
      port = atoi(++c1);

      /*
                       * hostname can be either dotted string of inet address or
                       * hostname
                      */
      p = (struct xtvsis_addr *) malloc(sizeof(struct xtvsis_addr));
      bzero((char *) &(p->addr), sizeof(p->addr));
      p->addr.sin_family      = AF_INET;
      if ((host=gethostbyname(hostname)) != NULL)
         bcopy(host->h_addr, (char *)&(p->addr.sin_addr), host->h_length
             );
      else
         p->addr.sin_addr.s_addr = inet_addr(hostname);
      p->addr.sin_port        = htons(port);
      p->next                 = NULL;

      if (xslist) {
         p->next = xslist;
         xslist = p;
      }
      else
         xslist = p;
   }

   fclose(fp);
}


/**************************************************************************
 * create_socket_for_xtvsis()
 *      creates socket to be used in XTVSIS interface.
 *
 *      calling routines : main() -- at the beginning.
 *************************************************************************/
create_socket_for_xtvsis()
{
   int     length;

   if ((sockfd=socket(AF_INET, SOCK_DGRAM, 0)) < 0)
      return(-1);

   bzero((char *) &myaddr, sizeof(myaddr));
   myaddr.sin_family       = AF_INET;
   myaddr.sin_addr.s_addr  = htonl(INADDR_ANY);
   myaddr.sin_port         = htons(0);

   if (bind(sockfd, (struct sockaddr *) &myaddr, sizeof(myaddr)) < 0)
      return(-1);

   /*
            * bind itself does not fill myaddr structure with values
            * getsockname call is made to fill myaddr
            */
   length = sizeof(struct sockaddr);
   if (getsockname(sockfd, (struct sockaddr *) &myaddr, &length) < 0)
      return(-1);

}


/**************************************************************************
 * send_REQUEST_XTVSIS_LIST()
 *************************************************************************/
send_REQUEST_XTVSIS_LIST()
{
   char         packet = REQUEST_XTVSIS_LIST;
   struct xtvsis_addr   *p;

   for (p=xslist; p ; p=p->next)
      sendto(sockfd, &packet, 1, 0,
      (struct sockaddr *) &(p->addr), sizeof(p->addr));

   accept_XTVSIS_LIST = TRUE;
}


/**************************************************************************
 * send_REQUEST_LAST_UPDATE_TIME()
 *************************************************************************/
send_REQUEST_LAST_UPDATE_TIME()
{
   struct xtvsis_addr   *p;
   /*  begin modifications by Pam Skillman  */
   time_t                  NetTime;
   char                    packet[sizeof(time_t) + 2];

   /*  send last_update_time with REQUEST_LAST_UPDATE_TIME packet  */
   packet[0] = REQUEST_LAST_UPDATE_TIME;
   NetTime = htonl(last_update_time_from_xtvsis);
   bcopy((char *) &(NetTime), &packet[1], sizeof(time_t));
   packet[sizeof(time_t)+1] = '\0';
   /*  end modifications by Pam Skillman  */

   for (p=xslist; p ; p=p->next)
      sendto(sockfd, packet, strlen(packet), 0,
      (struct sockaddr *) &(p->addr), sizeof(p->addr));

   accept_LAST_UPDATE_TIME = TRUE;
   alarm(TIMEOUT_LAST_UPDATE_TIME);
   signal(SIGALRM, send_REQUEST_LAST_UPDATE_TIME);
}


/**************************************************************************
 * xtvsis_service(client_data,sockfd, id)
 *      receives packet and calls proper routine according to the packet
 *      type. accept_XTVSIS_LIST is set to TRUE at the beginning but
 *      set to FALSE as soon as the first XTVSIS_LIST packet is received to
 *      discard any subsequent XTVSIS_LIST from other XTVSIS's.
 *      accept_SESSION_INFO and accept_LAST_UPDATE_TIME are set to TRUE
 *      at the beginning and set to FALSE when SESSION_INFO and
 *      LAST_UPDATE_TIME respectively are received. they are set to TRUE
 *      again whenever send_REQUEST_SESSION_INFO and send_LAST_UPDATE_TIME
 *      respectively are executed.
 *
 *      calling routines : from XtInputReadMask -- when input is signaled
 *                              in sockfd.
 *************************************************************************/
void xtvsis_service(client_data, sfd, id)
caddr_t client_data;
int   *sfd;
XtInputId *id;
{
   int                     rc, fromlen;
   char                    packet[PACKET_SIZE];
   struct sockaddr_in      fromaddr;
   time_t         LastTime;

   fromlen = sizeof(fromaddr);
   if ((rc=recvfrom(*sfd, packet, PACKET_SIZE, 0,
   (struct sockaddr *) &fromaddr, &fromlen)) < 0) {
      printf("error recvfrom\n");
      return;
   }
   packet[rc] = '\0';

   switch (packet[0]) {

   case SESSION_INFO :
      if (accept_SESSION_INFO) {

         /*****************  begin modification by cai *******************/
         NewData_FLAG = FALSE;
         if (Icon_FLAG) {
         	Icon_FLAG = FALSE;
   	 	change_flag ();
         }
         /*****************  end modification by cai *******************/


         accept_SESSION_INFO = FALSE;
                                         XtSetSensitive(AllInfoButton, FALSE);
         
                                         /** modifed by Angela Yeung **/
         if (First_Time){
            /* show the conference before this interface starts */
            First_Time= 0;
            process_SESSION_INFO(&packet[1]);
            OldList= build_db(packet);
         }
         else
            /* add an intermediate step before update the list */
            Handle_Diff(packet);
         /** end Angela Yeung **/
      }
      accept_SESSION_INFO = FALSE;
      break;

   case XTVSIS_LIST :
      if (accept_XTVSIS_LIST) {
         FILE    *fp;

         accept_XTVSIS_LIST = FALSE;
         if ((fp=fopen(XTVSIS_LIST_file, "w")) != NULL) {
            fputs(&packet[1], fp);
            fclose(fp);
            build_xslist();
         }
      }
      accept_XTVSIS_LIST = FALSE;
      break;

   case LAST_UPDATE_TIME :
      bcopy(&packet[1], (char *)&LastTime, sizeof(time_t));
      last_update_time_from_xtvsis = ntohl(LastTime); 
      if (accept_LAST_UPDATE_TIME) {
         accept_LAST_UPDATE_TIME = FALSE;
         if (last_update_time_from_xtvsis > last_update_time) {

            /*****************  begin modification by cai *******************/
            NewData_FLAG = TRUE;
            if (!Icon_FLAG) {
            	Icon_FLAG = TRUE;
            	change_flag ();
            }
            /*****************  end modification by cai *******************/
            last_update_time = last_update_time_from_xtvsis;
            XtSetSensitive(AllInfoButton, TRUE);
#if   RingBell
            XBell(XtDisplay(topLevel), 5);
# endif
         }
      }
      accept_LAST_UPDATE_TIME = FALSE;
      break;

      /*  begin modifications by Pam Skillman  */

   case ID_LIST :

      bcopy(&packet[1], LastVersion, sizeof(LastVersion));
      last_version_from_xtvsis = atoi(LastVersion);
      /*  only update our XTVSIS_LIST if our version # is lower  */


      if (last_version_from_xtvsis > XTVLIST_VERSION) {
         FILE       *fp;
         XTVLIST_VERSION = last_version_from_xtvsis;
         if ((fp=fopen(XTVSIS_LIST_file, "w")) != NULL)
         {
            fputs(&packet[1], fp);
            fclose(fp);
            build_xslist();
         }
      }

      break;
      /*  end modifications by Pam Skillman  */
   default :
      break;

   }
}


/**************************************************************************
 * process_SESSION_INFO(packet)
 *      updates
 *
 *      calling routines : xtvxix_service -- when SESSION_INFO packet is
 *                                      received from a XTVSIS.
 *************************************************************************/
process_SESSION_INFO(packet)
char *packet;
{
   int             line_count;
   struct session   *p,*q;
   time_t LastTime;


   bcopy(packet, (char *)&LastTime, sizeof(time_t));
   last_update_time = ntohl(LastTime);
   packet += sizeof(time_t);

   for (p=slist; p ; p=q) {
      q = p->next;
      free(p->session_id);
      free(p->info);
   }

   slist = NULL;

   for (; *packet != '\0' ; packet++) {
      char   session_id[133], *c;
      char   info[PACKET_SIZE];

      for (c=session_id; *packet != END_OF_SESSION_ID ; packet++, c++)
         *c = *packet;
      *c = '\0';

      p = (struct session *) malloc(sizeof(struct session));
      p->session_id = malloc((size_t) strlen(session_id)+1);
      strcpy(p->session_id, session_id);

      for (c=info, packet++; *packet != END_OF_SESSION ; packet++, c++)
         *c = *packet;
      *c = '\0';

      p->info = malloc((size_t) strlen(info)+1);
      strcpy(p->info, info);
      p->next = NULL;

      if (slist) {
         p->next = slist;
         slist = p;
      }
      else
         slist = p;
   }

   for (line_count=0 ; line_count<ALL_LIST_LINES ; line_count++)
      all_list[line_count] = NULL;
   all_list[0] = ALL_LIST_HEADER;

   if (slist) {
      for (line_count=1,p=slist; p ; p=p->next, line_count++) {
         all_list[line_count] = p->session_id; 
      }
      XtSetSensitive(JoinButton, TRUE);
      XtSetSensitive(OneSessionButton, TRUE);
   }
   else {
      all_list[0] = NO_SESSION;
      line_count = 1;
      XtSetSensitive(JoinButton, FALSE);
      XtSetSensitive(OneSessionButton, FALSE);

   }

   XawListChange(SessionList, all_list, line_count, 0, TRUE);
   XtSetSensitive(SessionListViewport, TRUE);
#if   RingBell
   XBell(XtDisplay(topLevel), 5);
# endif
}
/* handling zombies */
void sig_child()
{
   int pid;
   union wait   status;


   while ( (pid=wait3(&status, WNOHANG, (struct rusage *) 0)) > 0){
      ( stderr, "sig_child(): child process have terminated\n");
   }

}
/************************************************************************
 * send_ID_LIST_REQUEST(fromaddr)
 *      sends ID_LIST_REQUEST to fromaddr
 *
 *      calling routines : send_ID_LIST_REQUEST_TO_ALL()
 *
 *      ID_LIST_REQUEST packet :
 *              ID_LIST_REQUEST
 *              version
 *
 ****************************************************************************/
send_ID_LIST_REQUEST(fromaddr)
struct sockaddr_in      fromaddr;
{
   int     fd;
   char    c, *s;
   char    packet[PACKET_SIZE];

   packet[0] = ID_LIST_REQUEST;
   sprintf(LastVersion,"%d",XTVLIST_VERSION);
   bcopy(LastVersion, &packet[1], sizeof(LastVersion));

   packet[sizeof(LastVersion)+1] = '\0';

   sendto(sockfd, packet, strlen(packet), 0,
   (struct sockaddr *) &fromaddr, sizeof(fromaddr));

}
/************************************************************************
 * send_ID_LIST_REQUEST_TO_ALL()
 *      sends ID_LIST_REQUEST to all daemons when we start up
 *
 *      calling routines : main()
 *
 *      called routines  : send_ID_LIST_REQUEST(fromaddr)  --  send a request
 *                                    to all XTVSIS daemons.
 *
 ****************************************************************************/
send_ID_LIST_REQUEST_TO_ALL()
{
   FILE                    *fp;
   char                    c, *s, localhost[80];
   char                    aline[80], *c1, *c2;
   char                    hostname[HOSTNAME_SIZE];
   int                     port, n;
   struct hostent          *gethostbyname();
   struct hostent          *host, *hp;
   struct sockaddr_in      fromaddr, ouraddr;


   if ((fp=fopen(XTVSIS_LIST_file, "r")) == NULL) {
      return;
   }
   for (; fgets(aline, 80, fp) ;) {
      if (strstr(aline, ":") != NULL)  {

         for (c1=aline, c2=hostname; *c1!=':' ; c1++,c2++)
            *c2 = *c1;
         *c2 = '\0';
         port = atoi(++c1);
         /*
                          * hostname can be either dotted string of inet address or
                          * hostname
                         */

         bzero((char *) &fromaddr, sizeof(fromaddr));
         fromaddr.sin_family      = AF_INET;

         if ((host=gethostbyname(hostname)) != NULL)
            bcopy(host->h_addr, &(fromaddr.sin_addr), host->h_length);
         else
            fromaddr.sin_addr.s_addr = inet_addr(hostname);
         fromaddr.sin_port        = htons(port);

         gethostname(localhost, HOSTNAME_SIZE);

         if ((hp=gethostbyname(localhost)) == NULL)  {
            continue;
         }
         bcopy(hp->h_addr, &(ouraddr.sin_addr), hp->h_length);

         /*  We want to send to ALL daemons, even our own     */

         /*                if ((n = strcmp(fromaddr.sin_addr,ouraddr.sin_addr)) !=0)  */
         send_ID_LIST_REQUEST(fromaddr);
      }
   }
   fclose(fp);
}
/*  end modifications by Pam Skillman  */

/*****************  begin modification by cai *******************/

/************************************************************************
 *      called routines  : change_flag()
 *      called routines  : create_flag()   
 *      called routines  : rediplay()
 *      called routines  : resize()
 *      called routines  : SetHints()
 ***********************************************************************/
void change_flag()
{
   XBell(XtDisplay(FlagForm),0);
   redisplay (FlagForm, NULL, NULL);
}

Pixmap create_flag(w, gc, bits, width, height)
Widget   w;
char    *bits;
int      width, height;
{
   Pixmap bitmap;

   /* 
      * Create a bitmap containing the flag.
      */
   bitmap=XCreateBitmapFromData(XtDisplay(w), 
   RootWindowOfScreen(XtScreen(w)),
   bits, width, height);
   return bitmap;
}

void redisplay (w, clidata, call_data)
Widget         w;    
pixmap_data    *clidata;
caddr_t       call_data;
{
   Arg wargs[10];
   int n;
   Dimension widget_width, widget_height;
   pixmap_data    *cdata;
   /* 
      * Get the current size of the widget window.
      */
   cdata = &data[Icon_FLAG];
   n = 0;
   XtSetArg(wargs[n], XtNwidth,  &widget_width);
   n++;
   XtSetArg(wargs[n], XtNheight, &widget_height);
   n++;
   XtGetValues(w, wargs, n);
   /*
      * Copy plane 1 of the bitmap to the center
      * of the window, using the widget's foreground 
      * and background color.
      */
   XCopyPlane(XtDisplay(w), cdata->pix, XtWindow(w), cdata->gc,
   0, 0,  cdata->width, cdata->height, 
   (widget_width - cdata->width)/2,  
   (widget_height - cdata->height)/2, 1);
}


void resize (w, data, call_data)
Widget                       w;
caddr_t                      data;
caddr_t              call_data;
{
   if(XtIsRealized(w))
      XClearArea(XtDisplay(w), XtWindow(w), 0, 0, 0, 0, TRUE);
}

/*  set icon image   */
void SetHints()
{
   Display   *dpy;
   XWMHints   wm_hints, *hints;
   Pixmap         icon;
   XSizeHints              Hint;

   dpy = XtDisplay(topLevel);
   hints = XGetWMHints(dpy, XtWindow(topLevel));

   wm_hints.input = True;
   wm_hints.initial_state = hints->initial_state;   /* take care of iconic state */
   wm_hints.icon_x = hints->icon_x;      /* restore icon position */
   wm_hints.icon_y = hints->icon_y;
   wm_hints.flags  = InputHint | StateHint;

   XFree(hints);

   wm_hints.flags |= IconPixmapHint;
   wm_hints.icon_pixmap = XCreateBitmapFromData(dpy, XtScreen(topLevel)->root,
   conf_bits, conf_width, conf_height);

   XSetWMHints(dpy, XtWindow(topLevel), &wm_hints);


} /* SetHints */

/*****************  end modification by cai *******************/

/**  The following functions are added by Angela Yeung    **/

/***************************************************************************************
 int Handle_Diff(pkt_buf)       -- Angela Yeung
 
 Compare the Difference bewteen last and current packet.  Call functions to
 create a Motif bulletin board to interface with user.  Add a intermediate step before
 update the Athena conference listing.  

 This function is called by  xtvsis_service().
 Input parameter:  pkt_buf      packet got from the communication stuff 
 Functions called: build_db, cmp_key, cmp_info, create_dialog, show_dialog  
***************************************************************************************/
int Handle_Diff(pkt_buf)
char  *pkt_buf;                /* packet buffer from communication */
{


# if DEBUGINF
	fprintf( stderr,"Handle_Diff:Staring\n");
# endif

   /* build database link list base on packet */
   NewList= build_db(pkt_buf);     

   /* get the list of new conferences, deleted conference and modified cconference */
   cmp_key(NewList, OldList, &AddList);  
   if (AddList != NULL)
   	XtSetSensitive(NewInfoButton, TRUE);
   else
	XtSetSensitive(NewInfoButton, FALSE);
   cmp_key(OldList, NewList, &DelList);
   if (DelList != NULL)
   	XtSetSensitive(DeletedInfoButton, TRUE);
   else
	XtSetSensitive(DeletedInfoButton, FALSE);
   cmp_info(NewList, OldList, &ModList);
   if (ModList != NULL)
        XtSetSensitive(ModifiedInfoButton, TRUE);
   else
        XtSetSensitive(ModifiedInfoButton, FALSE);


   strcpy(conf_pkt, pkt_buf);

   if ((AddList != NULL) || (DelList != NULL) || (ModList != NULL) ){
        Update_FLAG = TRUE;
   	XtSetSensitive(UpdateInfoButton, TRUE);
   }else{
	Update_FLAG = FALSE;
        XtSetSensitive(UpdateInfoButton, FALSE);
   }

   startAllInfo(AllInfoButton, NULL, NULL);
}

/***************************************************************************
  update_conf_ls(w, call_data, call_data)  -- Angela Yeung
  Update conference list shown on the XTVsisint window and assign value for
  OldList for next time to use.
  This is a callback function for button "Update".
  It calls process_SESSION_INFO() to update the conference list.
  
  Input parameter:  w --  push button
                    packet-- packet recevied for socket.
****************************************************************************/
void update_conf_ls(w, packet, call_data)
Widget     w;                  /* activated pushbotton widget */
char  *packet;                 /* packet buffer from communication */
caddr_t call_data;
{


   Update_FLAG = FALSE;
   XtSetSensitive(UpdateInfoButton, FALSE);  
   XtSetSensitive(NewInfoButton, FALSE);  
   XtSetSensitive(DeletedInfoButton, FALSE);  
   XtSetSensitive(ModifiedInfoButton, FALSE);  
   DoneAllInfo(w, NULL, NULL);
   process_SESSION_INFO(&packet[1]);  

   /* Assign value for OldList for next time to use only if user press "update" */
   OldList= NewList;
}

void Show_Diff(w, DiffList, call_data)
Widget   w;
struct  info_db *DiffList;                /* list of conference to show */
caddr_t  call_data;
{
   int             line_count;
   struct  info_db *cur;

# if DEBUGINF
	fprintf( stderr,"Show_Diff:Staring\n");
# endif
   for (line_count=0 ; line_count<ALL_LIST_LINES ; line_count++)
      detail_list[line_count] = NULL;
   detail_list[0] = DETAIL_LIST_HEADER;

  /* Create a list on the bulletin board dialog */
   line_count=1;
   /* get the content of DiffList to build the scroll list  */
   if (w==NewInfoButton) {
   	cur = AddList;
        detail_list[0] = NEW_LIST_HEADER;
   }else if(w== DeletedInfoButton) {
        cur = DelList;
        detail_list[0] = DELETED_LIST_HEADER;
   }else if(w== ModifiedInfoButton) {
        cur = ModList;
        detail_list[0] = MODIFIED_LIST_HEADER;
   }

   while (cur!= NULL){
      detail_list[line_count] = cur->conf_id;
# if DEBUGINF
	fprintf(stderr,"Show_Diff: %s\n", cur->conf_id);
# endif
      line_count++;
      cur= cur->next;
   }

   /* if there is nothing in the list, display this statement in the list */
   if (line_count==1){
      detail_list[line_count] = "No conferences in this category";
      line_count++;
   }
   XawListChange(OneListWidget, detail_list, line_count, 0, TRUE);
   XtSetSensitive(OneViewport, TRUE);
#if   RingBell
   XBell(XtDisplay(topLevel), 5);
# endif
}





/********************************************************************
  int  build_db(packet)     -- Angela Yeung
  build the conference_info data base from the received packet.
  This is called by Handle_diff().
  input parameter:  pkt_buf          
  RETURN: linked list of conf.
********************************************************************/
struct info_db *build_db(pkt_buf)
char pkt_buf[PACKET_SIZE];                /* packet buffer from communication */
{
   int  n=0;                                 /* index of array */

   struct info_db *dblist= 0, *newconf;  
   char *pr;     

   pr= &pkt_buf[5];                   /* pass the conference info and time */

   while ( *pr != '\0'){
      /* build & add a node of the dblist */
      if ((newconf= (struct info_db *) malloc(sizeof(struct info_db))) == 0) 
         printf("error in malloc\n");

      newconf->next= dblist;
      dblist= newconf;


      /* find out the number of bytes for conference-id */
      n= 0;
      while(*pr++ != END_OF_SESSION_ID) n++;

      /* get data for the field conf_id */
      if ((dblist->conf_id = malloc(n+1)) == NULL)
         printf("error in malloc\n");
      bcopy(pr-n-1, dblist->conf_id, n);
      *(dblist->conf_id + n)= '\0';


      /* find out the number of bytes for info */
      n= 0;
      while(*pr++ != END_OF_SESSION) n++;

      /* get value for the field info */
      if ((dblist->info = malloc(n+1)) == NULL)
         printf("error in malloc\n");
      bcopy(pr-n-1, dblist->info, n);
      *(dblist->info + n)= '\0';

   }  
   return dblist;
}

/**********************************************************************
  int cmp_key(list1, list2, list3)    -- Angela Yeung
  This is called by Handle_diff().
  Compare list 1 and list 2 by using their conf_id as a key.
  Input parameter: list1, list2
  Output parameter: list3, contains the list of records which is difference.
    (records in list 1, but not in list 2);
**********************************************************************/
int cmp_key(list1, list2, list3)
struct info_db *list1, *list2;
struct info_db **list3;
{
   struct info_db *lspr, *cur,                  /* pointer to the list */
   *new, *outlist= 0;            /* for build the output list */

   int done;                                    /* flag */

   cur= list1;
   while (cur != NULL){
      done= 0;
      lspr= list2;
      while (!done){

         if (lspr == NULL){         /* element not found in list 2 */
            /* build & add a node of the new list */
            if ((new= (struct info_db *) malloc(sizeof(struct info_db))) == 0) 
               printf("error in malloc\n");
            *new= *cur;
            new->next= outlist;
            outlist= new;
            done= 1;
         }
         else if (strcmp(cur->conf_id,lspr->conf_id)== 0)
            done= 1;
         else
            lspr= lspr->next;                  /* traverse list 2 */
      }
      cur= cur->next;                        /* traverse list 1 */
   }

   *list3= outlist;                         /* assign the output list */
}

/************************************************************************
  int cmp_info(list1, list2, list3)   -- Angela Yeung
  This is called by Handle_diff().
  Compare the list1 (new) & list2(old) list by the field, info, in the 
    database.
  Input paraneter: list1, list2
  Output parameter: list3.  contains the list of records which have been
  updated.
************************************************************************/
int cmp_info(list1, list2, list3)
struct info_db *list1, *list2;
struct info_db **list3;                      /* the list of difference */
{
   struct info_db *lspr, *cur,                  /* pointer to the list */
   *new, *outlist= 0;            /* for build the output list */


   int done;                                    /* flag */


   cur=  list1;
   while (cur != NULL){
      done= 0;
      lspr= list2;
      while (!done){
         if (lspr == NULL){         /* element not found in list 2 */
            done= 1;
         }
         else if (strcmp(cur->conf_id,lspr->conf_id)== 0){
            if (strcmp(cur->info, lspr->info)!= 0){

               /* build & add a node of the output list */
               if ((new= (struct info_db *) malloc(sizeof(struct info_db))) == 0) 
                  printf("error in malloc\n");
               *new= *cur;
               new->next= outlist;
               outlist= new;
            }
            done= 1;
         }
         else
            lspr= lspr->next;                  /* traverse list 2 */
      }
      cur= cur->next;                        /* traverse list 1 */
   }

   *list3= outlist;                         /* assign the output list */
}


/*
 *
 * OptionChange()
 *
 * toggle states of options
 *
 */
void OptionChange(w, option)
Widget w;
char option;
{
  int  n;
  Arg  args[1];

  switch (option){
    case 'B':
      if(Option.b)
        Option.b = 0;
      else
        Option.b = 1;      /* invoke -B */
break;
case 'T':
if(Option.t)
Option.t = 0;
else
   Option.t = 1;      /* invoke -T */
break;
case 'V':
if(Option.v){
   Option.v = 0;
   XtSetSensitive(WMButton, TRUE);
}
else
{
   Option.v = 1;      /* invoke -V */
   XtSetSensitive(WMButton, FALSE);
}
break;
case 'F':
if(Option.f){
   Option.b = 0;      /* make not FAST by not invoking -B -T -V */
   Option.t = 0;
   Option.v = 0;
   XtSetSensitive(WMButton, TRUE);

}
else{
   Option.b = 1;      /* make FAST by invoking -B -T -V */
   Option.t = 1;
   Option.v = 1;
   XtSetSensitive(WMButton, FALSE);
}
n = 0;
XtSetArg(args[n], XtNstate, Option.b );
n++;
XtSetValues(Board, args, n);
n = 0;
XtSetArg(args[n], XtNstate, Option.t );
n++;
XtSetValues(Talk, args, n);
n = 0;
XtSetArg(args[n], XtNstate, Option.v );
n++;
XtSetValues(Virtual, args, n);
}
/* recalculate the state of FAST */

Option.f = Option.b && Option.t && Option.v;
n = 0;
XtSetArg(args[n], XtNstate, Option.f);
n++;
XtSetValues(Fast, args, n);
}


void WMCallBack(w, client_data, call_data)
Widget w;
caddr_t client_data, call_data;
{
int             winx, winy, rwinx, rwiny;
unsigned int    maskret;
Window          rootret, childret;
Arg             arg[2];

XQueryPointer(XtDisplay(topLevel), XtWindow(topLevel), &rootret,
&childret, &rwinx, &rwiny, &winx, &winy, &maskret);

/* move the popup shell to pointer position */
XtSetArg(arg[0], XtNx, rwinx);
XtSetArg(arg[1], XtNy, rwiny);
XtSetValues(WMTranShell, arg, 2);

XtSetSensitive(WMButton, FALSE);
XtPopup(WMTranShell, XtGrabNonexclusive);
}

void WMCancel(w, client_data, call_data)
Widget w;
caddr_t client_data, call_data;
{
XtPopdown(WMTranShell);
XtSetSensitive(WMButton, TRUE);
}

void WMAccept(w, client_data, call_data)
Widget w;
caddr_t client_data, call_data;
{
/* copy any arguments entered in the dialog box */
if (*XawDialogGetValueString(WMDialog))
sprintf(WM_dialog_str, "-M %s", 
XawDialogGetValueString(WMDialog));
else
strcpy(WM_dialog_str, "-M twm");


XtPopdown(WMTranShell);
XtSetSensitive(WMButton, TRUE);
}
