/*
 * Copyright (c) 1991, 1993 Old Dominion University.
 * Copyright (c) 1991, 1993 University of North Carolina at Chapel Hill.
 * All rights reserved.
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  Old Dominion University and the University of North
 * Carolina at Chapel Hill make no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 */

#include <signal.h>
#include <stdio.h>
#include <errno.h>
#include <sys/time.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netdb.h>
#include <X11/X.h>
#include <time.h>


#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <unistd.h>


#include <X11/Xproto.h>

#include <X11/Xlib.h>

#include <X11/Xresource.h>

#include <X11/Xutil.h>

#include "debug.h"
#include "vs.h"

char vswm[10];
char DISPLAY_VAR[256]="";
main(argc, argv)
   int argc;
   char *argv[];
{
   XrmDatabase        resource_data_base;
   vs_type            vs;
   int                remaining_sockets_to_set;
   int                new_display_number;
   char               *GetNewDisplayVariable();
   filter_type        filter[FILTER_NUMBER];
   int                debug = 0;
   fd_set             socket_set;
   int                filter_index;
   BYTE               buffer[MAXIMUM_SIZE_OF_A_PROTOCOL];
   Window             dummy_window;
   XEvent             xevent;
   
/* begin VS-XTV modification */
   FILE *fp,*fdopen();
   int pipefd;
   pipefd = atoi(argv[1]);
   strcpy(vswm, argv[2]);
/* end VS-XTV modification */
   if (debug) PrintArgument(argc,argv);
   DEBUG("Maximum file descriptor number is %d\n",getdtablesize());

   ParseParameters(&argc,argv, &resource_data_base, &vs);


   if ( ( vs.real_display = XOpenDisplay(vs.real_display_variable) ) == NULL ) {
      printf("ERROR: DISPLAY %s can't be opened!\n",vs.real_display_variable);
      exit(0);
   }
   vs.real_X_socket = ConnectionNumber(vs.real_display);
   DEBUG("real_X_socket = %d\n",vs.real_X_socket);

   CreateVirtualRoot(&vs);

/* begin VS-XTV modification */

/* write down the virtual_root window ID to pipe so that XTV can find 
   it later on if it needs it.
*/
     if((fp = fdopen(pipefd,"w")) == NULL) {
         perror("opening pipe for writing");
         exit(1);
        }
     fprintf(fp,"0x%08x\n",vs.virtual_root.window);
     fflush(fp);

/*
     fclose(fp);
*/
/* end VS-XTV modification */



   vs.public_socket = OpenPublicDisplaySocketAtLocalHost(&new_display_number);
   DEBUG("public_socket = %d\n",vs.public_socket);
   fcntl(vs.public_socket,F_SETFL,FNDELAY);
   vs.virtual_display_variable = GetNewDisplayVariable(new_display_number);
   DEBUG(" real display varaibe = %s\n",vs.real_display_variable);

   ConsoleProcessInitialization(&vs);
   DEBUG(" real display varaibe = %s\n",vs.real_display_variable);

   InactivateAllFilters(filter,&(vs.number_of_active_filter));
   DEBUG(" =====================================\n");

   for (;;) {
      DEBUG("------------------------------------\n");
      SetSockets(&vs,filter,&socket_set);
      select(256,&socket_set,(fd_set *) 0,(fd_set *) 0, (struct timeval *) 0);

      if ( FD_ISSET(vs.real_X_socket,&socket_set) ) {
         DEBUG("  REAL EVENT\n");
         XNextEvent(vs.real_display,&xevent);
         DEBUG("  4444\n");
         DEBUG("xevent.xconfigure.window =0x%08x\n",xevent.xconfigure.window);
         if ( (xevent.type == ConfigureNotify) &&
              (xevent.xconfigure.window == vs.virtual_root.window) ) {
            XTranslateCoordinates(vs.real_display,
                                  vs.virtual_root.window,vs.real_root.window,
                                  0,0,
                                  &(vs.virtual_root.x),&(vs.virtual_root.y),
                                  &dummy_window);
            DEBUG("  virtual root at (%d,%d)\n",
                  vs.virtual_root.x,vs.virtual_root.y);
         }
      }
      if ( FD_ISSET(vs.public_socket,&socket_set) ) {
         DEBUG("  A client is coming!\n");
         filter_index = NewFilterInitialization(filter,&(vs));
         DEBUG(" Filter[%d] is used.\n",filter_index);
         if (filter_index<0) {
            printf("VS Error: Can't allocate any new filter!\n");
         }
      }
      for ( filter_index=0 ; filter_index<FILTER_NUMBER ; filter_index++) {
         if ( filter[filter_index].is_active) {
            if ( FD_ISSET(filter[filter_index].socket_to_X,&socket_set) ) { 
               if ( filter[filter_index].next_action == HANDLE_SETUP_REPLY ) {
                  DEBUG("HANDLE_SETUP_REPLY\n");
                  ReceiveSetupReply(filter[filter_index].socket_to_X,buffer);
                  VSHandleSetupReply( buffer, &(vs.virtual_root) );
                  SendSetupReply(filter[filter_index].socket_to_client,buffer);
                  filter[filter_index].next_action = HANDLE_USUAL_PROTOCOL;
               }
               else if ( ReceiveReplyOrEvent(filter[filter_index].socket_to_X,
                         buffer) > 0 ) {
                  DEBUG("ReceiveReplyOrEvent\n");
                  VSHandleReplyOrEvent(buffer,vs.real_root.window,
                                       filter[filter_index].root,
                                       filter[filter_index].request_status);
                  SendReplyOrEvent(filter[filter_index].socket_to_client,
                                   buffer);
               }
               else {
                  printf("Logic Error:socket is set, but no data!\n");
                  printf("            socket to X at # %s\n",filter_index);
               }
            }
            if ( FD_ISSET(filter[filter_index].socket_to_client,&socket_set) ){
               if ( filter[filter_index].next_action == HANDLE_SETUP_REQUEST) {
                     DEBUG("HANDLE_SETUP_REQUEST\n");
                     ReceiveSetupRequest(filter[filter_index].socket_to_client,
                                         buffer);
                     VSHandleSetupRequest(buffer);
                     SendSetupRequest(filter[filter_index].socket_to_X,buffer);
                     filter[filter_index].next_action = HANDLE_SETUP_REPLY;
               }
               else if ( ReceiveRequestProtocol(
                         filter[filter_index].socket_to_client,buffer) > 0) {
                  DEBUG("ReceiveRequestProtocol\n");
                  VSHandleRequestProtocol( buffer, &(filter[filter_index]) );
                  SendRequestProtocol(filter[filter_index].socket_to_X,
                                      buffer);
               }
               else {
                  DEBUG("A client is dead!\n");
                  filter[filter_index].is_active= False;
                  close(filter[filter_index].socket_to_X);
                  close(filter[filter_index].socket_to_client);
                  vs.number_of_active_filter--;
               }
            } 
         } 
      } 
/*
 Gary Carlosn fix. It Only works for SUN. did't work for DEC at UNC
*/
 write(pipefd,"",0);

   } /* forever loop */
}

SetSockets(vs,filter,socket_set)
   vs_type     *vs;
   filter_type filter[];
   fd_set      *socket_set;
{
   int remaining_sockets_to_set,
       filter_index;
   int debug = 0;

   FD_ZERO(socket_set);
   FD_SET(vs->public_socket,socket_set);
   FD_SET(vs->real_X_socket,socket_set);
   remaining_sockets_to_set = vs->number_of_active_filter;
   for (filter_index = 0; remaining_sockets_to_set > 0; filter_index++) {
      if (filter[filter_index].is_active) {
         FD_SET(filter[filter_index].socket_to_X,socket_set);
         FD_SET(filter[filter_index].socket_to_client,socket_set);
         remaining_sockets_to_set--;
      }
   }
}



int NewFilterInitialization(filter,vs)
   filter_type filter[];
   vs_type     *vs;
{
   int filter_id;

   for (filter_id=0; filter_id<FILTER_NUMBER; filter_id++) {
      if (filter[filter_id].is_active == False) {
         filter[filter_id].is_active = True;
         filter[filter_id].socket_to_client = 
                         AcceptSocketConnection(vs->public_socket);
         filter[filter_id].socket_to_X = 
                         ConnectSocketToDisplay(vs->real_display_variable);
         filter[filter_id].root = &(vs->virtual_root);
         filter[filter_id].request_status = 0;
         filter[filter_id].is_window_manager = False;
         filter[filter_id].rx = 0;
         filter[filter_id].ry = 0;
         filter[filter_id].bottom_window = NULL;
         filter[filter_id].next_action = HANDLE_SETUP_REQUEST;
         vs->number_of_active_filter++;
         return filter_id;
      }
   }
   return (-1);
}


InactivateAllFilters(filter,used_filter)
   filter_type  filter[];
   int          *used_filter;
{
   int index;

   *used_filter = 0;
   for (index=0; index<FILTER_NUMBER; index++) {
      filter[index].is_active = False;
   }
}
    

ConsoleProcessInitialization(vs)
   vs_type *vs;
{
   int debug = 0;

   /* -- to be added -- */
   /* Addition to reserve signal stutus is necessary */
   /*to have normal forks later!                     */

   void TerminateVirtualScreen();

   DEBUG("   NEW DISPLAY = %s\n",vs->virtual_display_variable);
   signal(SIGCLD,TerminateVirtualScreen);
   DEBUG("   OLD DISPLAY = %s\n",vs->real_display_variable);
/*
   setenv("DISPLAY",vs->virtual_display_variable,1);
*/
   strcat(DISPLAY_VAR,"DISPLAY");
   strcat(DISPLAY_VAR,"=");
   strcat(DISPLAY_VAR,vs->virtual_display_variable);
   putenv(DISPLAY_VAR);
   DEBUG("   OLD DISPLAY = %s\n",vs->real_display_variable);
   if ( fork() == 0 ) {
/*
      DEBUG("   Putting initial xterm!\n");
      if ( (execlp("xterm","xterm","-n","XTV-VS Console", "-geometry", "40x10", (char *)0) ) == (-1) ) {
         printf(" Error :  Console process can not be created!\n");
      } 
*/
        if ( (execlp(vswm,vswm,(char *)0) ) == (-1) ) {
         printf(" Error : %s can't created!\n", vswm);
      }
      exit(0);
   }
   DEBUG("   OLD DISPLAY = %s\n",vs->real_display_variable);
/*
   setenv("DISPLAY",vs->real_display_variable,1);
*/
   strcat(DISPLAY_VAR,"DISPLAY");
   strcat(DISPLAY_VAR,"=");
   strcat(DISPLAY_VAR,vs->real_display_variable);
   putenv(DISPLAY_VAR);
   DEBUG("   OLD DISPLAY = %s\n",vs->real_display_variable);
}

void TerminateVirtualScreen()
{
   int debug = 0;

   DEBUG("Virtual screen is normally terminated.\n");
   exit(0);
}




   
CreateVirtualRoot(vs)
   vs_type *vs;
{
   long    event_mask_for_virtual_root = StructureNotifyMask;

   vs->real_screen = DefaultScreen(vs->real_display);

   vs->real_root.window = RootWindow(vs->real_display,vs->real_screen);
   vs->virtual_root.window = XCreateSimpleWindow(vs->real_display, 
                            vs->real_root.window, 0,0,
                            vs->virtual_root.width, vs->virtual_root.height,2,
                            BlackPixel(vs->real_display,vs->real_screen),
                            WhitePixel(vs->real_display,vs->real_screen) );
   XStoreName(vs->real_display,vs->virtual_root.window,vs->title);
   XSelectInput(vs->real_display,vs->virtual_root.window,
                event_mask_for_virtual_root);
   XMapWindow(vs->real_display,vs->virtual_root.window);
   XFlush(vs->real_display);
}
                                      

