/*
 * Copyright (c) 1991, 1993 Old Dominion University.
 * Copyright (c) 1991, 1993 University of North Carolina at Chapel Hill.
 * All rights reserved.
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  Old Dominion University and the University of North
 * Carolina at Chapel Hill make no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 */


#include <stdio.h>
/*
#include <X11/X.h>
*/
#include <X11/Xproto.h> 

#include <X11/Xlib.h>

#include <X11/Xresource.h>

#include <X11/Xutil.h>

#include "debug.h"
#include "vs.h"

#define default_root_width  1000
#define default_root_height 800


ParseParameters(argc,argv,      resource_data_base,  vs)
   int          *argc;
   char         *argv[];
   XrmDatabase  *resource_data_base;
   vs_type      *vs;
{
   static XrmOptionDescRec opTable[] = {
   {"\-help",       ".help",         XrmoptionNoArg,   (caddr_t) "on"},
   {"\-title",      ".title",        XrmoptionSepArg,  (caddr_t) NULL},
   {"\-display",    ".display",      XrmoptionSepArg,  (caddr_t) NULL},
   {"=",            ".geometry",     XrmoptionIsArg,   (caddr_t) NULL},

   {"\-asyn",      ".synchrounous",  XrmoptionNoArg,   (caddr_t) "off"},

   {"\-bordercolor",".borderColor",  XrmoptionSepArg,  (caddr_t) NULL},
   {"\-borderwidth",".borderWidth",  XrmoptionSepArg,  (caddr_t) NULL},
   };

   char         *dummy_str_type[20];
   XrmValue     resource_value;
   int          geometry_mask;
   char         guest_socket_port[4];
   unsigned int width,height;
   char         geometry_string[30];
   int          dummy_x,
                dummy_y;
   int          debug = 0;

   if (debug) PrintArgument(*argc,argv);
   XrmInitialize();
   XrmParseCommand(resource_data_base, opTable, 4, "vs", argc, argv);
   if (debug)  XrmPutFileDatabase(*resource_data_base,"tmp.resource");
   if (debug) PrintArgument(*argc,argv);

   if (XrmGetResource(*resource_data_base, "vs.help", "Vs.Help",
                      dummy_str_type, &resource_value)== True) {
      printf("Usage: vs -help       --  print this message\n");
      printf("          -display    --  DISPLAY variable for host\n");
      printf("          -title      --  title fo this virtual screen\n");
      printf("          =<width>x<height>\n"); 
      printf("                      --  set geometry of virtual screen\n");
      exit(0);
   }

   
   if (XrmGetResource(*resource_data_base, "vs.title", "Vs.Title",
                       dummy_str_type, &resource_value)== True) {
        (void) strncpy(vs->title, resource_value.addr,
                       (int) resource_value.size);
   }
   else {
         (void) strcpy(vs->title,"Virtual Screen");
   }
   DEBUG("title = %s\n",vs->title);

   if (XrmGetResource(*resource_data_base, "vs.display", "Vs.Display",
                       dummy_str_type, &resource_value)== True) {
        (void) strncpy(vs->real_display_variable, resource_value.addr,
                       (int) resource_value.size);
   }
   else {
       strcpy(vs->real_display_variable,(char *) getenv("DISPLAY")); 
       /*
       vs->real_display_variable = (char *) getenv("DISPLAY");
       */
       if ( vs->real_display_variable  == NULL ) {
          printf("ERROR: DISPLAY variable is NULL\n" );
          exit(0);
       }
   }
   DEBUG("DISPLAY = %s\n",vs->real_display_variable);

   vs->virtual_root.width  = 0;
   vs->virtual_root.height = 0;
   if (XrmGetResource(*resource_data_base, "vs.geometry", "Vs.Geometry",
                       dummy_str_type, &resource_value)== True) {
      DEBUG("geometry assigned!\n"); 
      (void) strncpy(geometry_string, resource_value.addr,
                                (int) resource_value.size);
      DEBUG("   geometry string is %s\n",geometry_string);
      geometry_mask = XParseGeometry(geometry_string,&dummy_x,&dummy_y,
                                     &width,&height);
      if (geometry_mask && WidthValue) {
           vs->virtual_root.width  = (int) width ;
      }
      if (geometry_mask && HeightValue) {
           vs->virtual_root.height = (int) height;
      }
   }
   if (vs->virtual_root.width == 0) { 
      vs->virtual_root.width = default_root_width;  
   }
   if (vs->virtual_root.height== 0) { 
      vs->virtual_root.height= default_root_height;  
   }
   DEBUG("width = %d, height = %d\n",vs->virtual_root.width,
                                     vs->virtual_root.height);
}


PrintArgument(argc,argv)
   int argc;
   char *argv[];
{
       int           index;

    for (index = 0; index<argc ; index++) {
       printf(" %s |",argv[index]);
    }
    printf("\n");
}

   


   
