/*
 * Copyright (c) 1991, 1993 Old Dominion University.
 * Copyright (c) 1991, 1993 University of North Carolina at Chapel Hill.
 * All rights reserved.
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  Old Dominion University and the University of North
 * Carolina at Chapel Hill make no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 */

/*
	chat.c
	chat routines

*/

/* Code enhancements designed by Dick At Lee (4/93) are marked: -RYA */

#include "chat.h"
#include "subscriber.h"
#include "tool.h"
#include "protocol.h"
#include "network.h"
#include "memory.h"
#include "token.h"
#include "subscriber2.h"
#include "tool2.h"
#include "popups.h"
#include "particInfo.h"


#define  DEBUGInitializeChatPanel   0&DEBUGON
#define  DEBUGInitializeChat        0&DEBUGON
#define  DEBUGChatCheck             0&DEBUGON
#define  DEBUGAddChatZone           0&DEBUGON
#define  DEBUGDeleteChatZone        0&DEBUGON
#define  DEBUGChatEventHandler      0&DEBUGON
#define  DEBUGChatSwitch            0&DEBUGON
#define  DEBUGPackegeHandle         0&DEBUGON
#define  DEBUGChatHandler           0&DEBUGON
#define  DEBUGChatDisplay           0&DEBUGON
#define  DEBUGConvertChat           0&DEBUGON


extern BYTE  MulticastFlag[];  /* If TRUE, include Subscriber in Multicast*/

extern int chat_port;
extern Socket ChatSocket;
extern int CHAT;
Widget    chat_top,chat_pane,chatpub,chatpublab;   /*for chat windows -RYA*/
Widget    anon_top,anon_pane,anonpub,anonpublab;   /*for anon windows -RYA*/
Bool   mode; 
static int chat_count,pub_count;
char   *blank=" ";
char   *ret="\n";
char hidding[512];
int  r_count = 0;
int s_count = 0;

extern int SessionNumber;
extern Subscriber LocalInfo;
extern BYTE LocalSubID;
Dimension  chat_height=56;
XImage *Inimage;

/****************************************************************************
** FUNCTION NAME: beep
** 
** PURPOSE: Sound bell when typing in window other than your own.
** 
** CALLED BY: chat_wind widget created in AddChatZone() in chat.c
** 
** INPUTS: Standard callback parameters (most not used).
** 
** CALLS: Nothing.
** 
** 
*/
void beep(w,event,params,num_params)
Widget w;
XEvent *event;
String *params;
int num_params;
{

  XBell(XtDisplay(w),15);

}

/***************************************************************************/
/* FUNCTION NAME: InitializeChatPanel
** 
** PURPOSE: Creates "chat" initial chat windows for subscribers which include
**   basic overall window containing name at top, a top window labeled 
**   Anonymous which displays anonymous messages from any user, a thin
**   label/toggle widget with subscriber's name and [ON] or [OFF] message,
**   and finally, a text entry window below the label widget.
** 
** CALLED BY: InitializeSubscriber() in subscriber.c
** 
** INPUTS: A parent widget name.
** 
** CALLS: GetMode(), GetAppCon()
** 
** 
*/
InitializeChatPanel(toplevel)
Widget toplevel;
{

char   *dp="Anonymous";
static char TITLE[100];
char SessNum[5];


#        if DEBUGInitializeChatPanel
		fprintf( stderr, "InitializeChatPanel():\n" );
#        endif


/*
   strcat(TITLE, "Chat/Session#");     
   sprintf(SessNum,"%d/",SessionNumber);
   strcat(TITLE, SessNum);          
*/
   if (GetMode())
   	strcat(TITLE, "Chat/Chairman");
    else 
   	strcat(TITLE, "Chat/Participant");

   chat_top = XtVaAppCreateShell(GetAppCon(),"xtv",topLevelShellWidgetClass,
                      XtDisplay(toplevel),
                      XtNtitle,TITLE,
		      NULL);
   chat_pane =  XtVaCreateManagedWidget("chatpane",
                                       panedWidgetClass,chat_top ,
                      XtNallowResize,        True,
                      XtNresizeToPreferred,  True,
                      NULL);

   chatpublab  = XtVaCreateManagedWidget( "chatpublabel",
                   labelWidgetClass,chat_pane,
                   XtNshowGrip,  False,
                   XtNmax,       15,
                   XtNheight,    15,
                   XtNwidth,     400,
                   XtNlabel,     dp,
                   NULL);     

   chatpub  = XtVaCreateManagedWidget( "Chatpub",
                     asciiTextWidgetClass,chat_pane, 
                     XtNmin,40,
                     XtNwidth,     400,
                     XtNscrollVertical,XawtextScrollWhenNeeded ,
                     XtNtype,XawAsciiString,
                     XtNwrap,XawtextWrapWord,
                     XtNeditType, XawtextAppend,
                     NULL);     

 } /* end InitializeChatPanel() */



/****************************************************************************
** FUNCTION NAME: InitializeAnonPanel   DESIGNED BY RYA  (Dick At Lee)
** 
** PURPOSE: To create Anonymous Typing Window for text entry under the 
**    anonymous mode.  Window has Title in frame, user info in label, and
**    a text window which allows normal text entry.  Very similar to 
**    InitializeChatPanel().
** 
** CALLED BY: ChatSwitch() in chat.c
** 
** INPUTS: A parent widget.
** 
** CALLS: GetAppCon(), callback: ChatEventHandler()
** 
** 
*/
static char   TITLE[100]="ANONYMOUS TYPING WINDOW";
InitializeAnonPanel(anon)
Widget anon;
{
  char SessNum[5];
  char   *dp=chatinf[LocalSubID].name;


   anon_top = XtVaAppCreateShell(GetAppCon(),"xtv",    /*toplevel widget*/
		      topLevelShellWidgetClass,
                      XtDisplay(anon),
                      XtNtitle,TITLE,
		      NULL);

   anon_pane =  XtVaCreateManagedWidget("anonpane",    /*container widget*/
                      panedWidgetClass,anon_top,
                      XtNallowResize,        True,
                      XtNresizeToPreferred,  True,
                      NULL);

   anonpublab  = XtVaCreateManagedWidget("anonpublabel",/*user info label*/
                   labelWidgetClass,anon_pane,
                   XtNshowGrip,  False,
                   XtNmax,       15,
                   XtNheight,    15,
                   XtNwidth,     400,
                   XtNlabel,     dp,
                   NULL);     

   anonpub  = XtVaCreateManagedWidget( "Anonpub",      /*text entry window*/
                     asciiTextWidgetClass,anon_pane, 
                     XtNmin,     40,
                     XtNwidth,   400,
                     XtNscrollVertical,XawtextScrollWhenNeeded ,
                     XtNtype,XawAsciiString,
                     XtNwrap,XawtextWrapWord,
                     XtNeditType, XawtextAppend,
                     NULL);    

  XtAddEventHandler(anonpub,KeyPressMask,   /*handles text entry/xmission*/
		    FALSE,ChatEventHandler,LocalSubID);

  SetPlace(anon_top,chat_top);
            /*show widgets on display*/
  XtRealizeWidget(anon_top);    

 } /* end of InitializeAnonPanel() */







/****************************************************************************
** FUNCTION NAME: InitializeChat
** 
** PURPOSE: Initializes chat flags to zero, establishes a comms port for the
**    chat window, and realizes the chat window.
** 
** CALLED BY: main() in main.c
** 
** INPUTS: A parent widget and hostname.
** 
** CALLS: 
** 
** 
*/
InitializeChat(toplevel,hostname)
Widget toplevel;
char   *hostname;
{

   Socket connect; 
   int n,m;
   BYTE *TempByteP;
   unsigned long Temp;
 
#   if DEBUGInitializeChat
   fprintf( stderr, "InitializeChat(): \n");
#   endif

   mode = GetMode();
   chat_count = 0;
   pub_count = 0;
  
   if (mode) {

     XtAppAddInput( GetAppCon(), ChatSocket, XtInputReadMask, ChatCheck, NULL );

     for (n=0;n<=256;n++)
         for(m=0;m<=256;m++)
             chatflag[n][m]=0;     

#   if DEBUGInitializeChat
   fprintf( stderr, "InitializeChat(): Chat port open at display port %d\n", chat_port );
#   endif

   /*  fcntl( ChatSocket, F_SETFL, FNDELAY );
*/
   } else {
     
     	if ((connect = ConnectTo( hostname, htons(chat_port ))) < 0 ) {
#  if DEBUGInitializeChat
		fprintf( stderr, "InitializeChat(): Fail to Connect to %s:%d\n", hostname,chat_port);
#  endif
		return -1;

	}

#  if DEBUGInitializeChat
		fprintf( stderr, "InitializeChat(): Connect to %s:%d\n", hostname,chat_port);
#  endif

        Temp =htonl(R_XTV_CONNECT_MAGIC) ;
        if (SendData(connect, &Temp , sizeof(Temp)) == ERROR) EndSession();

#      if DEBUGInitializeChat
	fprintf( stderr, "InitializeChat(): Sent R_ XTV_CONNECT_MAGIC \n" );
	fprintf( stderr, "InitializeChat(): Waiting for reply\n" );
#      endif

	TempByteP = (BYTE *)ReadData( connect, sizeof(BYTE), TRUE );
     
	if (*TempByteP) {
#        if DEBUGInitializeChat
		fprintf( stderr, "InitializeChat(): Accepted.\n" );
#        endif

		if (SendData( connect,&LocalInfo,sizeof(LocalInfo))==ERROR) EndSession();

        	TempByteP = (BYTE *)ReadData( connect, sizeof(BYTE), TRUE );
                
                n = *TempByteP; 
                          
		XtAppAddInput( GetAppCon(), connect, XtInputReadMask,
		    ChatHandler,n );
		FreeMem( TempByteP );	
       ChatSocket = connect;
#        if DEBUGInitializeChat
       fprintf(stderr,"chatsocket=%d",ChatSocket);
#        endif
       } else {

#        if DEBUGInitializeChat
		fprintf( stderr, "InitializeChat(): Connection not been Accepted.\n" );
#        endif
         EndSession(); 
	      }
      }

#        if DEBUGInitializeChat
		fprintf( stderr, "InitializeChat(): Realize the chat window.\n" );
#        endif


if (CHAT)
    XtRealizeWidget(chat_top);


 }






/****************************************************************************
** FUNCTION NAME: ChatCheck
** 
** PURPOSE: 
** 
** CALLED BY: 
** 
** INPUTS: 
** 
** CALLS: 
** 
** 
*/
XtInputCallbackProc ChatCheck( client_data, connect, ID )
XtPointer client_data;
Socket *connect;
XtInputId *ID;
{
   Subscriber    *Who;
   Socket        NewSocket;
   struct sockaddr Newaddr; 
   int      n = sizeof(struct sockaddr);
   unsigned long *magicNumber;
   Subscriber    *newname;
   int           id;

#    if DEBUGChatCheck 
	fprintf( stderr, "ChatCheck()\n" );
#    endif

	XtRemoveInput( *ID );
	XtAppAddInput( GetAppCon(), *connect, XtInputReadMask, ChatCheck, NULL );

	if ((NewSocket = accept( ChatSocket, &Newaddr, &n )) < 0 ) {
#    if DEBUGChatCheck 
		fprintf( stderr, "ChatCheck(): No Connection\n" );
#    endif

	        XtRemoveInput( *ID );
		return FALSE;
	}

#  if DEBUGChatCheck
	fprintf( stderr, "ChatCheck(): Got connection\n" );
#  endif


	magicNumber=(unsigned long *)ReadData( NewSocket, sizeof(unsigned long), TRUE );
        *magicNumber = ntohl(*magicNumber); 
	if ( *magicNumber == R_XTV_CONNECT_MAGIC ) {
		SendByte( NewSocket, (BYTE)01);

#        if DEBUGChatCheck
		fprintf( stderr, "ChatCheck(): Waiting for data...\n" );
#        endif
		newname =(Subscriber *)ReadData( NewSocket,sizeof(Subscriber), TRUE );

#        if DEBUGChatCheck
		fprintf( stderr, "ChatCheck(): Adding new user : %s\n",newname->LoginName );
#        endif

         for ( Who=FirstSubscriber(); Who != NULL; Who=NextSubscriber() ) {

             n = strcmp(Who->LoginName,newname->LoginName);
             if (!n) {
                   n = strcmp(Who->HostName,newname->HostName);
                   if (!n && (Who->ChatConnect == newname->ChatConnect)) {           

                          Who->ChatConnect = NewSocket;
			  SendByte( NewSocket, (BYTE)Who->SubscriberID);
                          id = Who->SubscriberID;
                          break;
			}
		 }
	   }
             XtAppAddInput( GetAppCon(), NewSocket, XtInputReadMask, ChatHandler,id );

	      } else {

#        if DEBUGChatCheck
		fprintf( stderr, "ChatCheck(): Wrong Connection format .");
#        endif
   
                  close (NewSocket);

             }

 }



/****************************************************************************
** FUNCTION NAME: ChatHandler
** 
** PURPOSE: 
** 
** CALLED BY: 
** 
** INPUTS: 
** 
** CALLS: 
** 
** 
*/
XtInputCallbackProc ChatHandler( Subid, connect, ID )
int    Subid;
Socket *connect;
XtInputId *ID;
{
   Packege *Inc; 
   Socket  channel;
   int n,j;
   XImage *image;
   char *image_data;

#  if DEBUGChatHandler
	fprintf( stderr, "ChatHandler(): \n" );
#  endif

	XtRemoveInput( *ID );
        if (chatinf[Subid].chat_wind) {
           channel = *connect;
	Inc = (Packege *)ReadData (*connect,sizeof(Packege), TRUE );
#  if DEBUGChatHandler
        fprintf( stderr, "ChatHandler(): came from ReadData \n" );
#  endif
        if (Inc){
	ntohConvertChat(Inc);
#  if DEBUGChatHandler
	fprintf( stderr, "ChatHandler(): Received packege from id : %d\n",Inc->id);
#  endif
              if (Inc->id > 256 || Inc->id < 0) {
          	fprintf( stderr, "ChatHandler(): Bad Packege ID Received : id= %d\n",Inc->id );
                goto bad;
	      }
 
              r_count = 0;
              if (Inc->flag != BOARD_IMAGE_ID) {
                  XtAppAddInput( GetAppCon(), channel, XtInputReadMask,ChatHandler, Subid );
#  if DEBUGChatHandler
fprintf(stderr, " ChatHandler(): %d != BOARD_IMAGE_ID\n", Inc->flag);
#  endif
                  PackegeHandle(Inc);
		} else {

                  j = Inc->p_keysym * Inc->y1 * Inc->depth;                     
#  if DEBUGChatHandler
fprintf(stderr, " ChatHandler()image size: j %d\n", j);
#  endif
                  image_data = (char *) ReadData(*connect,j,TRUE);
                  XtAppAddInput( GetAppCon(), channel, XtInputReadMask,ChatHandler, Subid );
                  Inimage->width = Inc->x1;
                  Inimage->height = Inc->y1;
                  Inimage->bytes_per_line = Inc->p_keysym;

#  if DEBUGChatHandler
	fprintf( stderr, "ChatHandler(): calling ImageData\n");
#  endif
                  ImageData(Inc->x,Inc->y,j,image_data,Inc);
           } 
       } else {

#  if DEBUGChatHandler
	fprintf( stderr, "ChatHandler(): Bad Packege Received..\n" );
#  endif

bad :             r_count++;
             if (r_count < 5) 
                XtAppAddInput( GetAppCon(), channel, XtInputReadMask,ChatHandler, Subid );
           }
     }
}







/****************************************************************************
** FUNCTION NAME: ChatDisplay
** 
** PURPOSE: 
** 
** CALLED BY: 
** 
** INPUTS: 
** 
** CALLS: 
** 
** 
*/
void ChatDisplay(packege)
Packege *packege;
{

   KeySym keysym;
   char   key[3];
   XawTextBlock  block;
   int j;

#  if DEBUGChatDisplay
	fprintf( stderr, "ChatDisplay() ..\n" );
#  endif

  if (CHAT) {

  if (packege->p_keysym >= XK_space && packege->p_keysym <= XK_asciitilde) {
    block.firstPos = 0;
    block.length = 1;
    block.ptr = &packege->string[0];

    j = chat_count;
    chat_count++;
    XawTextReplace(chatinf[packege->id].chat_wind,j,j,&block);
    XawTextSetInsertionPoint(chatinf[packege->id].chat_wind,chat_count);


  }  else
       if (packege->p_keysym == 65288 || packege->p_keysym == 65535) {
              block.firstPos = 0;
              block.length = 0;
              block.ptr = blank;

              j = chat_count;
              chat_count--;
              XawTextReplace(chatinf[packege->id].chat_wind,j-1,j+1,&block);
              XawTextSetInsertionPoint(chatinf[packege->id].chat_wind,chat_count);

            } else
                if (packege->p_keysym == 65293 ) {
                  block.firstPos = 0;
                  block.length = 1;
                  block.ptr = ret;

                  j = chat_count;
                  chat_count++;
                  XawTextReplace(chatinf[packege->id].chat_wind,j,j,&block);
                  XawTextSetInsertionPoint(chatinf[packege->id].chat_wind,chat_count);
                  }
   }
}







/****************************************************************************
** FUNCTION NAME: AddChatZone
** 
** PURPOSE: Adds new user info label, text entry window, and ChatEventHandler
**    callback routine for each new subscriber.  Adds the label and window
**    to exiting chat windows for itself and other subscribers.
** 
** CALLED BY: AddSubscriber() in subscriber.c
** 
** INPUTS: Pointer to structure type Subscriber
** 
** CALLS: 
** 
** 
*/
void AddChatZone(Newcomer)
Subscriber *Newcomer;
{

   int   id,bg,fg;
   Dimension  cheight;
   char  dp[100];
   Arg   arg;
   XWindowChanges values;


#  if DEBUGAddChatZone
	fprintf( stderr, "Chat():AddChatZone(): \n" );
#  endif

   
   id =(int)Newcomer->SubscriberID;
   sprintf(chatinf[id].name,"%s@%s",Newcomer->LoginName,Newcomer->HostName);

   sprintf(dp," [ ON ]   %s",chatinf[id].name);   

   chatinf[id].chat_label =  XtVaCreateManagedWidget( "chatlabel",
                   toggleWidgetClass,chat_pane,
                   XtNshowGrip,  False,
                   XtNmax,       15,
                   XtNheight,    15,
                   XtNlabel,     dp,
                   NULL);  
   XtAddCallback( chatinf[id].chat_label, XtNcallback, ChatSwitch, id);

   if (id == LocalSubID) {
        chatinf[id].chat_wind = XtVaCreateManagedWidget( "chatown",
                     asciiTextWidgetClass,chat_pane,
                     XtNmin,30,
                     XtNscrollVertical,XawtextScrollWhenNeeded,
		     XtNtype,XawAsciiString,
                     XtNwrap,XawtextWrapWord,
                     XtNeditType, XawtextAppend,
                     NULL );

        XtAddEventHandler(chatinf[id].chat_wind,
                KeyPressMask,FALSE,ChatEventHandler,id);

     }  
    else {
        chatinf[id].chat_wind = XtVaCreateManagedWidget( "chat",
                     asciiTextWidgetClass,chat_pane,
                     XtNmin,30,
                     XtNscrollVertical,XawtextScrollWhenNeeded,
		     XtNtype,XawAsciiString,
                     XtNwrap,XawtextWrapWord,
                     XtNeditType, XawtextAppend,
                     NULL );

        XtAddEventHandler(chatinf[id].chat_wind,
                KeyPressMask,FALSE,beep,NULL);

      }
  chat_height = chat_height+45;
  if (XtIsRealized(chat_top)) {

     XtSetArg(arg,XtNheight,&cheight);
     XtGetValues(chat_top,&arg,1);
     if (chat_height > cheight) cheight = chat_height;
        else  chat_height = cheight;

     values.height = cheight+45;
     XConfigureWindow(XtDisplay(chat_top),XtWindow(chat_top),CWHeight,&values);

   }
}


/****************************************************************************
** FUNCTION NAME: ChatSwitch
** 
** PURPOSE: Callback procedure for user info label/toggle widgets.  When
**    label is clicked on, the widget reverses video automatically.  This
**    is used to signal whether you are hiding/unhiding text from other 
**    subscribers or turning on/off the anonymous mode for yourself.  
**    Whatever action you take, several structure values are set and an
**    information packet is sent off.
** 
** CALLED BY: chat_label widget created in AddChatZone() in chat.c
** 
** INPUTS: widget, user number, and normal widget callback data
** 
** CALLS: InitializeAnonPanel() in chat.c
**        PackegeHandle() in chat.c
** 
*/
void ChatSwitch(w, n, call_data)
Widget    w;
int       n;
XtPointer call_data;
{
  Arg arg;
  char dp[100];
  Packege packege;

  XtSetArg(arg,XtNlabel,dp);

  if(chatflag[LocalSubID][n]) {  /* check if label's turned "off".          */
                                 /* then test if it's user's own label      */
				    /* and if anonymous mode is on...       */
     if(LocalSubID == n && chatinf[n].anon_on) {               /*       -RYA*/
       XtDestroyWidget(anon_top);     /*if so, kill toplvl anon widget  -RYA*/
       chatinf[n].anon_on = FALSE;    /* set anon mode flag to off...   -RYA*/
       XtSetSensitive(chatinf[n].chat_wind, TRUE);  /* turn normal text -RYA*/
     }                                             /*      window back on.  */
     sprintf(dp," [ ON ]   %s",chatinf[n].name);   /* new label */
     chatflag[LocalSubID][n] = 0;               /* set user's chatflag to 0 */
     packege.id = LocalSubID;                   /* set pkg id to user       */
     packege.flag = CHAT_FLAG_ON;               /* set pkg chatflag on      */
     sprintf(packege.string,"%d",n);            /* place user # in pkg      */
     strcpy(hidding,"");                        /* set hidding string       */

   } 
  else {
                                    /* otherwise, label is "on", so...      */
                                    /* test if user's own lable and if      */
                                    /* anonymous mode is off....            */
     if(LocalSubID == n && !chatinf[n].anon_on) {              /*       -RYA*/
       chatinf[n].anon_on = TRUE;      /* if so, set anon flag on...    -RYA*/
       InitializeAnonPanel(w);         /* create anon typing window...  -RYA*/
       XtSetSensitive(chatinf[n].chat_wind, FALSE); /* turn normal text -RYA*/
                                                    /*      window off...   */
                                              /* and replace label wording. */
       sprintf(dp," [OFF - Use Anonymous Typing Window For %s]",     /* -RYA*/
	       chatinf[n].name);                                     /* -RYA*/
     }

     else            /* otherwise, it's not own label, so just turn off.-RYA*/
       sprintf(dp," [ OFF ] %s",chatinf[n].name);
     chatflag[LocalSubID][n] = 1;        /* set this chatflag value to 1    */
     packege.id = LocalSubID;            /* set pkg id to user              */
     packege.flag = CHAT_FLAG_OFF;       /* set pkg chatflag on             */
     sprintf(packege.string,"%d",n);     /* place user # in pkg             */
     strcpy(hidding,"");                 /* set hidding string              */

   }

   XtSetValues(chatinf[n].chat_label,&arg,1); /* set widget values          */
   PackegeHandle(&packege);                   /* send out the package       */

} /* end ChatSwitch */






/****************************************************************************
** FUNCTION NAME: DeleteChatZone
** 
** PURPOSE: Remove chat zone for departed subscriber.
** 
** CALLED BY: DeleteSubscriber() in subscriber.c
** 
** INPUTS: pointer to structure type Subscriber.
** 
** CALLS:
** 
** 
*/
void DeleteChatZone(Target)
Subscriber *Target;
{


#  if DEBUGDeleteChatZone
	fprintf( stderr, "Chat():DeleteChatZone(): \n" );
#  endif


   XtDestroyWidget(chatinf[Target->SubscriberID].chat_wind);
   XtDestroyWidget(chatinf[Target->SubscriberID].chat_label);

   chatinf[Target->SubscriberID].chat_wind = 0;
}





/****************************************************************************
** FUNCTION NAME: ChatEventHandler
** 
** PURPOSE: Takes action based on keypresses within chat windows.
** 
** CALLED BY: InitializeAnonPanel() as callback for anonpub widget,
**            AddChatZone() as callback for each user's chat window widget.
** 
** INPUTS: widget, user id, event type Xevent
** 
** CALLS:
** 
** 
*/
void ChatEventHandler(w,id,event)
Widget w;
int id;
XEvent *event;
{


  char            key[5];
  long            stdinout;
  XawTextBlock    block;
  int             keycount;
  KeySym          keysym;
  Packege         Outgoing;

#  if DEBUGChatEventHandler
	fprintf( stderr, "ChatEventHandler(): \n" );
#  endif

  keycount = XLookupString(event,key,10,&keysym,&stdinout);

  key[keycount] = NULL;

  if (!chatflag[LocalSubID][LocalSubID]) { 

     Outgoing.id = LocalSubID;
     Outgoing.flag = CHAT_CONTENCE;
     Outgoing.p_keysym = keysym;
     strcpy(Outgoing.string,key);
  
     PackegeHandle(&Outgoing);

  } else {

       if (keysym >= XK_space && keysym <= XK_asciitilde) {
           
          if (!strlen(hidding)) strcpy(hidding,key);
          else strcat(hidding,key);
          hidding[strlen(hidding)] = NULL;

       } else  

          if (keysym == 65288 || keysym == 65535) 
             hidding[strlen(hidding)-1] = NULL;
         
           else 
            
             if (keysym == 65293 ) {

                Outgoing.id = LocalSubID;
                Outgoing.flag = CHAT_HIDDEN;
                sprintf(Outgoing.string,"%s\n",hidding);
                strcpy(hidding,"");
                if (CHAT) ChatHidDisplay(&Outgoing);
                PackegeHandle(&Outgoing);

	      }
     }
}


/****************************************************************************
** FUNCTION NAME: PackegeHandle
** 
** PURPOSE: 
** 
** CALLED BY: 
** 
** INPUTS: 
** 
** CALLS: 
** 
** 
*/
void PackegeHandle(pack)
Packege *pack;
{

register Subscriber *Who;
int                n;
Packege OutPack;


#  if DEBUGPackegeHandle
	fprintf( stderr, "Chat():PackegeHandle(): %d\n",pack->id );
#  endif
#  if DEBUGPackegeHandle
fprintf(stderr,
"id %d | flag %d | depth %d | x %d | y %d | x1 %d | y1 %d \n",
 pack->id , pack->flag , pack->depth , pack->x ,
        pack->y , pack->x1, pack->y1);
#  endif



if (mode) {
     
     if (pack->flag == BOARD_TYPING && pack->id != LocalSubID) 
         PutChar(pack);
    
     if (pack->flag == BOARD_DRAWING && pack->id != LocalSubID)
         FreeDraw(pack);

/* wahab pointer */
     if (pack->flag == BOARD_POINTER_START && pack->id != LocalSubID){
/*
         printf("Got BOARD_POINTER_STAR\n");
*/
         FreePointerStart(pack);
     }
     if (pack->flag == BOARD_POINTER && pack->id != LocalSubID)
         FreePointer(pack);
     if (pack->flag == BOARD_POINTER_END && pack->id != LocalSubID)
         FreePointerEnd(pack);

     if (pack->flag == BOARD_ERASSING && pack->id != LocalSubID)
         Erasor(pack);

     if (pack->flag == CHAT_FLAG_ON) {
     
         n = atoi(pack->string);
         chatflag[pack->id][n] = 0;
         return;

       }

     if (pack->flag == CHAT_FLAG_OFF) {
     
         n = atoi(pack->string);
         chatflag[pack->id][n] = 1;
         return;

       }


     if (pack->flag == CHAT_CONTENCE && pack->id != LocalSubID 
              && (!chatflag[pack->id][LocalSubID])) {
         ChatDisplay(pack);
}

     if (pack->flag == CHAT_HIDDEN && pack->id != LocalSubID) {    
         ChatHidDisplay(pack);
}

/*
REPLACING THIS FOR LOOP WITH Multicast CALLS


     for ( Who=FirstSubscriber(); Who != NULL; Who=NextSubscriber() ) {

         if (Who->SubscriberID != LocalSubID ) {
   
               if (Who->SubscriberID == (BYTE) pack->id) continue;
               if (chatflag[pack->id][Who->SubscriberID] && pack->flag == CHAT_CONTENCE)
                    continue;

#  if DEBUGPackegeHandle
	fprintf( stderr, "Chat():PackegeHandle(): Sending packege to id : %d\n",Who->SubscriberID );
#  endif

           if (Who->ChatConnect != 0) {
 
bcopy(pack, &OutPack, sizeof(Packege));
htonConvertChat(&OutPack);
		if (SendData( Who->ChatConnect,&OutPack,sizeof(Packege))==ERROR) EndSession();
              
	    }
           }
        }
END OF REPLACING LOOP
*/
MulticastFlag[0]=FALSE;
MulticastFlag[pack->id]=FALSE;
     for ( Who=FirstSubscriber(); Who != NULL; Who=NextSubscriber() ) 
              if (chatflag[pack->id][Who->SubscriberID] && pack->flag == CHAT_CONTENCE)
		MulticastFlag[Who->SubscriberID]=FALSE;

bcopy(pack, &OutPack, sizeof(Packege));
htonConvertChat(&OutPack);
MulticastChat(&OutPack,sizeof(Packege));
MulticastFlag[0]=TRUE;
MulticastFlag[pack->id]=TRUE;
     for ( Who=FirstSubscriber(); Who != NULL; Who=NextSubscriber() )
              if (chatflag[pack->id][Who->SubscriberID] && pack->flag == CHAT_CONTENCE)
                MulticastFlag[Who->SubscriberID]=TRUE;



    } 
      else  {

         if (LocalSubID != (BYTE) pack->id) {

            if (pack->flag == CHAT_CONTENCE){
               ChatDisplay(pack);
	     }

            if (pack->flag == CHAT_HIDDEN){
               ChatHidDisplay(pack);
	     }
   
            if(pack->flag == BOARD_RESIZING)
               Resize(pack);

            if (pack->flag == BOARD_TYPING) 
               PutChar(pack);
    
            if (pack->flag == BOARD_DRAWING)
               FreeDraw(pack);
/* wahab pointer */

     if (pack->flag == BOARD_POINTER_START)
         FreePointerStart(pack);
     if (pack->flag == BOARD_POINTER)
         FreePointer(pack);
     if (pack->flag == BOARD_POINTER_END)
         FreePointerEnd(pack);

            if (pack->flag == BOARD_ERASSING)
               Erasor(pack);

         } else {

#  if DEBUGPackegeHandle
	fprintf( stderr, "Chat():PackegeHandle(): Sending packege to Chairman ..\n" );
#  endif
           if (ChatSocket != 0){
bcopy(pack, &OutPack, sizeof(Packege));
htonConvertChat(&OutPack);
		if (SendData( ChatSocket,&OutPack,sizeof(Packege))==ERROR) EndSession();
	     }
         }
      }

}








/****************************************************************************
** FUNCTION NAME: ChatHidDisplay
** 
** PURPOSE: 
** 
** CALLED BY: 
** 
** INPUTS: 
** 
** CALLS: 
** 
** 
*/
void ChatHidDisplay(packege)
Packege *packege;
{

   XawTextBlock  block;
   int j,n;

#  if DEBUGChatHidDisplay
	fprintf( stderr, "ChatHidDisplay(): Displaying hidding messege ..\n" );
#  endif

   if (CHAT){
   n = strlen(packege->string);
   
   block.firstPos = 0;
   block.length = n;
   block.ptr = &packege->string[0];

    j = pub_count;
    pub_count = pub_count + n;
    XawTextReplace(chatpub,j,j,&block);
    XawTextSetInsertionPoint(chatpub,pub_count);
 }

}


/****************************************************************************
** FUNCTION NAME: ntohConvertChat
** 
** PURPOSE:
** 
** CALLED BY:
** 
** INPUTS:
** 
** CALLS:
** 
** 
*/
ntohConvertChat(InPackege)
Packege *InPackege;
{
#  if DEBUGConvertChat
fprintf(stderr,
"ntohConvertChat: before: id %d | flag %d | depth %d | x %d | y %d | x1 %d | y1 %d | p_keysym  %d\n",
 InPackege->id , InPackege->flag , InPackege->depth , InPackege->x ,
        InPackege->y , InPackege->x1, InPackege->y1, InPackege->p_keysym);
#  endif
InPackege->id = ntohl(InPackege->id);
InPackege->flag = ntohl(InPackege->flag);
InPackege->depth = ntohl(InPackege->depth);
InPackege->x = ntohl(InPackege->x);
InPackege->y = ntohl(InPackege->y);
InPackege->x1 = ntohl(InPackege->x1);
InPackege->y1 = ntohl(InPackege->y1);
InPackege->p_keysym = ntohl(InPackege->p_keysym);
#  if DEBUGConvertChat
fprintf(stderr,
"ntohConvertChat: after: id %d | flag %d | depth %d | x %d | y %d | x1 %d | y1 %d | p_keysym %d \n",
 InPackege->id , InPackege->flag , InPackege->depth , InPackege->x ,
        InPackege->y , InPackege->x1, InPackege->y1, InPackege->p_keysym);
#  endif

}


/****************************************************************************
** FUNCTION NAME: htonConvertChat
** 
** PURPOSE: 
** 
** CALLED BY: 
** 
** INPUTS: 
** 
** CALLS: 
** 
** 
*/
htonConvertChat(OutPackege)
Packege *OutPackege;
{
#  if DEBUGConvertChat
fprintf(stderr,
"htonConvertChat: before : id %d | flag %d | depth %d | x %d | y %d | x1 %d | y1 %d | p_keysym %d \n",
 OutPackege->id , OutPackege->flag , OutPackege->depth , OutPackege->x ,
        OutPackege->y , OutPackege->x1, OutPackege->y1, OutPackege->p_keysym);
#  endif
OutPackege->id = htonl(OutPackege->id);
OutPackege->flag = htonl(OutPackege->flag);
OutPackege->depth = htonl(OutPackege->depth);
OutPackege->x = htonl(OutPackege->x);
OutPackege->y = htonl(OutPackege->y);
OutPackege->x1 = htonl(OutPackege->x1);
OutPackege->y1 = htonl(OutPackege->y1);
OutPackege->p_keysym = htonl(OutPackege->p_keysym);
#  if DEBUGConvertChat
fprintf(stderr,
"htonConvertChat: after : id %d | flag %d | depth %d | x %d | y %d | x1 %d | y1 %d | p_keysym %d \n",
 OutPackege->id , OutPackege->flag , OutPackege->depth , OutPackege->x ,
        OutPackege->y , OutPackege->x1, OutPackege->y1, OutPackege->p_keysym);
#  endif

}
