/*
 * Copyright (c) 1991, 1993 Old Dominion University.
 * Copyright (c) 1991, 1993 University of North Carolina at Chapel Hill.
 * All rights reserved.
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  Old Dominion University and the University of North
 * Carolina at Chapel Hill make no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 */

/*
	protocol.c
	Routines to handle the protocol

*/

#define DEBUGHandleStoHProtocol	0&DEBUGON
#define DEBUGHandleHtoSProtocol	0&DEBUGON

#include <stdio.h>
#include <X11/Intrinsic.h>

#include "protocol.h"
#include "subscriber.h"
#include "network.h"
#include "htos.h"
#include "stoh.h"
#include "tool.h"


/*
	Define the protocol table.  This is simply a table of
	routines to call when certain requests come in.

	Each function takes the following arguments:

		Socket		Port	- File descriptor of the socket in question
		Subscriber	*Who	- Structure showing who sent it (NULL for HtoS)

*/

typedef struct _ProtoTableEntry {
	void	(*StoHFunc)();	/* Function to call in Viewer-To-Host Protocol */
	void	(*HtoSFunc)();	/* Function to call in Host-To-Viewer Protocol */
} ProtoTableEntry;

ProtoTableEntry	ProtoTable[31] = {
/*** Basic control mechanisms */
/* 0 */        { NULL,                 NULL                    },
/* 1 */        { NULL,                 NULL                    },
/* 2 */	{ STOH_Reply,		HTOS_X11Packet		},
/* 3 */	{ NULL,			HTOS_InternalAtom	},
/* 4*/	{ NULL,			HTOS_YouAre		},
/* 5 */  { NULL,                 HTOS_AllocColor         },
/* 6 */   { NULL,			HTOS_AllocNamedColor	}, /* FSH 16-SEP-91 */
/* 7 */	{ STOH_LeavingSession,	HTOS_EndSession		},
/* 8 */	{ STOH_Event,		HTOS_DroppedYou		},
/* wang_x add */
/* 9 */        { STOH_ResizeWindowFunc,HTOS_ResizeWindowFunc   },
/* end wang_x add */

/** Client and tool list management */
/* 10 */	{ STOH_ConfirmNewTool,	HTOS_AddNewSubscriber	},
/* 11 */	{ NULL,			HTOS_DeleteSubscriber	},
/* 12 */	{ STOH_NewToolLocal,	HTOS_AddNewTool		},
/* 13 */        { STOH_NewToolRemote,   HTOS_IncomingTools      },
/* 14 */	{ STOH_TerminateTool,	HTOS_DeleteTool		},
/* 15 */        { STOH_StopTool,        HTOS_EndIncoming        },
/* 16 */        { STOH_WantJoinTools,   HTOS_ToolsToJoin        },
/* 17 */        { STOH_JoinTools,       HTOS_CatchUpStart	},
/* 18 */        { NULL,                 HTOS_CatchUpEnd		},
/* 19 */        { NULL,                 NULL                    },
/** Token passing mechanisms */
/* 20 */	{ STOH_WantToken,	HTOS_TakeToken		},
/* 21 */	{ STOH_GiveTokenToMe,	HTOS_ForceTokenRelease	},
/* 22 */	{ STOH_ReleasingToken,	HTOS_AskTokenRelease	},
/* 23 */	{ STOH_DeQueToken,      HTOS_CreatorTakeToken 	},
/* 24 */   { NULL,                 HTOS_ContinueTakeToken   },
/** others */
/*
 * Begin Murray Addition
 */
/** list requests of the host **/
/* 25 */{ STOH_QueuedList,      HTOS_QueuedList         },
/*
 * End Murray Addition
 */
/* 26 */        { NULL,                 NULL                    },
/* 27 */        { NULL,                 NULL                    },
/* 28 */        { NULL,                 NULL                    },
/* 29 */        { NULL,                 NULL			},
/* 30 */        { STOH_SendArchivePacket, NULL			}
};


/*
 *	HandleStoHProtocol()
 *
 *	Handle incoming viewer-to-host requests.
 *
 *	RETURN VALUE:	void
 */
XtInputCallbackProc HandleStoHProtocol( Who, Port, ID )
Subscriber *Who;
Socket *Port;
XtInputId *ID;
{
	BYTE	Request;

	/* Check for a bogus request */
	if ( (Request=CheckByte( *Port )) == NothingAvail) {
		return;
	}


	/* Reset the toolkit before doing anything else */
	XtRemoveInput( *ID );
	Who->InputCallbackID = XtAppAddInput( GetAppCon(), *Port, XtInputReadMask,
	    HandleStoHProtocol, Who );

	if ( (Request <= MAX_PROTO_REQUEST) && ((ProtoTable[Request].StoHFunc) != NULL) ) {
#		if DEBUGHandleStoHProtocol
		fprintf( stderr, "STOH(): Performing call [%03d]\n", Request );
#		endif
		(ProtoTable[Request].StoHFunc)( *Port, Who );
	} else {
#		if DEBUGHandleStoHProtocol
		fprintf( stderr, "STOH(): Got NULL request [%03d] from host\n", Request );
#		endif
	}
}


/*
 *	HandleHtoSProtocol()
 *
 *	Handle incoming host-to-viewer requests.
 *
 *	RETURN VALUE:	None to speak of
 */
XtInputCallbackProc HandleHtoSProtocol( Who, Port, ID )
Subscriber *Who;
Socket *Port;
XtInputId *ID;
{
	BYTE	Request;

	/* Check for a bogus request */
	if ( (Request=CheckByte( *Port )) == NothingAvail) {
		return;
	}

	/* Reset the toolkit before leaving */
	XtRemoveInput( *ID );
	XtAppAddInput( GetAppCon(), *Port, XtInputReadMask, HandleHtoSProtocol, Who );
			
	if ( (Request <= MAX_PROTO_REQUEST) && ((ProtoTable[Request].HtoSFunc) != NULL) ) {
#		if DEBUGHandleHtoSProtocol
		if ( Request != HS_X11Packet ) {
			fprintf( stderr, "HTOS(): Performing call [%03d]\n", Request );
		}
#		endif
		(ProtoTable[Request].HtoSFunc)( *Port, Who );	/* "Who" gets ignored here */
	} else {
#		if DEBUGHandleHtoSProtocol
		fprintf( stderr, "HTOS(): Got NULL request [%03d] from host\n", Request );
#		endif
	}

}
