/*
 * Copyright (c) 1991, 1993 Old Dominion University.
 * Copyright (c) 1991, 1993 University of North Carolina at Chapel Hill.
 * All rights reserved.
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  Old Dominion University and the University of North
 * Carolina at Chapel Hill make no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 */

/*
   session_info.c
   XTV Information Interface portion of XTV

*/


#include <stdio.h>
#include <errno.h>
#include <sys/time.h>

#include <sys/param.h>
/*
#include <sys/socket.h>
*/
#include <sys/file.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
/*
#include <netdb.h>
*/

#include <X11/Intrinsic.h>

#include "xtvinfo.h"
#include "subscriber.h"
#include "tool.h"

#define MAXHOSTNAME 80
#define BUFSIZE 1024
#define MAX_SESSIONID 128


int     socket_id;                     /* datagram socket id             */
int     daemon_known = FALSE;          /* true of daemon data file there */

char buf[BUFSIZE];                     /* packer buffer                  */
char session_id[MAX_SESSIONID];        /* XTV SII session identifier     */

struct sockaddr_in server;             /* XTV SII daemon address         */

extern XtAppContext   AppCon;          /* application context            */


void socket_read();

/******************************************************************************
socket_read() - reads a datagram packet from a socket
******************************************************************************/

void socket_read(client_data, s_id, id)
caddr_t client_data;
int     *s_id;
XtInputId *id;
{
   int    rc;
   int    fromlen;
   char   readbuf[BUFSIZE];
   struct sockaddr_in from;

   fromlen = sizeof(from);
   if ((rc=recvfrom(*s_id, readbuf, sizeof(readbuf), 0, &from, &fromlen)) < 0)
      perror("receiving datagram  message");

    readbuf[rc] = '\0';

   if (readbuf[0] == SESSION_QONEINFO)
      process_oneinfo(readbuf, *s_id, from);
   else if (readbuf[0] == SESSION_QALIVE)
      process_alive(readbuf, *s_id, from);
}


/******************************************************************************
process_alive() - processes an ?ALIVE? request packer from the daemon
******************************************************************************/

process_alive(buf, socket_id, send_addr)
char   *buf;
int     socket_id;
struct  sockaddr_in send_addr;
{
   /* ignore packets that do not match the session_id */
   if (!strcmp(buf+1, session_id))
   {
      /* resend same message but with AALIVE instead of QALIVE */
      buf[0] = (char)  SESSION_AALIVE;

      if (sendto(socket_id, buf, strlen(buf), 0, &send_addr,
                                                     sizeof(send_addr)) <0 )
         perror("sending datagram message");
   }
}


/******************************************************************************
proceses_oneinfo() - processes a ?ONEINFO? request from XTV SII User Interface
******************************************************************************/

process_oneinfo(buf, socket_id, send_addr)
char   *buf;
int     socket_id;
struct  sockaddr_in send_addr;
{
   int    num_lines;
   char   *buf_ptr;
   Subscriber  *a_subscriber;
   Tool        *a_tool;

   buf[0] = (char)  SESSION_AONEINFO;
   sprintf(buf, "%c       ", (char) SESSION_AONEINFO);
   buf_ptr = buf + strlen(buf);
   num_lines = 1;

   strcpy(buf_ptr, "Subscriber Information:\n");
   buf_ptr = buf_ptr + strlen(buf_ptr);

   a_subscriber = FirstSubscriber();
   if (!a_subscriber)
   {
      num_lines++;
      strcpy(buf_ptr, "No Subscribers\n");
   }
   else while (a_subscriber)
   {
      num_lines++;

      sprintf(buf_ptr, "%s@%s [%s]\n", a_subscriber->LoginName,
                        a_subscriber->FullHostName, a_subscriber->FullName);
      buf_ptr = buf_ptr + strlen(buf_ptr);
      a_subscriber = NextSubscriber();
   }

   num_lines++;
   strcpy(buf_ptr, "Tools in use:\n");
   buf_ptr = buf_ptr + strlen(buf_ptr);

   a_tool = FirstTool();

   if (!a_tool)
   {
      num_lines++;
      strcpy(buf_ptr, "No tools in use\n");
   }
   else while (a_tool)
   {
      num_lines++;
      sprintf(buf_ptr, "%s\n", a_tool->ToolCommand);
      buf_ptr = buf_ptr + strlen(buf_ptr);
      a_tool = NextTool();
   }

   sprintf(buf+1, "%d", num_lines);
   buf[strlen(buf)] = ' ';

   if (sendto(socket_id, buf, strlen(buf), 0, &send_addr,
                                                      sizeof(send_addr)) <0 )
      perror("sending datagram message");
}


/******************************************************************************
initialize_XTV_SII() - initializes processing for XTV SII
******************************************************************************/

initialize_XTV_SII(SessionNumber)
int SessionNumber;
{
   send_start_message(SessionNumber);

   XtAppAddInput(AppCon, socket_id, XtInputReadMask, socket_read, NULL);
}


/******************************************************************************
get_localhost_name() - gets the local hosts name
******************************************************************************/

/* get local host information, NAME and INET ADDRESS */
int get_localhost_name(hostname, inet_address)
char *hostname;
char *inet_address;
{
   struct hostent *localhost;
   struct hostent *gethostbyname();

   gethostname(hostname, MAXHOSTNAME);

   if  ((localhost = gethostbyname(hostname)) == NULL)
   {
      fprintf(stderr, "Can't find host %s\n", hostname);
      return -1;
   }

   strcpy(hostname, localhost->h_name);
   strcpy(inet_address, inet_ntoa(localhost->h_addr));

   return 1;
}


/******************************************************************************
get_daemon_info() - gets XTV SII daemon information
******************************************************************************/

int get_daemon_info(hostname, portid)
char *hostname;
char *portid;
{
   FILE *name;

   if ((name = fopen("/tmp/xtvinfo.dat", "r")) == NULL)
      return -1;
   else
   {
      fgets(hostname, MAXHOSTNAME, name);
      fgets(portid, 100, name);
      hostname[strlen(hostname)-1] = '\0';
      portid[strlen(portid)-1] = '\0';
      fclose(name);
   }

   return 1;
}


/******************************************************************************
create_socket_for_SII() - creates socket for XTV SII processing
******************************************************************************/

int create_socket_for_SII(hostname, portid, server)
char *hostname;
char *portid;
struct sockaddr_in *server;
{
   struct hostent     *host;
   struct hostent     *gethostbyname();

   if  ((host = gethostbyname(hostname)) == NULL )
   {
      struct sockaddr_in addr;

      addr.sin_addr.s_addr = inet_addr(hostname);
      if ((host = gethostbyaddr(&addr.sin_addr.s_addr,
          sizeof(addr.sin_addr.s_addr),AF_INET)) == NULL)
      {
         fprintf(stderr, "Can't find host %s\n", hostname);
         /*
         return -1;
         */
      }
   }

   if (host)
   bcopy (host->h_addr, &(server->sin_addr), host->h_length);

   /* Construct name of socket to send to. */
   server->sin_family = host->h_addrtype;
   server->sin_port   = htons(atoi(portid));

   /* Create socket on which to send  and receive */
   return (socket(host->h_addrtype, SOCK_DGRAM, 0));
}



/******************************************************************************
send_start_message() - sends <Start> message to daemon
******************************************************************************/

send_start_message(SessionNumber)
int SessionNumber;
{

char hostname[MAXHOSTNAME];
char portid[80];

int     port_num;
char    date_time[80];
time_t  t;
char    localhost_name[80];
char    inet_address[80];
extern char Subject[];
char    host_sn[100];
int fromlen;
int length;
int rc, cc;
int done;
int return_info;
struct sockaddr_in this_server;
Subscriber  *a_subscriber;

   if (get_localhost_name(localhost_name, inet_address) < 0)
   {
      perror("can not get localhost information");
      return -1;
   }

   /* get daemon info */
   if (get_daemon_info(hostname, portid) < 0)
   {
      daemon_known = FALSE;
      return;
   }

   daemon_known = TRUE;

   if ((socket_id = create_socket_for_SII(hostname, portid, &server)) < 0)
   {
      perror("can not create socket to xtv mgr");
      return -1;
   }

   this_server.sin_family = AF_INET;
   this_server.sin_addr.s_addr = htonl(INADDR_ANY);
   this_server.sin_port = htons(0);

   if (bind(socket_id, &this_server, sizeof(this_server) ) ) {
      close(socket_id);
      perror("binding name to datagram socket");
      exit(-1);
   }

   length = sizeof(this_server);
   if (getsockname (socket_id,&this_server,&length) ) {
      perror("getting socket name");
      exit(0);
   }

   port_num = ntohs(this_server.sin_port);

   time(&t);
   strcpy(date_time, ctime(&t));

   if (date_time[strlen(date_time)-1] == '\n') /* strip terminating linefeed */
      date_time[strlen(date_time)-1] = NULL;
   while (date_time[strlen(date_time)-1] != ' ') /* do not show the year     */
      date_time[strlen(date_time)-1] = NULL;
   while (date_time[strlen(date_time)-1] == ' ')
      date_time[strlen(date_time)-1] = NULL;

   a_subscriber = FirstSubscriber();
   if (a_subscriber)
      sprintf(host_sn, "%s:%d", a_subscriber->FullHostName, SessionNumber);
   else
      strcpy(host_sn," UNKNOWN_HOST:%d", SessionNumber);

   sprintf(buf, "%c%-22s / %s / %-15s / %-18s /%d", (char) SESSION_START, 
                host_sn, date_time, ChairmanName(), Subject, port_num);
   sprintf(session_id, "%-22s / %s", host_sn, date_time);

   rc = strlen(buf);

   if (sendto(socket_id, buf, rc, 0, &server, sizeof(server)) <0 )
      perror("sending datagram message");
}


/******************************************************************************
terminate_session() - sends <Terminate> messsage to daemon and closes socket
******************************************************************************/

terminate_session()
{
   int rc;

   if (daemon_known)
   {
      buf[0] = (char) SESSION_TERMINATE;
      rc = strlen(buf);

      if (sendto(socket_id, buf, rc, 0, &server, sizeof(server)) <0 )
        perror("sending datagram message");
   }

   close(socket_id);
}

