/*
 * Copyright (c) 1991, 1993 Old Dominion University.
 * Copyright (c) 1991, 1993 University of North Carolina at Chapel Hill.
 * All rights reserved.
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  Old Dominion University and the University of North
 * Carolina at Chapel Hill make no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 */


/*
 **  spacket.c  - Handle unique cases where HTOS must receive a special RID 
 **               from host. Notably : InternAtom, AllocColor and 
 **               AllocNamedColor.
 **
 **
 **   FSH 4-SEP-91
 */

#include <stdio.h>
#include <stdlib.h>
#include <X11/X.h>
#include <X11/Intrinsic.h>
#define  NEED_REPLIES      /* Tell the header to include the replies */
#define  NEED_EVENTS
#include <X11/Xproto.h>
#include "tool.h"
#include "protocol.h"
#include "memory.h"
#include "misc.h"
#include "spacket.h"


#define PAD(x) ((4 - (x%4)) % 4)
#define DEFENSE_ON 0             /* Only turn on if things go flakey */
#define NUM_SPECIAL_REQUESTS 3   

#define  DEBUGtoSTDOUT   1

#if DEBUGtoSTDOUT
#define  TRACE_PLACE  stdout
#else
#define  TRACE_PLACE  stderr
#endif

#define  DEBUGHandleClientSpecialRequest 0&DEBUGON
#define  DEBUGHandleSubSpecialRequest  0&DEBUGON
#define  DEBUGHandleJoinSpecialRequest  0&DEBUGON
#define  DEBUGHandleClientInternAtom  0&DEBUGON
#define  DEBUGHandleClientAllocColor  0&DEBUGON
#define  DEBUGHandleClientAllocNamedColor 0&DEBUGON 
#define  DEBUGHandleSubscrInternAtom  0&DEBUGON
#define  DEBUGHandleSubscrAllocColor  0&DEBUGON
#define  DEBUGHandleSubscrAllocNamedColor  0&DEBUGON 
#define  DEBUGHandleJoinInternAtom  0&DEBUGON
#define  DEBUGHandleJoinAllocColor  0&DEBUGON
#define  DEBUGHandleJoinAllocNamedColor 0&DEBUGON 
#define  DEBUGSpecialQuery 0&DEBUGON


#if _lint
static void HandleClientInternAtom(Tool *, BYTE *);
static void HandleClientAllocColor(Tool *, BYTE *);
static void HandleClientAllocNamedColor(Tool *, BYTE *);
static void HandleSubscrInternAtom(Tool *, BYTE *);
static void HandleSubscrAllocColor(Tool *, BYTE *);
static void HandleSubscrAllocNamedColor(Tool *, BYTE *);
static void HandleJoinInternAtom( Tool *, xReq *, BYTE *);
static void HandleJoinAllocColor( Tool *, xReq *, BYTE *);
static void HandleJoinAllocNamedColor( Tool *,xReq *,  BYTE *);
#else
static void HandleClientInternAtom();
static void HandleClientAllocColor();
static void HandleClientAllocNamedColor();
static void HandleSubscrInternAtom();
static void HandleSubscrAllocColor();
static void HandleSubscrAllocNamedColor();
static void HandleJoinInternAtom();
static void HandleJoinAllocColor();
static void HandleJoinAllocNamedColor();
#endif

extern BYTE  MulticastFlag[];  /* If TRUE, include Subscriber in Multicast*/


typedef struct _ReplyTableEntry {
   char   PacketName[30];      /* MIT Standard Packet Name */
   BYTE   HandlingCode;      /* How TranslatePacket is to handle the request */
   int   OffsetToID[10];      /* Table of offsets to simple XIDs */
   void   (*ProcessFunc)();   /* Function to handle any special processing */
} ReplyTableEntry;
extern ReplyTableEntry ReplyTable[];


/*
  ** Follow the table style format for these special calls.
  ** Not enough requests yet to justify full table. 
  */

typedef struct  SpecialRequestTableEntry {
   char  PacketName[30];    /* MIT Related Packet Name */
   BYTE    opcode;
   void  (*ClientProcessFunc)(); /* Treat Incoming packet */
   void  (*SubscrProcessFunc)(); /* Special to subscriber */
   void  (*JoinProcessFunc)();   /* Additional Join       */
} SpecialRequestTableEntry;

typedef struct  SpecialQueryTableEntry {
   char  *PacketName;    /* MIT Related Packet Name */
   BYTE    opcode;
} SpecialQueryTableEntry;

SpecialRequestTableEntry SpecialRequest[NUM_SPECIAL_REQUESTS] = {
   {"InternAtom",X_InternAtom,HandleClientInternAtom, HandleSubscrInternAtom,
   HandleJoinInternAtom},
   {"AllocColor",X_AllocColor,HandleClientAllocColor,HandleSubscrAllocColor,
   HandleJoinAllocColor},
   {"AllocNamedColor",X_AllocNamedColor,HandleClientAllocNamedColor,
   HandleSubscrAllocNamedColor,HandleJoinAllocNamedColor},
};


SpecialQueryTableEntry SpecialQuery[] = {
   {"GetMotionEvents",X_GetMotionEvents},
   {"GetInputFocus",X_GetInputFocus},
   {"GetImage",X_GetImage},
   {"GetWindowAttributes",X_GetWindowAttributes},
   {"GetGeometry",X_GetGeometry},
   {"GetSelectionOwner",X_GetSelectionOwner},
   {"GetPointerControl",X_GetPointerControl},
   {"GetPointerMapping",X_GetPointerMapping},
   {"GrabPointer",X_GrabPointer},
   {"GrabKeyboard",X_GrabKeyboard},
   {"QueryPointer",X_QueryPointer},
   {"QueryTree",X_QueryTree},
   {"TranslateCoords",X_TranslateCoords},
   {NULL,0}
};



SpecialQueryTableEntry HostQuery[] = {
   {"AllocNamedColor",X_AllocNamedColor},
   {"AllocColor",X_AllocColor},
   {"AllocColorCells",X_AllocColorCells},
   {"AllocColorPlanes",X_AllocColorPlanes},
   {"GetAtomName",X_GetAtomName},
   {"InternAtom",X_InternAtom},
   {"ListProperties",X_ListProperties},
   {"ListFonts",X_ListFonts},
   {"ListFontsWithInfo",X_ListFontsWithInfo},
   {"ListInstalledColormaps",X_ListInstalledColormaps},
   {"ListHosts",X_ListHosts},
   {"ListExtensions",X_ListExtensions},
   {"LookupColor",X_LookupColor},
   {"GetFontPath",X_GetFontPath},
   {"GetKeyboardMapping",X_GetKeyboardMapping},
   {"GetKeyboardControl",X_GetKeyboardControl},
   {"GetScreenSaver",X_GetScreenSaver},
   {"GetModifierMapping",X_GetModifierMapping},
   {"GetProperty",X_GetProperty},
   {"QueryKeymap",X_QueryKeymap},
   {"QueryFont",X_QueryFont},
   {"QueryTextExtents",X_QueryTextExtents},
   {"QueryColors",X_QueryColors},
   {"QueryBestSize",X_QueryBestSize},
   {"QueryExtension",X_QueryExtension},
   {"SetModifierMapping",X_SetModifierMapping},
   {"SetPointerMapping",X_SetPointerMapping},
   {NULL,0}
};



/*
    ** define and initialize our packet type
    */
static struct xtv_ACP xtv_ACPacket = {
   HS_AllocColor,
   (CARD8) 0,
   (CARD16) 0,
   (CARD16) 0,
   (CARD16) 0,
   (CARD32) 0,
   (CARD32) 0 
};



/*
    ** define and initialize our AllocateNamedColor packet type
    */
static struct xtv_ANCP xtv_ANCPacket = {
   HS_AllocNamedColor,
   (CARD8) 0,
   (CARD16) 0,
   (CARD32) 0,
   (CARD32) 0,
   NULL 
};



/*
  ** static variables
  */
static CARD32 AtomValue;        /* Loacl host atom RID  */
static CARD32 PixelValue;       /* Local host pixel RID */
static int dispatch_index = 0;  /* current dispatch table index */


/*
  ** externs 
  */
extern BYTE XCOLOR;             /* Color mode on/off */
extern BOOL SpecialQueryFlag;  /* used to indicate Special Query */
extern BOOL HostQueryFlag;  /* used to indicate Host Query */



/*
** HandleClientSpecialRequest() - Dispatch to functions requiring
** special processing.
*/

void
HandleClientSpecialRequest(WhatTool,Header,PacketStuff)
Tool *WhatTool;       /* Current Tool     */
xReq * Header;        /* x Header info    */
BYTE *PacketStuff;    /* Packet info from x Request   */
{
   register int i;

#if DEBUGHandleClientSpecialRequest 
   fprintf(TRACE_PLACE,"HandleClientSpecialRequest() :: opcode = %d \n",
       Header->reqType & 0x00ff);
#endif    

   dispatch_index = -1;  /* should always be called first so initialize */
   for (i = 0; SpecialRequest[i].ClientProcessFunc != NULL; i++) {
      if (SpecialRequest[i].opcode == Header->reqType) {
         dispatch_index = i; /* set dispatch index to save time */
         SpecialRequest[i].ClientProcessFunc(WhatTool,PacketStuff);
         return;
      }
   }
   return;

}

/*
** SpecialQuery() - set SpecialQueryFlag or HostQueryFlag to TRUE 
*/

HandleSpecialQuery(WhatTool,Request)
Tool *WhatTool;       /* Current Tool     */
BYTE Request;        /* x Request number    */
{
   register int i;

#if DEBUGSpecialQuery
   printf("Request[%d] %s\n", Request, ReplyTable[Request].PacketName);
#endif

   for (i = 0; SpecialQuery[i].PacketName != NULL; i++) {
      if (SpecialQuery[i].opcode == Request) {
         SpecialQueryFlag = TRUE;

#if DEBUGSpecialQuery
         fprintf(TRACE_PLACE, "SpecialQuery() :: Request_opCode %d PacketName = %s\n", 
             Request, SpecialQuery[i].PacketName);
#endif

         return;
      }
   }
   for (i = 0; HostQuery[i].PacketName != NULL; i++) {
      if (HostQuery[i].opcode == Request) {
         HostQueryFlag = TRUE;

#if DEBUGSpecialQuery
         fprintf(TRACE_PLACE, "HostQuery() :: Request_opCode %d PacketName = %s\n", 
             Request, HostQuery[i].PacketName);
#endif


         return;
      }
   }
   return;
}




/*
** HandleSubSpecialRequest() - Dispatch Subscriber sends for special packets 
*/
void
HandleSubSpecialRequest(WhatTool,Header,PacketStuff)
Tool * WhatTool;      /* Current Tool     */
xReq * Header;        /* x Header info    */
BYTE * PacketStuff;   /* Packet info from x Request   */
{

#if DEBUGHandleSubSpecialRequest
   fprintf(TRACE_PLACE,"HandleSubSpecialRequest() :: opcode = %d : index = %d \n",
       Header->reqType & 0x00ff,dispatch_index);
#endif

#if DEFENSE_ON  
   if (SpecialRequest[dispatch_index].opcode != Header->reqType) {
      /*
      ** index not previously set so reset it-somebody changed code if
      ** this happens
      */
      register int i;

      dispatch_index = -1;

      for (i = 0; SpecialRequest[i].SubscrProcessFunc != NULL; i++)
         if (SpecialRequest[i].opcode == Header->reqType) {
            dispatch_index = i;
            break;
         }

      if (dispatch_index == -1)  /* request not found */
         return;
   }
#endif   
   SpecialRequest[dispatch_index].SubscrProcessFunc(WhatTool,PacketStuff);
   return;
}




/*
** HandleJoinSpecialRequest() - Dispatch for handling archiving for
**  special packets.
*/
void
HandleJoinSpecialRequest(WhatTool,Header,PacketStuff)
Tool * WhatTool;     /* Current Tool      */
xReq * Header;       /* x Header info     */
BYTE * PacketStuff;  /* Packet info from x Request   */
{

#if DEBUGHandleJoinSpecialRequest
   fprintf(TRACE_PLACE,"HandleJoinSpecialRequest() :: opcode = %d \n",
       Header->reqType & 0x00ff);
#endif

#if DEFENSE_ON
   if (SpecialRequest[dispatch_index].opcode != Header->reqType) {
      /*
      ** reset index 
      */
      register int i;

      dispatch_index = -1;
      for (i = 0; SpecialRequest[i].SubscrProcessFunc != NULL; i++)
         if (SpecialRequest[i].opcode == Header->reqType){
            dispatch_index = i;
            break;
         }
      if (dispatch_index == -1)  /* request not found */
         return;
   }
#endif
   SpecialRequest[dispatch_index].JoinProcessFunc(WhatTool,Header,PacketStuff);

}




/*
** If we have an InternalAtom request, call the local display server
**  and look it up. 
*/
static void
HandleClientInternAtom(WhatTool,PacketStuff)
Tool *WhatTool;
BYTE *PacketStuff;
{

   AtomValue = XInternAtom( XtDisplay((Widget *)GetTopLevelWidget()),
       (char *)(PacketStuff+(sizeof(xInternAtomReq)-sizeof(xReq))), FALSE );
   AtomValue = SWAP_LONG( AtomValue,WhatTool->ConnectPrefix.byteOrder);

#if DEBUGHandleClientInternAtom
   fprintf(TRACE_PLACE, "HandleClientInternAtom(): New Atom: '%s' = %08x\n", 
       (char *)(PacketStuff+(sizeof(xInternAtomReq)-sizeof(xReq))), 
       AtomValue );
#endif
   return;
}



/*
** HandleClientAllocColor() - Handle Client Alloc Color request.
**
** If we have an AllocColor request, call the local display server
** and save pixel for shipment to subscriber. All values will be saved for
** shippment in TOOL order format for swap if needed on remote.
**
** 
**  xtv_ACPacket ->  1  HS_AllocColor 
**                   1  ToolID 
**                   2  RED
**                   2  GREEN
**                   2  BLUE
**                   4  CMAP 
**                   4  Pixel 
**  
*/
static void
HandleClientAllocColor(WhatTool,PacketStuff)
Tool * WhatTool;
BYTE * PacketStuff;    /* stuff after xReq */
{
   XColor x_color;      /* For local AllocColor call */
   CARD32 cmap;


   xtv_ACPacket.cmap =  (*(Colormap *) PacketStuff); 
   cmap =  SWAP_LONG ( (*(Colormap *) PacketStuff), 
       WhatTool->ConnectPrefix.byteOrder);
   /*
    ** Copy over RGB values 
    */
   memcpy(&x_color.red, PacketStuff + sizeof(CARD32), 6);

   /*
    ** local server is allowed to alter RGB values so save now and
    ** store in tool format
    */
   xtv_ACPacket.red = x_color.red;
   xtv_ACPacket.green = x_color.green;
   xtv_ACPacket.blue = x_color.blue;
   
   x_color.red = SWAP_INT(x_color.red,
       WhatTool->ConnectPrefix.byteOrder);
   x_color.green = SWAP_INT(x_color.green,
       WhatTool->ConnectPrefix.byteOrder);
   x_color.blue = SWAP_INT(x_color.blue,
       WhatTool->ConnectPrefix.byteOrder);

#if DEBUGHandleClientAllocColor
  fprintf(TRACE_PLACE, "HandleAllocColor(): cmap 0x%08x red %d green %d blue %d\n",
                      cmap, x_color.red, x_color.green, x_color.blue);
# endif
   XAllocColor( XtDisplay((Widget *)GetTopLevelWidget()),
       cmap,&x_color);

   /*
    ** Now build the rest of our packet for HS_AllocColor type packet
    */
   xtv_ACPacket.pixel= SWAP_LONG(x_color.pixel,
       WhatTool->ConnectPrefix.byteOrder);

#if DEBUGHandleClientAllocColor
   fprintf(TRACE_PLACE, "HandleAllocColor():PixelValue= %d \n",
       x_color.pixel);
#endif
   return;
}




/*
** HandleClientAllocNamedColor() - Make Local call and then build packet
**       to send over to subscriber and for archive
**
**      1       HS_AllocNamedColor  Protocol ID Tag
**      1       toolID              Tool ID of sender
**      2       name length         Color name length in bytes
**      4       cmap                Colormap
**      4       pixel               pixel from remote host
**      n       Name                actual name to lookup
**
**
*/
static void
HandleClientAllocNamedColor(WhatTool,PacketStuff)
Tool * WhatTool;
BYTE * PacketStuff;    /* Stuff after xReq ie CMAP */
{
   CARD16  NameLength;        /* Name length from client */
   XColor  ColorCellDef;      /* Closest RBG match       */
   XColor  RgbDbDef;          /* exact RBG equiv         */
   CARD32 cmap;


   xtv_ANCPacket.cmap =  (*(Colormap *)PacketStuff); 
   cmap =  SWAP_LONG ( (*(Colormap *) PacketStuff),
       WhatTool->ConnectPrefix.byteOrder);


   NameLength = SWAP_INT(*(CARD16 *)(PacketStuff+ sizeof(Colormap)),
       WhatTool->ConnectPrefix.byteOrder);

   /*
    ** Align on 32 bit word  ie. ((NameLength+1)/4) * 4)
    */
   NameLength++;
   NameLength = NameLength + PAD(NameLength);

   xtv_ANCPacket.nlen = SWAP_INT(NameLength,WhatTool->ConnectPrefix.byteOrder);
   xtv_ANCPacket.name = (BYTE *)GetMem(NameLength); /* succeed or BOMB out */
   memcpy(xtv_ANCPacket.name, (char *)(PacketStuff+8), NameLength - 1);
   xtv_ANCPacket.name[NameLength] = '\0';
   /*
    ** make the local call to get the pixel value to be shipped
    */
#if DEBUGHandleClientAllocNamedColor
   fprintf(TRACE_PLACE, "Calling XAllocNamedColor cmap: %08x, name: %s\n", 
       cmap, xtv_ANCPacket.name);
#endif
   XAllocNamedColor(XtDisplay((Widget *)GetTopLevelWidget()),
       cmap,xtv_ANCPacket.name, &ColorCellDef, &RgbDbDef);

   xtv_ANCPacket.pixel = SWAP_LONG( ColorCellDef.pixel,
       WhatTool->ConnectPrefix.byteOrder);

   return;

}




/*
** HandleSubscrInternAtom() - Handle Subscriber Internal Atom. 
**
** Handle send to subscriber of special internal atom request
*/
static void
HandleSubscrInternAtom(WhatTool,PacketStuff)
Tool * WhatTool;
BYTE * PacketStuff;
{
Subscriber * Who;

   CARD16 HowMany = SWAP_INT( *((CARD16 *)PacketStuff),
       WhatTool->ConnectPrefix.byteOrder);

      for ( Who=FirstSubscriber(); Who != NULL; Who=NextSubscriber() ) {
         if (Who->SubscriberID == 0) {
            MulticastFlag[0]=FALSE;
            continue;
         }


   MulticastByte(HS_InternalAtom );
   MulticastByte(WhatTool->ToolID);
   MulticastData(&AtomValue, sizeof(CARD32) );
   MulticastData(PacketStuff, sizeof(CARD16) );
   MulticastData( PacketStuff+(sizeof(xInternAtomReq)-sizeof(xReq)), HowMany );
   }

}



/*
** Packet Format : 
**              
**  xtv_ACPacket ->  1  HS_AllocColor 
**                   1  ToolID 
**                   2  RED
**                   2  GREEN
**                   2  BLUE
**                   4  CMAP
**                   4  Pixel 
**
*/
static void
HandleSubscrAllocColor(WhatTool,PacketStuff)
Tool *WhatTool;
BYTE * PacketStuff;
{
Subscriber * Who;

      for ( Who=FirstSubscriber(); Who != NULL; Who=NextSubscriber() ) {
         if (Who->SubscriberID == 0) {
            MulticastFlag[0]=FALSE;
            continue;
         }


#if DEBUGHandleSubscrAllocColor
   fprintf(TRACE_PLACE,"HandleSubscrAllocColor() :: Sending packet \n");
#endif
   xtv_ACPacket.ToolID = WhatTool->ToolID;
   /*
    ** Packet is now complete so send 
    */
   MulticastData(&xtv_ACPacket,sizeof(xtv_ACPacket));
}

}



/*
**
** HandleSubscrAllocNamedColor() - Send an HS_AllocNamedColor HTOS special
**  request.Everything except tool Id has aleady been set up in
**  HandleClientAllocNamedColor.
**
**      1       HS_AllocNamedColor  Protocol ID Tag
**      1       toolID              Tool ID of sender
**      2       name length         Color name length in bytes
**      4       cmap                Colormap
**      4       pixel               pixel from remote host
**      n       Name                actual name to lookup
**
*/
static void
HandleSubscrAllocNamedColor(WhatTool,PacketStuff)
Tool *WhatTool;
BYTE * PacketStuff;
{
Subscriber * Who;
CARD16  NameLength;        /* Name length from client */

   /*
   ** need to add in new length 
   */
   int packet_len = sizeof(xtv_ANCPacket) - sizeof ( char *);
      for ( Who=FirstSubscriber(); Who != NULL; Who=NextSubscriber() ) {
         if (Who->SubscriberID == 0) {
            MulticastFlag[0]=FALSE;
            continue;
         }
#if  DEBUGHandleSubscrAllocNamedColor 
   fprintf(TRACE_PLACE,"HandleSubscrAllocNamedColor() :: Sending packet \n");
#endif
   xtv_ANCPacket.ToolID = WhatTool->ToolID;
   /* 
    ** Packet is now complete so send 
    */
   MulticastData( &xtv_ANCPacket,packet_len);
   /*
    ** And name 
    */

   NameLength = SWAP_INT(xtv_ANCPacket.nlen, WhatTool->ConnectPrefix.byteOrder);
   MulticastData( xtv_ANCPacket.name, NameLength);

   /*
    ** a little leftover nasty which needs to be cleaned up here
    */
   free(xtv_ANCPacket.name);
   return;
}
}




/*
** HandleJoinInternAtom() - Handle archive for special
*/

static void
HandleJoinInternAtom(WhatTool,Header,PacketStuff)
Tool * WhatTool;
xReq * Header;
BYTE * PacketStuff;
{
   char *ArchPacketP;
   CARD32 TotalSize;
   CARD16   HowMany;

   HowMany = SWAP_INT( *((CARD16 *)PacketStuff),
       WhatTool->ConnectPrefix.byteOrder );
   TotalSize = 9+HowMany;
   ArchPacketP = GetMem(TotalSize);
   /* onlyIfExistsfield */
/*
   ArchPacketP[0] = Header->data;  
*/
   ArchPacketP[0] = FALSE;
   ArchPacketP[1] = HS_InternalAtom;
   ArchPacketP[2] = WhatTool->ToolID;
   memcpy(ArchPacketP+3,(char *) &AtomValue,4);
   memcpy(ArchPacketP+7,PacketStuff,2);
   memcpy(ArchPacketP+9,PacketStuff+    
       (sizeof(xInternAtomReq)-sizeof(xReq)),HowMany);
   if( (WhatTool->ToolInfo->InternalAtom =
       realloc(WhatTool->ToolInfo->InternalAtom,
       WhatTool->ToolInfo->InternalAtomSize+4+TotalSize)) == NULL){
      perror("Reallocating");
      exit(1);
   }

   memcpy(WhatTool->ToolInfo->InternalAtom+WhatTool->ToolInfo->InternalAtomSize
       ,(char *) &TotalSize,4);
   memcpy(WhatTool->ToolInfo->InternalAtom+WhatTool->ToolInfo->InternalAtomSize
       +4,ArchPacketP,TotalSize);
   WhatTool->ToolInfo->InternalAtomSize += ( 4+TotalSize );
   if( WhatTool->SubDirtyCount > 0 ){
      AppendRequestToArchiveBuf(WhatTool,ArchPacketP+1,TotalSize-1);
   }
   FreeMem(ArchPacketP);
   return;

}



/*
** HandleJoinAllocColor() - Handle archive for special packet. Everything's
**  all set up so just add it to object list and archive it
*/
static void
HandleJoinAllocColor(WhatTool,Header,PacketStuff)
Tool * WhatTool;
xReq * Header;
BYTE * PacketStuff;
{
   extern ColormapObj *FindColormapByID();
   static CARD32 Size_ACP = sizeof(xtv_ACPacket);
   ColormapObj * cop = FindColormapByID(SWAP_LONG(*((CARD32 *)PacketStuff),
       WhatTool->ConnectPrefix.byteOrder),WhatTool);
#if DEBUGHandleJoinAllocColor
   fprintf(TRACE_PLACE,"HandleJoinAllocColor() : Archiving \n");
#endif
   xtv_ACPacket.ToolID = WhatTool->ToolID;
   if ( cop == NULL ) {
      fprintf(TRACE_PLACE,"HandleJoinAllocColor: Couldn't find the colormap\n");
      exit(1);
   }
   cop->AllocPackets = realloc(cop->AllocPackets,cop->AllocPacketsSize+
       sizeof(CARD32) +Size_ACP);
   memcpy(cop->AllocPackets+cop->AllocPacketsSize,(char *) &Size_ACP,sizeof(CARD32));
   memcpy(cop->AllocPackets+cop->AllocPacketsSize+sizeof(CARD32),
       &xtv_ACPacket,Size_ACP);
   cop->AllocPacketsSize += (sizeof(CARD32) + Size_ACP);

   if( WhatTool->SubDirtyCount > 0 )
      AppendRequestToArchiveBuf(WhatTool,&xtv_ACPacket,Size_ACP);
   return;
}



/*
** Archive AllocNamedColor packet and add to Colormap object list 
*/
static void
HandleJoinAllocNamedColor(WhatTool,Header,PacketStuff)
Tool * WhatTool;
xReq * Header;
BYTE * PacketStuff;
{
   extern ColormapObj *FindColormapByID();
   CARD32 Size_ANCP = sizeof(xtv_ANCPacket) - sizeof(char *);
   CARD32 namelen =  SWAP_INT(xtv_ANCPacket.nlen, 
                                  WhatTool->ConnectPrefix.byteOrder);

   CARD32 TotalSize  = Size_ANCP + namelen;   /* blockk size to archive */
   ColormapObj *cop = FindColormapByID(SWAP_LONG(*((CARD32 *)PacketStuff),
       WhatTool->ConnectPrefix.byteOrder),WhatTool);

#if DEBUGHandleJoinAllocNamedColor
   fprintf(TRACE_PLACE,"HandleJoinAllocNamedColor() : Archiving packet\n");
   fprintf(TRACE_PLACE,"HandleJoinAllocNamedColor() : namelen %d\n", namelen);
#endif

   xtv_ANCPacket.ToolID = WhatTool->ToolID;
   if ( cop == NULL ) {
      fprintf(TRACE_PLACE,"HandleJoinAllocColor: Couldn't find the colormap\n");
      exit(1);
   }
   cop->AllocPackets = realloc(cop->AllocPackets,cop->AllocPacketsSize+
       sizeof(CARD32) + TotalSize);
   memcpy(cop->AllocPackets+cop->AllocPacketsSize,(char *) &TotalSize,
       sizeof(CARD32));
   memcpy(cop->AllocPackets+cop->AllocPacketsSize + sizeof(CARD32),
       &xtv_ANCPacket,Size_ANCP);
   memcpy(cop->AllocPackets+cop->AllocPacketsSize + sizeof(CARD32) +
       Size_ANCP,xtv_ANCPacket.name,namelen);
   cop->AllocPacketsSize += (sizeof(CARD32) + TotalSize);

   if( WhatTool->SubDirtyCount > 0 )
      AppendRequestToArchiveBuf(WhatTool,&xtv_ANCPacket,Size_ANCP + namelen);
   return;

}
