/*
 * Copyright (c) 1991, 1993 Old Dominion University.
 * Copyright (c) 1991, 1993 University of North Carolina at Chapel Hill.
 * All rights reserved.
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  Old Dominion University and the University of North
 * Carolina at Chapel Hill make no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 */

/*
   tool.c
   Routines to manage the tool list

*/

#define   DEBUGAddTool      0&DEBUGON
#define   DEBUGDeleteTool      0&DEBUGON
#define   DEBUGFindTool      0&DEBUGON
#define   DEBUGFindToolByPort    0&DEBUGON
#define   DEBUGBuildToolWidget   0&DEBUGON
#define   DEBUGRebuildToolList   0&DEBUGON
#define   DEBUGInitializeTool   0&DEBUGON
#define   DEBUGToolCheck      0&DEBUGON
#define   DEBUGStartNewTool      0&DEBUGON
#define   DEBUGFryTool    0&DEBUGON
#define   DEBUGConfirmFryTool    0&DEBUGON
#define   DEBUGCleanUpTool    0&DEBUGON
#define   DEBUGStopToolGroup    0&DEBUGON
#define   DEBUGStartLocalTool    0&DEBUGON
#define   DEBUGStartRemoteTool    0&DEBUGON
#define   DEBUGStartSpecialTool    0&DEBUGON
#define   DEBUGsig_child    0&DEBUGON
#define   DEBUGJoinTool         0&DEBUGON
#define   DEBUGCleanUpToolGroup         0&DEBUGON
#define   DEBUGCleanUpAllTools         0&DEBUGON
#define   DEBUGfindSubscr         0&DEBUGON


#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#include <X11/Xproto.h>         /* Needed for X_TCP_PORT */
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/List.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/Cardinals.h>
#include <signal.h>
#include <sys/wait.h>

#include "memory.h"
#include "protocol.h"
#include "connect.h"
#include "subscriber.h"
#include "token.h"
#include "tool.h"
#include "popups.h"
#include "network.h"

/*
 * Begin Murray Addition
 */

/*
 * arpa/inet.h has the definition of inet_ntoa that was producing warnings.
 * I was getting sick of the warnings.
 */
#include <arpa/inet.h>
#include "subscriber2.h"
#include "tool2.h"
#include "tokenInfo.h"

/*
 * End Murray Addition
 */
extern SessionNumber;
extern OutputCount ;
extern HostDisplayNumber;
extern char HHostName[];
extern BYTE VERBOSE;


extern BYTE XCOLOR;
extern BYTE POPUP;
extern BYTE RESIZE;
extern BYTE SEQNUM;
extern BYTE VS;

extern BYTE JOIN;
extern int MinFloorTime;
extern TimeOutFlag;
extern BYTE POLICY;
int sig_child();
extern BYTE vs_in_out;
extern BYTE  MulticastFlag[];  /* If TRUE, include Subscriber in Multicast*/

BYTE CleanUpToolFlag = FALSE;
Bool	ToolExpected = FALSE;	/* True if we're expecting a connection from a tool */
BYTE GroupID = 0;  			/* for tool groups */
Tool	*TempTool;		/* Pointer to temporary tool structure */
int      ToolCount = 0;


/* ------------------------------------------------------------------------------------------- */

static Tool   *ToolIDList[256];   /* List of Tools by ID */

#ifndef NoIDAvailable
#define   NoIDAvailable      255
#endif

/*
 *   InitializeToolIDList()
 *
 *   Clear out the entire list of IDs and initialize the
 *   CurrentID value to 0.  Note that element 255 is reserved as
 *   a value for un-assignables.
 *
 *   RETURN VALUE:   void
 */
static void InitializeToolIDList()
{
	int   i;

	for ( i=0; i<256; i++ )
		ToolIDList[i] = NULL;
}


/*
 *   AssignToolID()
 *
 *   Hunt down an empty slot in the ID list and
 *   return its index.  This will serve as the tool's
 *   ID number throughout its life.
 *
 *   RETURN VALUE:   Element number (a byte)
 *         NoIDAvailable if the list is full.
 */
BYTE AssignToolID( Who )
Tool *Who;
{
	BYTE   i=0;

	for ( i=0; ((ToolIDList[i] != NULL) && (i <= NoIDAvailable )) ; i++ );

	if ( i < NoIDAvailable ) {
		ToolIDList[i] = Who;
	}
	return i;
}

/*
 *   SetToolID()
 *
 *   Set a tool ID based on the byte in the tool structure passed in.
 *
 *   RETURN VALUE:   void
 */
void SetToolID( Who )
Tool *Who;
{
	ToolIDList[Who->ToolID] = Who;
}


/*
 *   ClearToolID()
 *
 *   Return a Tool ID to the free pool
 *
 *   RETURN VALUE:   void
 */
void ClearToolID( Which )
BYTE Which;
{
	ToolIDList[Which] = NULL;
}

/*
 *   FindToolByID()
 *
 *   Hunt down a tool by the ToolID field and
 *   return a pointer to the structure.
 *
 *   RETURN VALUE:   Pointer to tool structure
 *         NULL if not found
 */
Tool *FindToolByID(Target)
BYTE Target;
{
	return ToolIDList[Target];
}
/* --- Begin Change ---bcl */
/*
 *   FindToolByPort()
 *
 *   Hunt down a tool by the Port number it is associated
 *   with (actually by the file descriptor we use for that Port)
 *   and return a pointer to the structure. Given ToolGroups, one
 *   must find the Leader of the tool group (not strictly true, but
 *   this routine will).
 *
 *   RETURN VALUE:  Pointer to tool structure,
 *                  NULL if not found.
 *
 */
Tool *FindToolByPort( Port )
Socket *Port;
{
  
  int   i=0;
  /* This is an integer to avoid byte rollover.
   */

#if DEBUGFindToolByPort
  fprintf( stderr, "FindToolByPort(%d): Entering. NoIDAvailable = %d.\n",
	  *Port, NoIDAvailable );
#endif
  
  for ( i=0; ( i <= NoIDAvailable ) ; i++ )
    {
#    if DEBUGFindToolByPort
      if ( ToolIDList[i] != NULL )
	fprintf( stderr, "FindToolByPort(%d): ToolIDList[%d] ?= %d.\n",
		*Port,i,ToolIDList[i]->OriginalConnection);
#    endif
      if (( ToolIDList[i] != NULL ) &&
	  (( ToolIDList[i]->OriginalConnection == *Port ) &&
	   ( ToolIDList[i]->Leader ) ) )
	{
#        if DEBUGFindToolByPort
	  fprintf( stderr, "FindToolByPort(%d): Found.\n", *Port );
#        endif
	  return ToolIDList[i];
	}
    }
  
#if DEBUGFindToolByPort
  fprintf( stderr, "FindToolByPort(%d): Not found.\n", *Port );
#endif
  return NULL;
}

/* --- End Change ---bcl */


/* ------------------------------------------------------------------------------------------- */

ToolNode   *ToolList = NULL;
static   ToolNode   *FindPointer = NULL;

static char   *EmptyList[] = {
	"None"
};


Widget      WTool;
Widget      WToolLabel;
Widget      WToolViewport;
Widget      WToolList;
Widget      WToolNew;
Widget      WToolGToken;
Widget      WToolDToken;
Widget      WToolSToken;
/*
 * Begin Murray Addition
 */
Widget      WToolIToken;
/*
 * End Murray Addition
 */
Widget      WToolJoin;
Widget      WToolTerminate;


static Arg  args[20];

extern Widget GetTopLevelWidget();   /* Added 01 */

/* ------------------------------------------------------------------------------------------- */

/*
 *   FirstTool()
 *
 *   Reset FindPointer to the first tool in the list and return 
 *   a pointer to the structure.
 *
 *   RETURN VALUE:   Pointer to first tool
 *         NULL if list is empty.
 */
Tool *FirstTool()
{
	FindPointer = ToolList;
	return (FindPointer != NULL ? FindPointer->ToolS : NULL);
}


/*
 *   NextTool()
 *
 *   Move FindPointer to the next tool in the list and return 
 *   a pointer to the structure.
 *
 *   RETURN VALUE:   Pointer to nextt tool
 *         NULL if at end of list
 */
Tool *NextTool()
{
	if (FindPointer != NULL)
		FindPointer = FindPointer->Next;

	if ( FindPointer == NULL )
		return NULL;
	else
		return FindPointer->ToolS;
}


/*
 *   AddTool()
 *
 *   Add a new tool to the list.  Insert the new entry sorted by
 *   the ToolCommand field in the structure.  Also blank out the
 *   token list field.
 *
 *   RETURN VALUE:   (void)
 */
void AddTool(NewTool)
Tool *NewTool;
{
	char      *LastWord;
	ToolNode   *NewNode;
	ToolNode   *WhereAt;

#   if DEBUGAddTool
	fprintf( stderr, "AddTool(): Adding %s Leader is %d\n", 
	    NewTool->ToolCommand, NewTool->Leader );
#   endif

	NewNode=(ToolNode *)GetMem(sizeof(ToolNode));

	NewNode->ToolS = NewTool;

	if (NewTool->Leader == TRUE)
		ToolCount++;

	/* Check to see if this is the first-ever item */

	if ( ToolList == NULL ) {
#      if DEBUGAddTool
		fprintf( stderr, "AddTool(): First entry in empty list.\n" );
#      endif
		NewNode->Next = NULL;
		ToolList = NewNode;
		if (NewTool->Leader == TRUE)
			RebuildToolList();
		return;
	}

	/* Check for head-of-list */

	if ( strcmp(NewTool->ToolCommand, ToolList->ToolS->ToolCommand ) < 0 ) {
#      if DEBUGAddTool
		fprintf( stderr, "AddTool(): Added at head, before %s\n",
		    ToolList->ToolS->ToolCommand );
#      endif
		NewNode->Next = ToolList;
		ToolList = NewNode;
		FindPointer = NULL;
		if (NewTool->Leader == TRUE)
			RebuildToolList();
		return;
	}

	/* Otherwise, find the right place for it */

	for ( WhereAt=ToolList ; WhereAt->Next != NULL ; WhereAt=WhereAt->Next ) {
#      if DEBUGAddTool
		fprintf( stderr, "AddTool(): Checking %s\n", WhereAt->Next->ToolS->ToolCommand );
#      endif
		if ( strcmp(WhereAt->Next->ToolS->ToolCommand,NewTool->ToolCommand) >= 0 ) {
#      if DEBUGAddTool
			fprintf( stderr, "AddTool(): Stopped there.\n" );
#      endif
			break;
		}
	}

	NewNode->Next = WhereAt->Next;
	WhereAt->Next = NewNode;
	if (NewTool->Leader == TRUE)
		RebuildToolList();
}


/*
 *   DeleteTool()
 *
 *   Hunt down a Tool and kill it.  If the entry does
 *   not exist in the list, scrub it.
 *
 *   RETURN VALUE:   void
 */
void DeleteTool(Target)
Tool *Target;
{
	ToolNode   *WhereAt;
	ToolNode   *Temp;

#   if DEBUGDeleteTool
	fprintf( stderr, "DeleteTool(): Removing %s\n", Target->ToolCommand );
#   endif

	ClearToolID( Target->ToolID );


#   if DEBUGDeleteTool
	fprintf( stderr, "DeleteTool(): Looking at First Node \n");
	fprintf( stderr, "Contains %s\n", ToolList->ToolS->ToolCommand);
#   endif

	if ( Target == ToolList->ToolS ) {
#      if DEBUGDeleteTool
		fprintf( stderr, "DeleteTool(): First in list\n" );
#      endif
		Temp = ToolList;
		ToolList = ToolList->Next;
		Temp->Next=NULL;
		if( GetMode() == TRUE && Target->join ){
			FreeMem( Target->ArchiveBuf );
			FreeMem( Target->Archive2Buf );
			FreeToolInfo( Target->ToolInfo );
		}
		FreeMem(Target);
		FreeMem(Temp);

		if (Target->Leader){
			ToolCount--;
			RebuildToolList();
		}
		return;
	}
#   if DEBUGDeleteTool
	fprintf( stderr, "DeleteTool(): Looking after First Node\n");
#   endif


	for ( Temp = ToolList, WhereAt=ToolList->Next ; WhereAt!= NULL ; Temp=WhereAt, WhereAt=WhereAt->Next ) {
#      if DEBUGDeleteTool
		fprintf( stderr, "DeleteTool(): Looking at %s\n",
		    WhereAt->ToolS->ToolCommand );
#      endif
		if ( WhereAt->ToolS == Target ) {
#         if DEBUGDeleteTool
			fprintf( stderr, "DeleteTool(): Found!\n" );
#         endif
			break;
		}
	}

	/* See if we didn't find it */

	if ( WhereAt == NULL ) {
#      if DEBUGDeleteTool
		fprintf( stderr, "DeleteTool(): Not Found\n" );
#      endif
		return;
	}

	/* Fry the sucker! */

	Temp->Next = WhereAt->Next;
	WhereAt->Next=NULL;
	if( GetMode() == TRUE && WhereAt->ToolS->join ){
		FreeMem( WhereAt->ToolS->ArchiveBuf );
		FreeMem( WhereAt->ToolS->Archive2Buf );
		FreeToolInfo( WhereAt->ToolS->ToolInfo );
	}
	FreeMem( WhereAt->ToolS );
	FreeMem( WhereAt);

	if (Target->Leader){
		ToolCount--;
		RebuildToolList();
	}

}


/*
 *   FindTool()
 *
 *   Hunt down a tool by the ToolCommand field and
 *   return a pointer to the structure.
 *
 *   RETURN VALUE:   Pointer to tool structure
 *         NULL if not found
 */
Tool *FindTool(Target)
char *Target;
{
	ToolNode   *WhereAt;
	ToolNode   *Temp;

	for ( WhereAt=ToolList ; WhereAt != NULL ; WhereAt=WhereAt->Next ) {
#      if DEBUGFindTool
		fprintf( stderr, "FindTool(): Looking at %s\n",
		    WhereAt->ToolS->ToolCommand );
#      endif
		if ( !strcmp( WhereAt->ToolS->ToolCommand, Target ) ) {
#         if DEBUGFindTool
			fprintf( stderr, "FindTool(): Found!\n" );
#         endif
			break;
		}
	}

	return ( WhereAt ? WhereAt->ToolS : NULL );
}


/* ------------------------------------------------------------------------------------------- */

Socket         ToolSocket;
int         DisplayNumber;
struct sockaddr_in   ListenerName;

/*
 *   RebuildToolList()
 *
 *   Scoot through the tool list and build a list and build an array of pointers
 *   to the FullName field.
 *
 *   RETURN VALUE:   void
 */
void RebuildToolList()
{
	static char   **TList = NULL;
	Tool      *WhereAt;
	int      n;
	int      Counter = 0;

	/* Handle an empty list by shutting everything off */

	if ( (ToolList == NULL) || (ToolCount == 0) ) {
		if (TList != NULL)
			FreeMem(TList);
		n=0;
		XtSetArg( args[n],    XtNsensitive,      FALSE );
		n++;
		XtSetArg( args[n],    XtNlist,       EmptyList );
		n++;
		XtSetArg( args[n],    XtNnumberStrings,   1 );
		n++;
		XtSetValues( WToolList, args, n );

		n=1;
		XtSetArg( args[n], XtNwidth, 300);
		n++;
#ifdef rs6000
		XtSetArg( args[n], XtNheight, 40);
#else
		XtSetArg( args[n], XtNheight, 36);
#endif
		n++;
		XtSetValues( WToolViewport, args, n);
		n--;
		n--;

		XtSetValues( WToolGToken, args, 1 );
		XtSetValues( WToolDToken, args, 1 );
		XtSetValues( WToolSToken, args, 1 );
		/*
 * Begin Murray Addition
 */
		XtSetValues( WToolIToken, args, 1 );
		/*
 * End Murray Addition
 */
		XtSetValues( WToolTerminate, args, 1 );
		return;
	}

	/* Build the list */

#   if DEBUGRebuildToolList
	fprintf( stderr, "RebuildToolList(): %d items at %d bytes each\n",
	    ToolCount, sizeof(char *) );
#   endif

	TList=(char **)GetMem( ToolCount * sizeof(char *) );

	for ( WhereAt=FirstTool(); WhereAt != NULL; WhereAt=NextTool() ) {
#      if DEBUGRebuildToolList
		fprintf( stderr, "RebuildToolList(): Found %s\n", WhereAt->ToolCommand );
#   endif
		/* xyzzy - Extract actual command, not the whole mess */
		if (WhereAt->Leader == TRUE)
			TList[Counter++] = WhereAt->ToolCommand;
	}

	/*
 * For some reason, the resizing was messing up and clipping the viewport
 * So leave it alone and never resize.
 */

	XawListChange( WToolList, TList, ToolCount, 0, FALSE );


	n=0;
	XtSetArg( args[n],    XtNsensitive,      TRUE );
	n++;
	/*
 * Begin Murray Addition
 */
	XtSetValues( WToolViewport, args, n);
	/*
 * End Murray Addition
 */
	XtSetValues( WToolGToken, args, n );
	XtSetValues( WToolDToken, args, n );
	XtSetValues( WToolSToken, args, n );
	/*
 * Begin Murray Addition
 */
	XtSetValues( WToolIToken, args, n );
	/*
 * End Murray Addition
 */
	XtSetValues( WToolTerminate, args, n );
	XtSetValues( WToolList, args, n );

	n=1;
	XtSetArg( args[n], XtNwidth, 300);
	n++;
#ifdef rs6000
	XtSetArg( args[n], XtNheight, 40);
#else
	XtSetArg( args[n], XtNheight, 36);
#endif
	n++;
	XtSetValues( WToolViewport, args, n);
#      if DEBUGRebuildToolList
	fprintf( stderr, "RebuildToolList(): Finished\n");
#   endif


}

/* ------------------------------------------------------------------------------------------- */

/* Start Added 01 */
XtCallbackProc CancelDone(w, client_data, call_data)
Widget   w;
caddr_t client_data, call_data;
{
	XtPopdown(TranShellNewTool);
	XtSetSensitive(WTool, TRUE);
}

/*-------------------------------------------------------------------*/



/*
 *
 * DialogDoneHost()
 *
 * A call back when Host Button is pressed.
 *
 */
XtCallbackProc DialogDoneHost(w, dialog, call_data)
Widget   w;
Widget   dialog;   /* client_data */
caddr_t call_data;
{
	String      DCommand;
	static String   DefaultString = "";
	char      Command[TOOL_CMD_SIZE];
        char      Options[TOOL_OPT_SIZE];
	Arg     arg;

	XtPopdown(TranShellNewTool);
	XtSetSensitive(WTool, TRUE);
	DCommand = XawDialogGetValueString(dialog);
	Command[strlen(DCommand)-1]=NULL;
	strcpy(Command, DCommand);

	XtSetArg(arg, XtNvalue, DefaultString);
	XtSetValues(dialog, &arg, 1);

        strcpy(Options, "");
        if(Option.j)
          strcat(Options, " -J ");        
        if(Option.z)
          strcat(Options, " -Z ");        
        if(Option.c)
          strcat(Options, " -C ");        
        if(Option.o)
          strcat(Options, " -O 0 ");       
        if(Option.p)
          strcat(Options, " -P ");        
        if(Option.v)
          strcat(Options, " -V ");        
        if(Option.q)
          strcat(Options, " -Q ");        

	if (strlen(Command) <1) {
		UpdateMessage("\nNull Entry for tool !!\n\n");
		return;
	}


	if (GetMode()) {
		StartLocalTool(Command, Options, GetLocalSubID());
	} else { /*Send the tool create message to the chairman */
		if ( SendByte(GetHostConnection(), SH_NewToolLocal )==ERROR){
			EndSession();
			return;
		}
		if ( SendByte(GetHostConnection(), vs_in_out )==ERROR){
			EndSession();
			return;
		}
		if ( SendData(GetHostConnection(), (BYTE *)Command, TOOL_CMD_SIZE)==ERROR) {
			EndSession();
			return;
		}
		if ( SendData(GetHostConnection(), (BYTE *)Options, TOOL_OPT_SIZE)==ERROR) {
			EndSession();
			return;
		}
	}
}


/* ------------------------------------------------------------------------------------------- */
/*
 *   StartLocalTool()
 *
 *   Start up a new tool localy.
 *
 */
void StartLocalTool(command, options, creator)
char *command;
char *options;
BYTE creator;
{
	char            Command[TOOL_CMD_SIZE];
	static char     NewDisplay[HOST_NAME_SIZE];
	char            HostName[HOST_NAME_SIZE];
	char		*HostInet;
	char            *CommandArgs[10];
	char            CommandName[TOOL_CMD_SIZE];
	struct hostent *hp, *gethostbyname();
	struct sockaddr_in server;
	Socket          creconnection;

	char SesNum[5];
#   if DEBUGStartLocalTool
	fprintf( stderr, "StartLocalTool(): starting %s\n",
	    command);
#   endif

	strcpy(Command, command);
	/*
	strcat(Command, " -title "); 
	LGetName(CommandName,Command);
	strcat (Command, CommandName);    

	strcat(Command, "/Session#");    
	sprintf(SesNum,"%d",SessionNumber); 
	strcat(Command, SesNum);           
*/
	/* This can't work now, since all tools are created by chairman
   if (GetMode())
	strcat(Command,"/Chairman");
   else
	strcat(Command,"/Participant");
*/
	if (strlen(Command) <1) {
		UpdateMessage("\nNull Entry for tool !!\n\n");
		return;
	}
	TempTool=(Tool *)GetMem(sizeof(Tool));
	LGetArgs(CommandArgs,Command);
	if ( gethostname( HostName, HOST_NAME_SIZE ) < 0 ) {
#   if DEBUGStartNewTool
		fprintf( stderr, "StartLocalTool(): Can't find host Name\n" );
#   endif
		FatalError( "StartLocalTool(): Unable to get host name" );
		return;

	}

	if  ( (hp = gethostbyname(HostName)) == NULL ) {
#   if DEBUGStartNewTool
		fprintf( stderr, "StartLocalTool(): Can't find host %s\n", HostName);
#   endif
		FatalError("StartLocalTool(): Can't find entry for host %s\n", HostName);
		return;
	}

	bcopy ( hp->h_addr, &(server.sin_addr), hp->h_length);
	HostInet = inet_ntoa(server.sin_addr);
	sprintf( NewDisplay, "DISPLAY=%s:%d.0", HostInet, DisplayNumber );
	strcpy(CommandName, CommandArgs[0]);
	strcpy(TempTool->ToolCommand, CommandName);
        
	InitToolStruct(options, creator);

#   if DEBUGStartLocalTool
	fprintf( stderr, "StartLocalTool(): CreatorID %d\n", TempTool->ToolCreator);
	fprintf( stderr, "StartLocalTool(): Forknig %s\n", TempTool->ToolCommand);
	fprintf( stderr, "StartLocalTool(): NewDisplay %s\n", NewDisplay);
#   endif
	siginterrupt(SIGCHLD, 1);
	signal(SIGCHLD, sig_child);
	if ( (TempTool->ToolPID = fork()) == 0 ){
		/*
    	int fd;
    	if ((fd = open ("/dev/tty", O_RDWR)) >= 0){
		ioctl(fd, TIOCNOTTY, (char *) NULL);
		close(fd);
    	}
    	for (fd =0; fd < NOFILE; fd++) close (fd);
      	setpgrp(0,getpid());
*/
		putenv( NewDisplay );

		execvp (CommandName,&CommandArgs[0]);
		fprintf(stderr, "\nStartLocalTool(): Can't execute %s %s\n", CommandName,TempTool->ToolCommand);
		UpdateMessage( "\nCan't execute Tool\n");
		/*
                        UpdateMessage( "\nStartLocalTool(): Can't execute %s %s\n", CommandName,TempTool->ToolCommand);
*/
		exit(0);
	}
#   if DEBUGStartLocalTool
	fprintf( stderr, "StartLocalTool(): PID for %s : %d \n", CommandName,
	    TempTool->ToolPID );
#   endif

}

/* ------------------------------------------------------------------------------------------- */
/*
 *   StartRemoteTool()
 *
 *   Start up a new tool remotely.
 *
 */
void StartRemoteTool(tool, options, creator)
char *tool;
char *options;
BYTE creator;
{
	char      Command[TOOL_CMD_SIZE];
	static char     NewDisplay[HOST_NAME_SIZE];
	char            HostName[HOST_NAME_SIZE];
	char            *CommandArgs[10];
	char            CommandName[TOOL_CMD_SIZE];

	strcpy(Command, tool);
#   if DEBUGStartRemoteTool
	fprintf( stderr, "StartRemoteTool(): starting %s\n",
	    tool);
#   endif

	TempTool=(Tool *)GetMem(sizeof(Tool));
	LGetArgs(CommandArgs,Command);
	strcpy(CommandName, CommandArgs[0]);
	strcpy(TempTool->ToolCommand, CommandName);

	InitToolStruct(options, creator);

#   if DEBUGStartRemoteTool
	fprintf( stderr, "StartRemoteTool(): CreatorID %d\n", TempTool->ToolCreator);
	fprintf( stderr, "StartRemoteTool(): started %s\n",
	    TempTool->ToolCommand);
	fprintf( stderr, "StartRemoteTool(): ToolGID is %d\n", TempTool->ToolGID);
#   endif
}


/*
 *
 * DefaultHost()
 *
 * An action proc called when Return is pressed
 *
 */

/* begin Ramesh's modification for VS-XTV */

/* remove this function, now it is in the program popups.c with */
/* a different meaning */

/*
XtCallbackProc DefaultHost(w, event)
Widget w;
XButtonEvent *event;
{


   DialogDoneHost(w, NewToolDialog, NULL);

}
*/

/* end VS-XTV modification */
/* ------------------------------------------------------------------------------------- */

/*
 *
 *
 */

XtCallbackProc DialogDoneRemote(w, dialog, call_data)
Widget  w;
Widget dialog;
caddr_t call_data;
{
	String      DCommand;
	char        Command[TOOL_CMD_SIZE];
	char        Options[TOOL_OPT_SIZE];

	static String   DefaultString = "";
	Arg     arg;
	static char     NewDisplay[HOST_NAME_SIZE];
	char            HostName[HOST_NAME_SIZE];
	char           *HostInet;
	char           *CommandArgs[10];
	char            CommandName[TOOL_CMD_SIZE];
	struct hostent *hp, *gethostbyname();
	struct sockaddr_in server;



	XtPopdown(TranShellNewTool);
	XtSetSensitive(WTool, TRUE);
	DCommand = XawDialogGetValueString(dialog);
	Command[strlen(DCommand)-1]=NULL;
	strcpy(Command, DCommand);
	XtSetArg(arg, XtNvalue, DefaultString);
	XtSetValues(dialog, &arg, 1);

        strcpy(Options, "");
        if(Option.j)
          strcat(Options, " -J ");        
        if(Option.z)
          strcat(Options, " -Z ");        
        if(Option.c)
          strcat(Options, " -C ");        
        if(Option.o)
          strcat(Options, " -O 0 ");       
        if(Option.p)
          strcat(Options, " -P ");        
        if(Option.v)
          strcat(Options, " -V ");        
        if(Option.q)
          strcat(Options, " -Q ");        


	if (strlen(Command) <1) {
		UpdateMessage("\nNull Entry for tool !!\n\n");
		return;
	}

	if (GetMode() == TRUE) {
		if ( gethostname( HostName, HOST_NAME_SIZE ) < 0 ) {
#   if DEBUGStartNewTool
			fprintf( stderr, "StartLocalTool(): Can't find host Name\n" );
#   endif
			FatalError( "StartLocalTool(): Unable to get host name" );
			return;
		}
		StartRemoteTool(Command, Options, GetLocalSubID());
	}
	else {
#   if DEBUGStartNewTool
		printf(stderr,"DialogDoneRemote():HHostName %s\n", HHostName);
#   endif
		strcpy(HostName, HHostName);
		DisplayNumber = HostDisplayNumber;
		if ( SendByte(GetHostConnection(), SH_NewToolRemote )==ERROR){
			EndSession();
			return;
		}
		if ( SendByte(GetHostConnection(), vs_in_out )==ERROR){

			EndSession();
			return;
		}
		if ( SendData(GetHostConnection(), (BYTE *)Command, TOOL_CMD_SIZE)==ERROR) {
			EndSession();
			return;
		}
		if ( SendData(GetHostConnection(), (BYTE *)Options, TOOL_OPT_SIZE)==ERROR) {
			EndSession();
			return;
		}
#   if DEBUGDialogDoneRemote
		fprintf( stderr,
		    "DialogDoneRemote(): Sent  SH_NewToolRemote %s to start %s\n",
		    HostName, Command);
#   endif

	}

	if  ( (hp = gethostbyname(HostName)) == NULL ) {
#   if DEBUGStartNewTool
		fprintf( stderr, "StartNewTool(): Can't find host %s\n", HostName);
#   endif
		FatalError("StartNewTool(): Can't find entry for host %s\n", HostName);
		return;
	}

	bcopy ( hp->h_addr, &(server.sin_addr), hp->h_length);
	HostInet = inet_ntoa(server.sin_addr);

	UpdateMessage("Login to remote host and then Type:     \n setenv DISPLAY  %s:%d.0\n OR setenv DISPLAY  %s:%d.0     \n then run X client application:     \n (%s)\n ",
	    HostName, DisplayNumber, HostInet, DisplayNumber, Command);
	return;
}


/* ------------------------------------------------------------------------------------------- */
/*
 *   StartNewTool()
 *
 *   Start up a new tool.
 *
 *   RETURN VALUE:   rv
 */
XtCallbackProc StartNewTool( widget, TopLevel, callData )
Widget widget;
Widget TopLevel;
XtPointer callData;
{
	XtSetSensitive(WTool, FALSE);
	PopUpBox ( widget, TranShellNewTool, callData );
}

static BYTE id[100];
static BYTE *name[100];
static int JoinToolNum;
static Widget TmpWidget;
static XtPointer TmpCallData;

/*
 *   JoinTool()
 *
 *   Join a tool.
 *
 *   RETURN VALUE: rv
 */
XtCallbackProc JoinTool( widget, TopLevel, callData )
Widget widget;
Widget TopLevel;
XtPointer callData;
{
	TmpWidget = widget;
	TmpCallData = callData;
	if( SendByte(GetHostConnection(), SH_WantJoinTools)==ERROR)
		EndSession();
}

HandleToolsToJoinList()
{
	int n;
	int i;

	XtSetSensitive(WTool, FALSE);

	i = 0;
	while(1){
		id[i] = GetByte(GetHostConnection());
		if( id[i] == 0xff ) break;
		name[i] = ReadData(GetHostConnection(),100,BLOCK);
#       if DEBUGJoinTool
		fprintf(stderr," HandleToolsToJoinList: got %d %s\n",id[i],name[i]);
#       endif
		i++;
	}
	JoinToolNum = i;
	if( i == 0 ){
		XtSetSensitive(WTool, TRUE);
		UpdateMessage("No tool available to join");
		return;
	}

	XawListChange( JoinToolList, name, i, 0, FALSE );

	n=0;
	XtSetArg( args[n], XtNsensitive, TRUE );
	n++;
	XtSetValues( JoinToolList, args, n);

	XtSetArg( args[n], XtNwidth, 150);
	n++;
	XtSetArg( args[n], XtNheight, 50);
	n++;
	XtSetValues( JoinToolViewPort, args, n);

	PopUpBox ( TmpWidget, TranShellJoinTool, TmpCallData );
}

/*
 * DialogDoneJoinAll()
 *
 * A call back when Join_All Button is pressed
 *
 */
XtCallbackProc DialogDoneJoinAll(w, dialog, call_data)
Widget   w;
Widget   dialog;   /* client_data */
caddr_t call_data;
{
	int i;

	XtPopdown(TranShellJoinTool);
	XtSetSensitive(WTool, TRUE);

	SendByte(GetHostConnection(), SH_JoinTools);
#       if DEBUGJoinTool
	fprintf(stderr, "DialogDoneJoinAll: sent SH_JoinTools\n");
#       endif

	for( i=0; i<JoinToolNum; i++)
		SendByte(GetHostConnection(), id[i]);
	SendByte(GetHostConnection(),0xff);
}

/*
 *
 * DialogDoneJoin()
 *
 * A call back when Join Button is pressed
 *
 */
XtCallbackProc DialogDoneJoin(w, dialog, call_data)
Widget   w;
Widget   dialog;   /* client_data */
caddr_t call_data;
{
	XawListReturnStruct     *Selected;
	int i;

	XtPopdown(TranShellJoinTool);
	XtSetSensitive(WTool, TRUE);

	Selected = XawListShowCurrent( JoinToolList );

	/* Bail out of nothing's highlighted */
	if( Selected->list_index == XAW_LIST_NONE ){
		UpdateMessage( "No tool selected!");
		return;
	}

	for( i=0; i<JoinToolNum && strcmp(name[i],Selected->string); i++);

	if( JoinToolNum == i ){
		fprintf(stderr, "Tool list for joining screwed up!\n");
		exit(1);
	}

	SendByte(GetHostConnection(), SH_JoinTools);
	SendByte(GetHostConnection(), id[i]);
	SendByte(GetHostConnection(), 0xff);
#       if DEBUGJoinTool
	fprintf(stderr, "DialogDoneJoin: sent SH_JoinTools\n");
#       endif

}

XtCallbackProc JoinCancelDone(w, client_data, call_data)
Widget   w;
caddr_t client_data, call_data;
{
	XtPopdown(TranShellJoinTool);
	XtSetSensitive(WTool, TRUE);
}

Tool *FriedTool;

/* ------------------------------------------------------------------------------------------- */

/*
 *      FryTool
 *
 *      Send a subscriber a HS_DeleteTool and close it out
 *
 *      RETURN VALUE:   void
 */
XtCallbackProc FryTool( widget, clientData, callData )
Widget widget, clientData;
XtPointer callData;
{
	XawListReturnStruct     *Selected;
	Tool              *What;



	Selected = XawListShowCurrent( WToolList );

	/* Bail out of nothing's highlighted */
	if ( Selected->list_index == XAW_LIST_NONE ) {
		UpdateMessage( "No Tool is selected." );
#               if DEBUGFryTool
		fprintf( stderr, "FryTool(): NoTool selected.  Returning.\n" );
#               endif
		return;
	}

#       if DEBUGFryTool
	fprintf( stderr, "FryTool(): Cooking %s\n", Selected->string );
#       endif

	if ( (What=FindTool(Selected->string)) == NULL ) {
		FatalError( "FryTool(): ERROR: List widget inconsistent with tool list" );
#       if DEBUGFryTool
		fprintf( stderr, "FryTool(): Can't find info about %s\n", Selected->string );
#       endif
		return;
	}

	GetToolLeader(What);
	/* allow a participnat to stop a tool */
	/*
        if (!GetMode() && What->ToolCreator != GetLocalSubID()) {
                UpdateMessage("Not the owner for this tool\n");
		return;
	}
*/

#       if DEBUGFryTool
	fprintf( stderr,
	    "FryTool(): Killing %s whose PID is: %d and whose ToolID is: %d and ToolGID is %d\n",
	    What->ToolCommand, What->ToolPID, What->ToolID, What->ToolGID);
#        endif

	FriedTool = What;

	XtSetSensitive(WTool, FALSE);
	PopUpBox ( widget, TranShell[TermTool], callData );
}

XtCallbackProc CancelFryTool (w, client_data, call_data)
Widget   w;
caddr_t client_data, call_data;

{
	XtPopdown(TranShell[TermTool]);
	XtSetSensitive(WTool, TRUE);

}

XtCallbackProc ConfirmFryTool (w, client_data, call_data)
Widget   w;
caddr_t client_data, call_data;

{
	XtPopdown(TranShell[TermTool]);
	XtSetSensitive(WTool, TRUE);


	if (GetMode() == TRUE) {
# if DEBUGConfirmFryTool
		fprintf( stderr, "ConfirmFryTool(): calling CleanUpToolGroup for tool %d\n", FriedTool->ToolPID);
# endif
		/*
    	CleanUpTool(FriedTool);
*/
		CleanUpToolGroup(FriedTool);
	}
	else {
#       if DEBUGConfirmFryTool
		fprintf( stderr, "ConfirmFryTool(): ToolCreator: %d, GetLocalSubID():%d\n ", 
		    FriedTool->ToolCreator, GetLocalSubID());
#       endif
		/* Allow participnats to stop tools */
		/*
      	if (FriedTool->ToolCreator == GetLocalSubID()) {
*/
		/* tool creator can kill tool */
		/*
		if ( SendByte(GetHostConnection(), SH_StopTool)==ERROR) {
			EndSession();
			return;
		}
		if ( SendByte(GetHostConnection(), FriedTool->ToolID)==ERROR){
			EndSession();
			return;
		}
*/
		StopToolGroup(FriedTool);
		/*
		CleanUpTool(FriedTool);
*/
		CleanUpToolGroup(FriedTool);

		/*
      	} else 
         	UpdateMessage("Not the owner for this tool\n");
*/
	}
}

/* ------------------------------------------------------------------------------------------- */

void CleanUpTool(What)
Tool *What;
{

	BYTE                    DeadID;
	Subscriber              *Who;
	/* Cook it, but save a copy of its ID number */
	CleanUpToolFlag = TRUE;
	DeadID = What->ToolID;
#       if DEBUGCleanUpTool
	fprintf( stderr,
	    "CleanUpTool(): Killing %s whose PID is: %d and whose ToolID is: %d and ToolGID is %d\n",
	    What->ToolCommand, What->ToolPID, DeadID, What->ToolGID);
#       endif
	if (GetMode() == TRUE)
	{
		if (POLICY == 1) XtRemoveTimeOut(What->floor_timer);
		XtRemoveInput(What->InID);
		XtRemoveInput(What->InExceptID);
		shutdown( What->InConnection, 2 );
#       if DEBUGCleanUpTool
		fprintf( stderr, "CleanUpTool(): shutdown Tool socket\n");
#       endif
	}

	shutdown( What->OutConnection,2 );
	XtRemoveInput(What->OutID);
	XtRemoveInput(What->OutExceptID);
#       if DEBUGCleanUpTool
	fprintf( stderr, "CleanUpTool(): shutdown Server socket\n");
#       endif


	/*
	if( GetMode() != TRUE || What->join == 0 || What->SubDirtyCount == 0 )
*/

	DeleteTool( What );

	/* Send everybody the bad news. */

	if (What->Leader == TRUE) {
		if (TimeOutFlag == FALSE)
			UpdateMessage( "Tool %s has been terminated.", What->ToolCommand );
		else
			UpdateMessage( "You have been time out the Tool %s.",What->ToolCommand );
	}

	if (GetMode() == TRUE) {
		for ( Who=FirstSubscriber(); Who != NULL; Who=NextSubscriber() ) {
			if (Who->SubscriberID == 0) {
				MulticastFlag[0]=FALSE;
				continue;
			}
			if( What->join && ArchiveToBeSent(Who,What) ){
				MulticastFlag[Who->SubscriberID]=FALSE;
				continue;
			}
		}
		MulticastByte(HS_DeleteTool);
		MulticastByte(DeadID );
		{
			int  i;
			for ( i=0; i<256; i++ )
				MulticastFlag[i] = TRUE;
		}


		{
			char buf[2];

			What->ThisToolDead = 1;
			if( What->join && What->SubDirtyCount > 0 ){
				buf[0] = HS_DeleteTool;
				buf[1] = DeadID;
				AppendRequestToArchiveBuf(What,buf,2);
			}
		}
		if (What->ToolPID != 0) {
			kill(What->ToolPID,SIGKILL);
#       if DEBUGCleanUpTool
			fprintf( stderr, 
			    "CleanUpTool(): Killed %s whose PID is: %d and whose ToolID is: %d\n", 
			    What->ToolCommand, What->ToolPID, DeadID);
#       endif
		}
		if (What->Leader) {

			findTool(FindSubscriberByID(HOST_ID), What);
			if (!offrightTool(FindSubscriberByID(HOST_ID)))
				deleteTool(FindSubscriberByID(HOST_ID));
			else
				fprintf(stderr, "Unable to remove tool %s from HOST.\n\r", What->ToolCommand);

			for (startSubscr(What); !offrightSubscr(What); forthSubscr(What))
			{
				findTool(getSubscr(What), What);
				if (!offrightTool(getSubscr(What)))
					deleteTool(getSubscr(What));
			}


			deleteAllSubscr(What);

if (What->verbose)
			fprintf(stderr, "Done deleting %s.\n\r", What->ToolCommand);
		}

	}
	CleanUpToolFlag = FALSE;
}


/*
 *   ConnectAndRefuse()
 *
 *   Fake a connection with an X client and refuse to accept it
 *
 *   RETURN VALUE:   void
 */
void ConnectAndRefuse( Port )
Socket Port;
{
	xConnClientPrefix   *Prefix;
	BYTE         *Leftovers;
	xConnSetupPrefix   SPrefix;

	/* xyzzy - Do nothing for now... Fake a refused connection later */
}

/* ------------------------------------------------------------------------------------------- */

/*
 *   InitializeTool()
 *
 *   Initialize all work processes related to tools and
 *   build the widgets
 *
 *   RETURN VALUE:   void
 */
void InitializeTool( AppCon, Parent, Mode )
XtAppContext AppCon;
Widget Parent;
Bool Mode;      /* TRUE if in 'host' mode */
{
	int      n;

	InitializeNewTool( AppCon, Parent, Mode );
	InitializeJoinTool( AppCon, Parent );

	/* Hunt down an unused port above 6010 */
	if ( GetMode() ) {

		DisplayNumber = XTV_LOW_DISPLAY;
		do {
			ToolSocket = OpenListener( htons(X_TCP_PORT+DisplayNumber), 4, &ListenerName );
		} while ( (ToolSocket < 0) && ( DisplayNumber++ <= XTV_HIGH_DISPLAY ) );

		if (DisplayNumber > XTV_HIGH_DISPLAY) {
			FatalError("Unable to find an available tool port.");
#   if DEBUGInitializeTool
			fprintf( stderr, "InitializeTool(): Unable to allocate a port, current valus is: display port %d\n",
			    DisplayNumber );
#   endif
			return;
		}

		XtAppAddInput( GetAppCon(), ToolSocket, XtInputReadMask, ToolCheck, NULL );

#   if DEBUGInitializeTool
		fprintf( stderr, "InitializeTool(): Tool port open at display port %d\n", DisplayNumber );
#   endif

		fcntl( ToolSocket, F_SETFL, FNDELAY );
	}

	WTool = XtCreateManagedWidget( "ToolPanel", formWidgetClass, Parent, NULL, 0 );
	WToolLabel = XtCreateManagedWidget( "Label", labelWidgetClass, WTool, NULL, 0);
	WToolViewport = XtCreateManagedWidget( "VPort", viewportWidgetClass, WTool, NULL, 0 );

	WToolList = XtCreateManagedWidget( "ToolList", listWidgetClass, WToolViewport, NULL, 0 );
	XawListChange( WToolList, EmptyList, 1, 0, TRUE );

	WToolGToken = XtCreateManagedWidget( "GToken", commandWidgetClass, WTool, NULL, 0 );
	XtAddCallback( WToolGToken, XtNcallback, GetToken, NULL );

	WToolDToken = XtCreateManagedWidget( "DToken", commandWidgetClass, WTool, NULL, 0 );
	XtAddCallback( WToolDToken, XtNcallback, DropToken, NULL );

	WToolSToken = XtCreateManagedWidget( "SToken", commandWidgetClass, WTool, NULL, 0 );
	XtAddCallback( WToolSToken, XtNcallback, SnatchToken, NULL );

	WToolNew = XtCreateManagedWidget( "NewTool", commandWidgetClass, WTool, NULL, 0 );
	XtAddCallback( WToolNew, XtNcallback, StartNewTool, TranShellNewTool); /* Added 01 */

	/*
 * Begin Murray Addition
 */
	WToolIToken = XtCreateManagedWidget( "IToken", commandWidgetClass, WTool, NULL, 0 );
	XtAddCallback( WToolIToken, XtNcallback, tokenInfo, NULL );
	/*
 * End Murray Addition
 */

	WToolJoin = XtCreateManagedWidget( "Join", commandWidgetClass, WTool, NULL, 0 );
	XtAddCallback( WToolJoin, XtNcallback, JoinTool, TranShellJoinTool);

	WToolTerminate = XtCreateManagedWidget( "TermTool", commandWidgetClass, WTool, NULL, 0 );
	XtAddCallback( WToolTerminate, XtNcallback, FryTool, TranShell[TermTool]);

#   if DEBUGInitializeTool
	fprintf( stderr, "InitializeTool(): Done.\n" );
#   endif
}
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*
 *   ToolCheck();
 *
 *   Check for a new Tool and handle all of the formalities of connection
 *
 * Possible combinations of input conditions and what to do about them:
 * (Port==ToolSocket implies this is a connection on a port not yet
 * associated with a running tool.)
 *                Port==ToolSocket             Port!=ToolSocket
 *  ToolExpected  accept; update ToolSocket    accept to old tool (race)
 * !ToolExpected  refuse; ERROR                accept to old tool
 *
 *   RETURN VALUE:   rv
 */
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/

XtInputCallbackProc ToolCheck(   XWho, Port, ID )
Subscriber *XWho;
Socket *Port;
XtInputId *ID;

{
  Tool                *NewTool, *Leader;
  struct sockaddr_in   NewToolAddr;
  Socket               NewSocket;
  int                  n;
  Subscriber          *Who;
  CARD16               tmpHowMuchAuthData,
                       tmpNumVisualIDs;
  /* --- Begin Change ---bcl */
  Bool                 NeedNewPort = FALSE; 
  /* --- End Change ---bcl */

#if DEBUGToolCheck
  fprintf( stderr, "ToolCheck(): Entering.\n" );
#endif
  
  XtRemoveInput( *ID );
  XtAppAddInput( GetAppCon(), *Port, XtInputReadMask, ToolCheck, XWho );
  
  n = sizeof(struct sockaddr_in);
  if ((NewSocket=accept( *Port, &NewToolAddr, &n )) < 0 ) {
#  if DEBUGToolCheck 
    fprintf( stderr, "ToolCheck(): No connection\n" );
#  endif
    return FALSE;
  }

#if DEBUGToolCheck
  fprintf( stderr, "ToolCheck(): Got connection\n" );
#endif

  if ( !ToolExpected ) {
/*---------------------------------------------------------------------------
#  if DEBUGToolCheck
    fprintf( stderr, "ToolCheck(): Not waiting for a tool.  Refusing.\n" );
#  endif
    ConnectAndRefuse( NewSocket );
    close( NewSocket );
    return FALSE;
-----------------------------------------------------------------------------*/
    /* tool is not expected, treat this as new a tool in every aspect.
     * This will be fixed later to make it better
     * IBM saettle have pointed out to this problem before
     */
#  if DEBUGToolCheck
    fprintf( stderr, "ToolCheck():  Tool NOT expected.  Connecting.\n" );
#  endif
    /* --- Begin Changes ---bcl */
    /* Tool is NOT expected; need to find what tool connection is associated
     * with and then just "start" the tool. -bcl
     */
    Leader = FindToolByPort(Port);
    if (Leader == NULL)
      {
#      if DEBUGToolCheck
	fprintf(stderr,"ToolCheck(): Refusing, No ToolGroup leader found.\n");
#      endif
	ConnectAndRefuse( NewSocket );
	close( NewSocket );
	return FALSE;
      }
    StartSpecialTool(Leader);
  }
  else {
    /* New tool is expected; need to make sure the connection is not for an
     * old tool and just happened to come at the wrong moment. -bcl
     */
#  if DEBUGToolCheck
    fprintf( stderr, "ToolCheck():  Tool expected.\n" );
#  endif
    Leader = FindToolByPort(Port);
    if (Leader != NULL)
      {
#      if DEBUGToolCheck
	fprintf(stderr,"ToolCheck(): ToolGroup leader found! Connecting to\n");
	fprintf(stderr,"ToolCheck(): group. Tool still expected.\n");
#      endif
     StartSpecialTool(Leader);
      }
    else
      {
	/* The tool is connecting to a port not currently recognized in the
	 * list of active tools --- this could well be a new tool. It could
	 * also be a bogus tool. Need to make sure by checking that the
	 * connection comes on the ToolSocket port. -bcl
	 */
	NeedNewPort = ( *Port == ToolSocket );
	if (!NeedNewPort)
	  {
#          if DEBUGToolCheck
	    fprintf(stderr,"ToolCheck(): Refusing, Unknown port.\n");
#          endif
	    ConnectAndRefuse( NewSocket );
	    close( NewSocket );
	    return FALSE;
	  }
      }
  }
  /* Okay, if we are here we know that we have a tool connecting to the
   * ToolSocket port and we were expecting a new connection, so we should
   * go ahead and fire up the new tool. We will also need to get another
   * port for ToolSocket (for the next tool connected). -bcl
  /* --- End Changes ---bcl */


  /* xyzzy - Handle connection to tool */
#if DEBUGToolCheck
  fprintf( stderr, "ToolCheck(): Doing conneciton work.\n" );
#endif

  TempTool->InConnection = NewSocket;
  TempTool->OriginalConnection = *Port;
  TempTool->TokenList = NULL;
  TempTool->TokenHolder = HOST_ID;
  TempTool->FindAbsoluteWindow = FALSE;
  TempTool->ArchiveBuf = NULL;
  TempTool->ArchiveSize = 0;
  TempTool->Archive2Buf = GetMem(0);
  TempTool->Archive2Size = 0;
  TempTool->SubDirtyCount = 0;
  TempTool->ThisToolDead = 0;
  TempTool->pending_reply = FALSE;

/*
**%%%%%%%%%%%%%%%%%% Begin Kantsios Addition
*/
if (TempTool->seqnum){
        (void) IntCircSeqMap(&(TempTool->seqMap));
        TempTool->ToolRequestNumber = 0;
        TempTool->ServerRequestNumber = 0;
}
        TempTool->LastSequenceNumber = 0;
/*
**%%%%%%%%%%%%%%%%%% End Kantsios Addition
*/
  if ( POLICY )
    TempTool->floor_timer = XtAppAddTimeOut(GetAppCon(), MinFloorTime,
					    FloorTimeOut, TempTool);

  if ( ProcessHostClientConnection(TempTool) >= 0 ) {
    AddTool( TempTool );
    if (TempTool->Leader == TRUE)
      UpdateMessage( "Tool %s has just been connected.", TempTool->ToolCommand );
    /* immedaitely swith token ----------------------------------------------
       GiveTokenTo(FindSubscriberByID(TempTool->ToolCreator), TempTool);
     ------------------------------------------------------------------------*/
  } else {
    ToolExpected = FALSE;
    return FALSE;
  }

  if( TempTool->join )
    InitializeObjects(TempTool);
  
  if ( (TempTool->ToolID=AssignToolID(TempTool)) == NoIDAvailable ){
    FatalError( "Too many tools connected" );
#  if DEBUGToolCheck
    fprintf( stderr, "ToolCheck(): Too many tools connected\n");
#  endif
    return;
  }
  /* Inform everybody else about the new tool */
  
#if DEBUGToolCheck
  fprintf( stderr, "ToolCheck(): The VisualIDList Contains %d items\n",
	  TempTool->NumVisualIDs);
  fprintf( stderr, "ToolCheck(): The VisualIDList elements are\n");
  {
    int i = TempTool->NumVisualIDs;
    while (i)
      printf ("%08x\n",TempTool->VisualIDList[TempTool->NumVisualIDs - i--]);
  }
#endif

  tmpHowMuchAuthData = TempTool->HowMuchAuthData;
  tmpNumVisualIDs = TempTool->NumVisualIDs;
  
  TempTool->HowMuchAuthData = SWAP_INT(TempTool->HowMuchAuthData,
				       FROMLOCAL(TempTool->ConnectPrefix.byteOrder));
  TempTool->NumVisualIDs = SWAP_INT(TempTool->NumVisualIDs,
				    FROMLOCAL(TempTool->ConnectPrefix.byteOrder));
  
  if (TempTool->Leader == TRUE)
    insertTool(FindSubscriberByID(HOST_ID), TempTool);
  for ( Who=FirstSubscriber(); Who != NULL; Who=NextSubscriber() ) {
    if (Who->SubscriberID == 0) {
      MulticastFlag[0]=FALSE;
      continue;
    }
    if( TempTool->join && ArchiveToBeSent(Who,TempTool) ){
      MulticastFlag[Who->SubscriberID]=FALSE;
      continue;
    }
  }
  MulticastByte(HS_AddNewTool );
  MulticastData((BYTE *)(TempTool), sizeof(Tool) );
  MulticastData((BYTE *)(TempTool->AuthData), tmpHowMuchAuthData );
  MulticastData((BYTE *)(TempTool->VisualIDList),
		((tmpNumVisualIDs) * sizeof(XID)) );
  {
    int  i;
    for ( i=0; i<256; i++ )
      MulticastFlag[i] = TRUE;
  }

  TempTool->HowMuchAuthData = tmpHowMuchAuthData;
  TempTool->NumVisualIDs = tmpNumVisualIDs;
  
  ToolExpected = FALSE;
  /* --- Begin  Changes ---bcl */
  /* If we started a new tool, the value in ToolSocket is no longer valid.
   * Hence we need to find the next connection to the X-Server which is
   * not in use and assign DisplayNumber and ToolSocket accordingly.
   */
  if (NeedNewPort)
  {
    DisplayNumber++;
    do
      {
	ToolSocket = OpenListener( htons(X_TCP_PORT+DisplayNumber),
				  4, &ListenerName );
      } while ( (ToolSocket < 0) && ( DisplayNumber++ <= XTV_HIGH_DISPLAY ) );
    XtAppAddInput( GetAppCon(), ToolSocket, XtInputReadMask, ToolCheck, NULL );
  }
  /* --- End Changes ---bcl */

  return FALSE;   /* Always!! */
}



/* handling zombies */
sig_child()
{
  int pid;
  union wait	*status;
  
  Tool *WhatTool;


  while ( (pid=wait3(&status, WNOHANG, (struct rusage *) 0)) > 0){
#  if DEBUGsig_child
    fprintf( stderr, "sig_child(): process(pid) %dhave terminated\n", pid);
#  endif
    for ( WhatTool=FirstTool(); WhatTool != NULL; WhatTool=NextTool() ) {
      if (WhatTool->ToolPID == pid){
#      if DEBUGsig_child
	fprintf( stderr, "sig_child(): CleaningUp Tool %s\n whose pid is %d\n", WhatTool->ToolCommand,
		WhatTool->ToolPID);
#      endif
	if (!CleanUpToolFlag)
	  /*
	   *  CleanUpTool(WhatTool);
	   */
	  CleanUpToolGroup(WhatTool);
	
      }
    }
  }
}

/*
 * Begin Murray Addition
 */

void
#ifdef __STDC__
startSubscr(Tool *tool)
#else
startSubscr(tool)
Tool *tool;
#endif
{
  if (GetLocalSubID() != HOST_ID)
    return;
  
  start(tool->subscrList);
}

void
#ifdef __STDC__
forthSubscr(Tool *tool)
#else
forthSubscr(tool)
Tool *tool;
#endif
{
  if (GetLocalSubID() != HOST_ID)
    return;
  
  forth(tool->subscrList);
}

int
#ifdef __STDC__
offrightSubscr(Tool *tool)
#else
offrightSubscr(tool)
Tool *tool;
#endif
{
  if (GetLocalSubID() != HOST_ID)
    return 1;
  
  return offright(tool->subscrList);
}

void
#ifdef __STDC__
addSubscr(Tool *tool, Subscriber *newSubscr)
#else
addSubscr(tool, newSubscr)
Tool *tool;
Subscriber *newSubscr;
#endif
{
  if (GetLocalSubID() != HOST_ID)
    return;
  
  addItem(tool->subscrList, newSubscr);
}

void
#ifdef __STDC__
deleteSubscr(Tool *tool)
#else
deleteSubscr(tool)
Tool *tool;
#endif
{
  if (GetLocalSubID() != HOST_ID)
    return;
  
  deleteItem(tool->subscrList);
}

Subscriber *
#ifdef __STDC__
getSubscr(Tool *tool)
#else
getSubscr(tool)
Tool *tool;
#endif
{
  if (GetLocalSubID() != HOST_ID)
    return (Subscriber *) 0L;
  
  return ((Subscriber *) getItem(tool->subscrList));
}

void
#ifdef __STDC__
findSubscr(Tool *tool, Subscriber *targetSubscr)
#else
findSubscr(tool, targetSubscr)
Tool *tool;
Subscriber *targetSubscr;
#endif
{
#if DEBUGfindSubscr
  fprintf( stderr,"findSubscr: strating\n");
#endif
  if (GetLocalSubID() != HOST_ID) {
#  if DEBUGfindSubscr
    fprintf( stderr,"findSubscr: leaving1\n");
#  endif
    return;
  }

  for (start(tool->subscrList);
       (!offright(tool->subscrList) && (((Subscriber *) getItem(tool->subscrList)) != targetSubscr));
       forth(tool->subscrList))
    ;
#if DEBUGfindSubscr
  fprintf( stderr,"findSubscr: leaving2\n");
#endif
}

/*
 * This is a LOCAL function that removes the PTRLIST * of
 * subscribers for the given tool. This should 
 */
void
#ifdef __STDC__
deleteAllSubscr(Tool *tool)
#else
deleteAllSubscr(tool)
Tool *tool;
#endif
{
  if (GetLocalSubID() != HOST_ID)
    return;

  /*
   * This loop frees the nodes, BUT NOT the items they point to.
   */

  for (start(tool->subscrList); !offright(tool->subscrList); forth(tool->subscrList))
    deleteItem(tool->subscrList);

  /*
   * This gets rid of the list header.
   */
  free(tool->subscrList);
}

/*
 * End Murray Addition
 */

/*
 *   CloseAllTools()
 *
 *   Go through the entire list of tools and shut 'em off
 *   one at a time.
 *
 *   Last Modified: AKR 05/09 changed incrementer in the for loop to NextTool
 *                  from FirstTool.
 *   RETURN VALUE:   void
 */

void CloseAllTools()
{
  Tool   *WhichTool;


  /* Yes, we want to re-use FirstTool since we're going to change
   * the contents of the subscriber list inside the loop.
   */

  for ( WhichTool=FirstTool(); WhichTool != NULL; WhichTool=NextTool() ) {
    if ( GetMode() ) {
      shutdown( WhichTool->InConnection, 2 );
      XtRemoveInput( WhichTool->InID );
      XtRemoveInput( WhichTool->InExceptID );
      
    }
    DeleteTool( WhichTool );
  }
}

/*
 *   CleanUpAllTools()
 *
 *   Go through the entire list of tools and shut 'em off
 *   one at a time.
 *
 *   RETURN VALUE:   void
 */
void CleanUpAllTools()
{
  Tool   *WhichTool;
  
  while ( (WhichTool=FirstTool()) != NULL){
    if ( GetMode() ) {
#    if DEBUGCleanUpAllTools
      fprintf(stderr, "CleanUpAllTools: cleaning tool %d\n", WhichTool->ToolID);
#    endif
      /*
       * CleanUpTool( WhichTool );
       */
      CleanUpToolGroup( WhichTool );
    } 
  }
}
/* --- Begin Change ---bcl */
/* ------------------------------------------------------------------------- */
/*
 *   CleanUpToolGroup()
 *
 *   Find all tools under a given ToolGID and shut them off.
 *
 *   RETURN VALUE:   void
 */
/* ------------------------------------------------------------------------- */

CleanUpToolGroup(FriedTool)
Tool   *FriedTool;
{
  Tool   *WhichTool;
  BOOL Found = TRUE;
  while (Found)
    {
      Found = FALSE;
      for ( WhichTool=FirstTool(); WhichTool != NULL; WhichTool=NextTool() ) {
      
#    if DEBUGCleanUpToolGroup
      fprintf(stderr, "CleanUpToolGroup: examining tool %d\n", WhichTool->ToolID);
#    endif
      if (WhichTool->ToolGID == FriedTool->ToolGID) {
#      if DEBUGCleanUpToolGroup
	fprintf(stderr, "CleanUpToolGroup: Cleaning Up tool %d whose GID is %d\n",
		WhichTool->ToolID, WhichTool->ToolGID);
#      endif
	CleanUpTool( WhichTool );
	Found = TRUE;
#      if DEBUGCleanUpToolGroup
	fprintf(stderr, "CleanUpToolGroup: cleaned Up Tool\n");
#      endif
	break;
      }
    }
  }
#if DEBUGCleanUpToolGroup
  fprintf(stderr, "CleanUpToolGroup: Leaving \n");
#endif
}




/* ------------------------------------------------------------------------- */
/*
 *   StartSpecialTool()
 *
 *   Start up a new connection for same tool.
 *
 */
/* --- Begin Change ---bcl */
/* Added the leader for the "special tool." */

StartSpecialTool(Leader)
  Tool *Leader;
/* --- End Change ---bcl */
{
  Tool    *TempTool2;              /* Pointer to temporary tool structure */

#if DEBUGStartSpecailTool
  fprintf( stderr, "StartSpecialTool(%d): starting\n",GID );
#endif

  TempTool2=(Tool *)GetMem(sizeof(Tool));
  memcpy(TempTool2,TempTool,sizeof(Tool));
  TempTool = TempTool2;
  TempTool->Leader = FALSE;
  /* --- Begin Change ---bcl */
  TempTool->ToolGID = Leader->ToolGID;
  memcpy(TempTool->ToolCommand,Leader->ToolCommand,sizeof(Leader->ToolCommand));

#if DEBUGStartSpecailTool
  fprintf( stderr, "StartSpecialTool(%d): finishing\n",GID );
#endif
  /* --- End Change ---bcl */

}

StopToolGroup(FriedTool)
Tool   *FriedTool;
{
  Tool   *WhichTool;
  for ( WhichTool=FirstTool(); WhichTool != NULL; WhichTool=NextTool() ) {
    
#  if DEBUGStopToolGroup
    fprintf(stderr, "StopToolGroup: examining tool %d\n", WhichTool->ToolID);
#  endif
    if (WhichTool->ToolGID == FriedTool->ToolGID) {
#    if DEBUGStopToolGroup
      fprintf(stderr, "StopToolGroup: Stopping tool %d whose GID is %d\n",
	      WhichTool->ToolID, WhichTool->ToolGID);
#    endif

      if ( SendByte(GetHostConnection(), SH_StopTool)==ERROR) {
	EndSession();
	return;
      }
      if ( SendByte(GetHostConnection(), WhichTool->ToolID)==ERROR){
	EndSession();
	return;
      }
/* added by wahab */
	WhichTool->InCount = 0;
        WhichTool->OutCount = 0;
#    if DEBUGStopToolGroup
      fprintf(stderr, "StopToolGroup: Sent  SH_StopTool\n");
#    endif
    }
  }
#if DEBUGStopToolGroup
  fprintf(stderr, "StopToolGroup: Leaving \n");
#endif
}

InitToolStruct(options, creator)
char *options;
BYTE creator;
{
 char           *p;

        /* find all the option tokens */

        TempTool->xcolor  = XCOLOR;
        TempTool->join    = JOIN;
        TempTool->seqnum    = SEQNUM;
        TempTool->resize  = RESIZE;
        TempTool->verbose = VERBOSE;
 	OutputCount = 10;
        TempTool->popup  = POPUP;
        TempTool->vs_in_out  = VS;

        strcpy(TempTool->ToolOptions, options);
#   if DEBUGStartNewTool
	printf("Tool Options %s\n", options);
#   endif
	printf("Tool Options %s\n", options);

        p = strtok( options, " ");
        do {
          if(p){
            if(strlen(p) == 2 && p[0] == '-'){
              switch(p[1]){
                case 'C':
                  TempTool->xcolor = 0;
                  break;
                case 'J':
                  TempTool->join = 0;
                  break;
                case 'Q':
                  TempTool->seqnum = 0;
                  break;
	        case 'Z':
                  TempTool->resize = 0;
                  break;
	        case 'O':
                  TempTool->verbose = 0;
                  break;
	        case 'P':
                  TempTool->popup = 0;
                  break;
	        case 'V':
                  TempTool->vs_in_out = 0;
                  break;
              }
            }
          p = strtok('\0', " ");
          }
        }while(p);

	ToolExpected = TRUE;
	TempTool->Leader = TRUE;
	TempTool->ToolGID = ++GroupID;
	TempTool->InCount = 0;
	TempTool->OutCount = 0;
	TempTool->ToolPID = 0;
	TempTool->ToolCreator = creator;
	TempTool->AskedToGetToken = FALSE;

/*
 * Begin Murray Addition
 */

        TempTool->subscrList = ptrNew();

        /*
 * End Murray Addition
 */

}
