/*
 * Copyright (c) 1991, 1993 Old Dominion University.
 * Copyright (c) 1991, 1993 University of North Carolina at Chapel Hill.
 * All rights reserved.
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  Old Dominion University and the University of North
 * Carolina at Chapel Hill make no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 */

/*
	tool.h
	Declarations for tool management module

*/

#ifndef	TRUE
#define	TRUE	1
#endif

#ifndef FALSE
#define	FALSE	0
#endif

#ifndef WINDOW
#define WINDOW          Window
#endif


#ifndef TOOL__H
#define TOOL__H

#include <X11/Xmd.h>
#include <X11/Xproto.h>
#include <X11/Intrinsic.h>
#include "network.h"
#include "objects.h"
/*
 * Begin Murray Addition
 */
#include "ptrlists.h"
#ifndef SUBSCRIBER__H
#include "subscriber.h"
#endif
/*
 * End Murray Addition
 */
/*
** Begin Kantsios addition
*/
#define         MAX_SEQ_TABLE   1000

typedef struct _entry {
        CARD16  Server_SN;
        CARD16  Tool_SN;
} Entry_;
 
 
typedef struct _circ_seq_map {
        Entry_   table[MAX_SEQ_TABLE];
        int     head;
        int     tail;
        BOOL     first;
	BOOL	MakeRoomFlag;
	CARD16  MakeRoomSeqNum;
} CircSeqMap;

/*
** End Kantsios addition
*/

#ifdef SVR4
#include <sys/file.h>
#endif /* SVR4 */

#include "token.h"

#define	TOOL_CMD_SIZE	200
#define	TOOL_OPT_SIZE	200

#define	HOST_NAME_SIZE	200

typedef struct _Tool {

	/* Stuff supplied by the host.  All of this is in the host's native
		 * data format (as specified by HostByteOrder) and will need to be
		 * converted by the local machine before use.  Note that any XIDs
		 * stored will not need conversions as the bits will already be
		 * appropriately screwed up.
		 */
	char			ToolCommand[100];	/* Command used to start tool */
        char                    ToolOptions[100];       /* Options used to start tool */
        BYTE                     join;                   /* 0 means late joiners don't get tool */
        BYTE                     xcolor;                 /* 0 means not to translate colors */
        BYTE                    resize;                 /* 0 means not to resize window */
        BYTE                    seqnum;                 /* 0 means not to use cir. que */
        BYTE                     verbose;                /* 0 means no packet statistics */
        BYTE                     popup;                 /* 0 means  not intervene in popups & fix */
	BYTE			ToolID;			/* ID number assigned to tool */
	BYTE			vs_in_out;		/* in or oute Virtual Screen */
	int			ToolPID;		/* Process ID of tool */
	xConnClientPrefix	ConnectPrefix;		/* Connection prefix */
	CARD16			HowMuchAuthData;	/* Number of bytes of AuthData */
	BYTE			*AuthData;		/* Pointer to authorization data */

	xConnSetup		HostConnectInfo;	/* First Part of connect info */
	xWindowRoot		HostRootInfo;		/* Host root window information */
        xPixmapFormat           *HostPixmapFormats;     /* Host pixmap formats infomation */
	CARD16			NumVisualIDs;		/* How many visual IDs in the list */
	XID			*VisualIDList;		/* Array of visual IDs */
        CARD16                  NumColorCells;          /* How many color cells in the list */
        CARD32                  *ColorCellList;         /* Array of color cells */
	BYTE			BitsInHostMask;		/* Number of bits in Resource ID mask */
	BOOL			Leader; 		/* True if it is the tool first connection */

	/* These fields are used only at the host's end but are passed along
		 * anyway and replaced by the subscriber.
		 */
	Socket			InConnection;		/* Incoming connection from client */
	Socket			OutConnection;		/* Outgoing connection to server */
	TokenNode		*TokenList;		/* Token waiting list */
	BYTE			TokenHolder;		/* Subscriber ID of token holder */
	BYTE			ToolGID;		/* Tool Group ID for muli-connection tools */
	Socket                  OriginalConnection;     /* Port assigned to tool group. */

	/* Stuff defined locally during the connection process
		 */
	xConnSetup		LocalConnectInfo;	/* Local Connect Info */
	xWindowRoot		LocalRootInfo;		/* Local Root info */
        xPixmapFormat           *LocalPixmapFormats;    /* Local Pixmap Formats Info  */
	CARD16			LocalNumVisualIDs;	/* Number of local visual IDs */
	XID			*LocalVisualIDList;	/* List of Local Visual IDs */
        CARD16                  LocalNumColorCells;     /* Number of local color cells */
        CARD32                  *LocalColorCellList;    /* List of local Color cells */
	BYTE			BitsInLocalMask;	/* Number of bits in local RID mask */

	/* Stuff used for determining input availability and exceptions.
		 */
	XtInputId		InID;			/* ID for available client input */
	XtInputId		OutID;			/* ID for available server input */
	XtInputId		InExceptID;		/* ID for client exceptions */
	XtInputId		OutExceptID;		/* ID for server exceptions */

	/* Stuff used by the host in translating events
		 */

	CARD32			LastTimeStamp;		/* Last known time stamp */
	int			pending_reply;		/* Do't switch token if TRUE */
	int			server_pending_reply;
	int			InCount;		/* number of Incoming packets */
	int			OutCount;		/* number of Outgoing packets */

        /* Stuff used to calculate the correct position of Pop-up menu
                */

        BOOL                    FindAbsoluteWindow;
        Window                  AbsoluteWindow;

	/* Stuff used to archive packets */

	ToolState		*ToolInfo;
	char			*ArchiveBuf;		/* pointer to packet buffer */
	off_t			ArchiveSize;		/* size of the archive */
	char			*Archive2Buf;		/* pointer to packet buffer to be sent after an expose event occurs */
	off_t			Archive2Size;		/* size of archive2 */
	int			SubDirtyCount;		/* number of new subscribers catching up on this tool */
	int			ThisToolDead;		/* flag to indicate this tool is dead */

	/* Stuff needed to get archives */

	int			CatchUpMode;		/* flag to indicate that this subscriber is in CatchUpMode for this tool */

	/* AKR 05/08 Stuff to do floor policies */

	XtIntervalId		floor_timer;		/* handle for timer routine for this tool */

	BYTE                    ToolCreator;            /* Subscriber ID of creator */
	BOOL                    AskedToGetToken;            /* sent asking for token */
	/*
	 * Begin Murray Addition
	 */
	PTRLIST                *subscrList;
	/*
	 * A list of subscribers who use this tool. 
	 * This list is valid only on the host.
	 */

	/*
	 * End Murray Addition
	 */
        /*
         * Begin Kantsios Addition
         */
        CircSeqMap              seqMap;
        CARD16                  ToolRequestNumber;   /* Tool request # */
        CARD16                  ServerRequestNumber; /* Local Server Req. # */
        CARD16                  LastSequenceNumber;/* last seq numb forward to server */

        /*
         * End Kantsios Addition
         */

} Tool;

typedef struct _ToolNode {
	Tool			*ToolS;
	struct _ToolNode	*Next;
} ToolNode;

struct {
  BYTE j;                        /* 0 = late join on, 1 = late join off */
  BYTE z;                        /* 0 = resize, 1 = no resize */
  BYTE q;                        /* 0 = no cir que,  1 = use cir. que */
  BYTE c;                        /* 0 = color translations, 1 = no color translations */
  BYTE f;                        /* 0 = normal, 1 = fast */
  BYTE o;                        /* 0 = show statistics, 1 = suppress statistics */
  BYTE p;                        /* 1 = allow popup adjustment, 0 = disable popup adj */
  BYTE v;                        /* 0 = inside VS , 1 = outside VS */
} Option;

Tool *FirstTool();
Tool *NextTool();
void AddTool();
void DeleteTool();
Tool *FindTool();
void RebuildToolList();
void InitializeTool();
XtInputCallbackProc ToolCheck();
XtInputCallbackProc ConfirmFryTool();
void CloseAllTools();
BYTE AssignToolID();
void SetToolID();
void ClearToolID();
Tool *FindToolByID();
void StartLocalTool();
void StartRemoteTool();
void CleanUpTool();
void CleanUpAllTools();

#ifndef NoIDAvailable
#define	NoIDAvailable		255
#endif

#endif /* TOOL__H */
