/*
 * xtv_interface -- XTV's interface to XTVSIS
 */

/*
 * Copyright (c) 1991, 1993 Old Dominion University.
 * Copyright (c) 1991, 1993 University of North Carolina at Chapel Hill.
 * All rights reserved.
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  Old Dominion University and the University of North
 * Carolina at Chapel Hill make no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 */

/*
 * written by : Sun W. Chang
 * date       : April, 1992
 * location   : Old Dominion University, Norfolk Virginia
 */
/* modifications made by  :  Pam Skillman
*  date                   :  December, 1992
*  location               :  Old Dominion University, Norfolk Virginia
*/


#include <stdio.h>
#include <signal.h>
#include <setjmp.h>
#include <sys/types.h>
#include <errno.h>
#include <sys/time.h>
#include "memory.h"
#include "network.h"

#include	"xtvsis.h"
#include	"subscriber.h"
#include	"tool.h"

struct xtvsis_addr {	/* maintains all XTVSIS's internet addresses */
	struct sockaddr_in	addr;
	struct xtvsis_addr	*next;
} *xslist;

/*  begin modifications by Pam Skillman  */
struct session {
        char                    *session_id;
        char                    *info;
        struct sockaddr_in      addr;
        time_t                  caltime;
        struct session          *next;
} *slist;
/*  end modifications by Pam Skillman  */

char	session_id[SESSION_ID_SIZE]; /* hostname:session_number:time_started */

int	sockfd;

struct sockaddr_in	myaddr;

int	accept_XTVSIS_LIST = TRUE;	/* turned on only at the beginning */

int	xtvsis_service();

int 	resend_REGISTRATION();

/*  begin modifications by Pam Skillman  */
int     XTVLIST_VERSION, last_version_from_xtvsis;

char    LastVersion[VERSION_SIZE];
/*  end modifications by Pam Skillman  */

extern XtAppContext	AppCon;


/**************************************************************************
 * xtvsis_interface(session_number)
 *	builds session_id, XTVSIS list, creates socket and sends first
 *	REGISTRATION, REQUEST_XTVSIS_LIST and set the alarm for 
 *	sending REGISTRATION at an interval.
 *
 * calling routines : main() -- when XTV starts.
 *************************************************************************/
xtvsis_interface(session_number)
int	session_number;
{
	Subscriber	*first_subscriber;
	char		time_started[80];
	time_t		t;
	char *ptr;


	/*
	 * build session_id
	 */
	time(&t);
	strcpy(time_started, ctime(&t));
	/* stip year and seconds from date (wahab) */
	ptr = rindex(time_started,':');
	*ptr = '\0';
	/* remove '\n' 
	time_started[strlen(time_started)-1] = '\0';  
        */
	if (first_subscriber=FirstSubscriber())
		sprintf(session_id, "%s:%d:%s", first_subscriber->FullHostName,
					session_number, time_started);
	else
		sprintf(session_id, "UNKNOWN_HOST:%d:%s",
					session_number, time_started);


	build_xslist();		/* build XTVSIS list */

	if (xslist == NULL) 		/* no XTVSIS */
		return;			/* no further processing for XTVSIS */

	if (create_socket_for_xtvsis() < 0) {	/* if creating socket fails, */
		xslist = NULL;			/* set xslist to NULL to maintain */
		return;				/* the integrity */
	}

/*  begin modifications by Pam Skillman  */
	send_ID_LIST_REQUEST_TO_ALL();
/*  end modifications by Pam Skillman  */

	send_REGISTRATION();

	send_REQUEST_XTVSIS_LIST();

	alarm(TIMEOUT_REGISTRATION);
	signal(SIGALRM, resend_REGISTRATION);

	XtAppAddInput(AppCon, sockfd, XtInputReadMask, xtvsis_service, NULL);
}


/**************************************************************************
 * build_xslist()
 *	builds XTVSIS list from "xtvsis.list" file.
 * 	"xtvsis.list" contains hostname:port per line.
 *
 *	calling routines : xtvsis_interface() -- when executed
 *			   xtvsis_service() -- when new XTVSIS_LIST packet 
 *					is received from a XTVSIS.
 *************************************************************************/
build_xslist()
{
	FILE			*fp;
	char			aline[80], *c1, *c2;
	char			hostname[HOSTNAME_SIZE];
	int			port;
	struct hostent		*gethostbyname();
	struct hostent		*host;
	struct xtvsis_addr	*p;

	xslist = NULL;

	if ((fp=fopen(XTVSIS_LIST_file, "r")) == NULL) {
		return;
	}

	for (; fgets(aline, 80, fp) ;) {
/*  begin modifications by Pam Skillman  */
/*  first line in file is version number  */

		if (strstr(aline, ":") == NULL)	{
		  XTVLIST_VERSION = atoi(aline);

/*  end modifications by Pam Skillman  */
			continue;  }

		for (c1=aline, c2=hostname; *c1!=':' ; c1++,c2++)
			*c2 = *c1;
		*c2 = '\0';
		port = atoi(++c1);

		/*
	 	 * hostname can be either dotted string of inet address or 
		 * hostname
	 	*/
		p = (struct xtvsis_addr *) malloc(sizeof(struct xtvsis_addr));
		bzero((char *) &(p->addr), sizeof(p->addr));
		p->addr.sin_family	= AF_INET;
		if ((host=gethostbyname(hostname)) != NULL) 
			bcopy(host->h_addr, &(p->addr.sin_addr), host->h_length);
		else
			p->addr.sin_addr.s_addr = inet_addr(hostname);
		p->addr.sin_port	= htons(port);
		p->next			= NULL;

		if (xslist) {
			p->next = xslist;
			xslist = p;
		}
		else
			xslist = p;
	}

	fclose(fp);

}


/**************************************************************************
 * create_socket_for_xtvsis()
 *	creates socket to be used in XTVSIS interface.
 *
 *	calling routines : xtvsis_interface() -- at the beginning.
 *************************************************************************/
create_socket_for_xtvsis()
{
	int	length;


	if ((sockfd=socket(AF_INET, SOCK_DGRAM, 0)) < 0) 
		return(-1);

	bzero((char *) &myaddr, sizeof(myaddr));
	myaddr.sin_family	= AF_INET;
	myaddr.sin_addr.s_addr	= htonl(INADDR_ANY);
	myaddr.sin_port		= htons(0);

	if (bind(sockfd, (struct sockaddr *) &myaddr, sizeof(myaddr)) < 0)
		return(-1);

	/*
	 * bind call itself does not fill myaddr structure with values
	 * getsockname call is made to fill myaddr
	 */
	length = sizeof(myaddr);
	if (getsockname(sockfd, &myaddr, &length) < 0)
		return(-1);
}


/**************************************************************************
 * send_REGISTRATION()
 *	sends REGISTRATION packet to all XTVSIS's in the list at the
 *	beginning.
 * 
 *	calling functions : xtvsis_interface() -- at the beginning
 *
 * 	REGISTRATION packet :
 *		REGISTRATION 
 *		hostname:session_number:time_started END_OF_SESSION_ID
 *		chairman:subject:participant1/participant2/..../ END_OF_INFO
 *		time END_OF_SESSION
 **************************************************************************/
send_REGISTRATION()
{
	char			packet[PACKET_SIZE];
	char			participants_list[PACKET_SIZE];
	Subscriber		*participant;
	struct xtvsis_addr	*p;
	char			timebytes[sizeof(time_t)+1];
	time_t			ht,t;
	extern char		Subject[];


	/*
	 * build REGISTRATION packet
	 */
	participants_list[0] = '\0';
	for (participant = FirstSubscriber(); participant ;
				participant = NextSubscriber()) {
		strcat(participants_list, participant);
		strcat(participants_list, "/");
	}

	ht = time(NULL);
	t = htonl(ht);
	bcopy((char *) &t, timebytes, sizeof(time_t));
	timebytes[sizeof(time_t)] = '\0';
	sprintf(packet, "%c%s%c%s:%s:%s%c%s%c", REGISTRATION, session_id,
			END_OF_SESSION_ID, ChairmanName(), Subject,
			participants_list, END_OF_INFO,
			timebytes, END_OF_SESSION); 

	/*
	 * send the REGISTRATION packet to all XTVSIS in xslist
	 */
	for (p=xslist; p ; p=p->next)
		sendto(sockfd, packet, strlen(packet), 0,
			(struct sockaddr *) &(p->addr), sizeof(p->addr));
}


/**************************************************************************
 * send_REQUEST_XTVSIS_LIST()
 *	sends REQEUST_XTVSIS_LIST packet to all XTVSIS's in the list.
 *
 *	calling routines : xtvsis_interface() -- at the beginning
 *
 *	REQUEST_XTVSIS_LIST packet :
 *		REQUEST_XTVSIS_LIST
 *************************************************************************/
send_REQUEST_XTVSIS_LIST()
{
	char			packet = REQUEST_XTVSIS_LIST;	
	struct xtvsis_addr	*p;

	/*
	 * send the REQUEST_XTVSIS_LIST packet to all XTVSIS in xslist
	 */
	for (p=xslist; p ; p=p->next)
		sendto(sockfd, &packet, 1, 0, 
			(struct sockaddr *) &(p->addr), sizeof(p->addr));
}


/**************************************************************************
 * xtvsis_service(client_data, id)
 *	receives packet and calls proper routine according to the packet
 *	type. accept_XTVSIS_LIST is set to TRUE at the beginning but
 *	set to FALSE as soon as the first XTVSIS_LIST packet is received 
 *	in order to discard any subsequent XTVSIS_LIST from other XTVSIS's.
 *
 *	calling routines : from XtInputReadMask -- when input is signaled
 *				in sockfd.
 *
 * 	XTVSIS_LIST packet :
 *		XTVSIS_LIST
 *		hostname:port\n
 *		hostname:port\n
 *		.
 *		.
 *		.
 *
 * 	ARE_YOU_ALIVE packet :
 *		ARE_YOU_ALIVE
 *************************************************************************/
xtvsis_service(client_data, id)
caddr_t		client_data;
XtInputId	*id;
{
	int			rc;
	int			fromlen;
	char			packet[PACKET_SIZE];
	struct sockaddr_in	fromaddr;

	fromlen = sizeof(fromaddr);
	if ((rc=recvfrom(sockfd, packet, PACKET_SIZE, 0, 
			(struct sockaddr *) &fromaddr, &fromlen)) <0)
		return;

	packet[rc] = '\0';

	switch (packet[0]) {

		case XTVSIS_LIST :
			if (accept_XTVSIS_LIST) {
				FILE	*fp;

				accept_XTVSIS_LIST = FALSE;

				if ((fp=fopen(XTVSIS_LIST_file, "w")) != NULL) {
					fputs(&packet[1], fp);
					fclose(fp);
					build_xslist();
				}
			}

			break;

		case ARE_YOU_ALIVE :
			send_REGISTRATION();
			break;

/*  begin modifications by Pam Skillman  */
                case ID_LIST :
                        bcopy(&packet[1], LastVersion, sizeof(LastVersion));
                        last_version_from_xtvsis = atoi(LastVersion);
/*  only update our XTVSIS_LIST if our version # is lower  */

                        if (last_version_from_xtvsis > XTVLIST_VERSION) {
                          FILE       *fp;
                            XTVLIST_VERSION = last_version_from_xtvsis;
                            if ((fp=fopen(XTVSIS_LIST_file, "w")) != NULL)
			      {
                                        fputs(&packet[1], fp);
                                        fclose(fp);
                                        build_xslist();
                                      }
			}
                        break;
/*  end modifications by Pam Skillman  */

		default :	/* ignore all other packet types */
			break;

	}
}

/**************************************************************************
 * send_TERMINATION_NOTICE()
 *	sends TERMINATION_NOTICE to all XTVSIS's in the list when finally
 *	XTV session is terminating.
 *
 *	calling routines : main() -- when terminating XTV session.
 *
 * 	TERMINATION_NOTICE packet :
 *		TERINATION_NOTICE
 *      	hostname:session_number:time_started END_OF_SESSION_ID
 *************************************************************************/
send_TERMINATION_NOTICE()
{
	char			packet[SESSION_ID_SIZE+1];
	struct xtvsis_addr	*p;

	/*
	 * build TERMINATION_NOTICE packet
	 */
	sprintf(packet, "%c%s%c", TERMINATION_NOTICE, session_id,
				END_OF_SESSION_ID);

	/*
	 * send the TERMINATION_NOTICE packet to all XTVSIS in xslist
	 */
	for (p=xslist; p ; p=p->next)
		sendto(sockfd, packet, strlen(packet), 0,
			(struct sockaddr *) &(p->addr), sizeof(p->addr));
}

resend_REGISTRATION()
{
	send_REGISTRATION();
	alarm(TIMEOUT_REGISTRATION);
	signal(SIGALRM, resend_REGISTRATION);
}
/************************************************************************
 * send_ID_LIST_REQUEST(fromaddr)
 *      sends ID_LIST_REQUEST to fromaddr
 *
 *      calling routines : send_ID_LIST_REQUEST_TO_ALL()
 *
 *      ID_LIST_REQUEST packet :
 *              ID_LIST_REQUEST
 *              version
 *
 ****************************************************************************/
send_ID_LIST_REQUEST(fromaddr)
struct sockaddr_in      fromaddr;
{
        int     fd;
        char    c, *s;
        char    packet[PACKET_SIZE];

        packet[0] = ID_LIST_REQUEST;
        sprintf(LastVersion,"%d",XTVLIST_VERSION);
        bcopy(LastVersion, &packet[1], sizeof(LastVersion));

        packet[sizeof(LastVersion)+1] = '\0';

        sendto(sockfd, packet, strlen(packet), 0,
               (struct sockaddr *) &fromaddr, sizeof(fromaddr));

      }
/************************************************************************
 * send_ID_LIST_REQUEST_TO_ALL()
 *      sends ID_LIST_REQUEST to all daemons when we start up
 *
 *      calling routines : main()  
 *
 *      called routines  : send_ID_LIST_REQUEST(fromaddr)  --  send a request
 *                                    to all XTVSIS daemons.
 *
 ****************************************************************************/
send_ID_LIST_REQUEST_TO_ALL()
{
        FILE                    *fp;
        char                    c, *s, localhost[80];
        char                    aline[80], *c1, *c2;
        char                    hostname[HOSTNAME_SIZE];
        int                     port, n;
        struct hostent          *gethostbyname();
        struct hostent          *host, *hp;
        struct sockaddr_in      fromaddr, ouraddr;


            if ((fp=fopen(XTVSIS_LIST_file, "r")) == NULL) {
              return;
            }
            for (; fgets(aline, 80, fp) ;) {
              if (strstr(aline, ":") != NULL)  {

                for (c1=aline, c2=hostname; *c1!=':' ; c1++,c2++)
                        *c2 = *c1;
                *c2 = '\0';
                port = atoi(++c1);
                /*
                 * hostname can be either dotted string of inet address or
                 * hostname
                */

                bzero((char *) &fromaddr, sizeof(fromaddr));
                fromaddr.sin_family      = AF_INET;

                if ((host=gethostbyname(hostname)) != NULL)
                        bcopy(host->h_addr, &(fromaddr.sin_addr), host->h_length);
                else
                     fromaddr.sin_addr.s_addr = inet_addr(hostname);
                fromaddr.sin_port        = htons(port);

                gethostname(localhost, HOSTNAME_SIZE);

                if ((hp=gethostbyname(localhost)) == NULL)  {
                    continue;
                }
                bcopy(hp->h_addr, &(ouraddr.sin_addr), hp->h_length);

/*  We want to send to ALL daemons, even our own     */

/*                if ((n = strcmp(fromaddr.sin_addr,ouraddr.sin_addr)) !=0)  */
                  send_ID_LIST_REQUEST(fromaddr);
              }
	    }
        fclose(fp);
      }
/*  end modifications by Pam Skillman  */
